/*
 * cp40cmd_dump.c --
 *
 *	Source code for command to dump CP40 photons position as a FITS
 *	image.
 *
 * Copyright and warranty:
 *	Copyright (c) 1998 Eric THIEBAUT.
 *	
 *	This program is  free  software;  you  can  redistribute  it and/or
 *	modify it under the terms of  the  GNU  General  Public  License as
 *	published by the Free Software Foundation; either version 2  of the
 *	License, or (at your option) any later version.
 *	
 *	This program is distributed in the hope that it will be useful, but
 *	WITHOUT  ANY  WARRANTY;  without  even  the  implied   warranty  of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.   See  the GNU
 *	General Public License for more details.
 *	
 *	You should have received a copy of the GNU  General  Public License
 *	along with this program (see file ``LICENSE'' at the  top directory
 *	of  this  distribution);  if  not,  write  to  the   Free  Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *	
 *	To contact me, mail at: Eric THIEBAUT
 *				Observatoire de Lyon
 *				9, avenue Charles Andre
 *				F-69561 Saint Genis Laval Cedex
 *	or e-mail at:           thiebaut@obs.univ-lyon1.fr
 *
 * History:
 *	$Id$
 *	$Log$
 *_____________________________________________________________________________
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

/*---------------------------------------------------------------------------*/
#undef LITTLE_ENDIAN
#define	LITTLE_ENDIAN 1234

#undef BIG_ENDIAN
#define	BIG_ENDIAN    4321

#undef PDP_ENDIAN
#define	PDP_ENDIAN    3412

static int ByteOrder(void);

static int ByteOrder(void)
{
    int i= 1;

    if (sizeof(char) != 1 || sizeof(short) != 2 || sizeof(int) != 4
	    || sizeof(long) != 4 || sizeof(float) != 4
	    || sizeof(double) != 8) {
	fprintf(stderr, "failed assumption about data type size(s)\n");
	exit(1);
    }

    return *(char *)&i ? LITTLE_ENDIAN : BIG_ENDIAN;
}
/*---------------------------------------------------------------------------*/
static void SwapByte(char *ptr, int elsize, long number);
static void SwapByte(char *ptr, int elsize, long number)
{
    long i;
    char c;
    
    if (elsize==1) {
	return;
    }
    number*= elsize;
    if (elsize==2) {
	for (i=0 ; i<number ; i+=2) {
	    c= ptr[i]; ptr[i]= ptr[i+1]; ptr[i+1]= c;
	}
    } else if (elsize==4) {
	for (i=0 ; i<number ; i+=4) {
	    c= ptr[i];   ptr[i]=   ptr[i+3]; ptr[i+3]= c;
	    c= ptr[i+1]; ptr[i+1]= ptr[i+2]; ptr[i+2]= c;
	}
    } else if (elsize==8) {
	for (i=0 ; i<number ; i+=8) {
	    c= ptr[i];   ptr[i]=   ptr[i+7]; ptr[i+7]= c;
	    c= ptr[i+1]; ptr[i+1]= ptr[i+6]; ptr[i+6]= c;
	    c= ptr[i+2]; ptr[i+2]= ptr[i+5]; ptr[i+5]= c;
	    c= ptr[i+3]; ptr[i+3]= ptr[i+4]; ptr[i+4]= c;
	}
    } else {
	fprintf(stderr, "bad element size\n");
	exit(1);
    }
}

/*---------------------------------------------------------------------------*/

#define FLAG_RAW  1
#define FLAG_SWAP 2

typedef struct RawFrame RawFrame;
struct RawFrame {
    long  number;	/* Frame number. */
    long  flux;		/* Number of photons. */
    long  size;		/* Size of malloc'ed buffer. */
    long *buffer;	/* Photon buffer. */
};
static int NextFrame(RawFrame *framePtr, FILE *stream, int flags);

/*---------------------------------------------------------------------------*/

static void FITS_PrintInteger(FILE *stream, const char *keyword,
		long value, const char *comment);
static void FITS_PrintReal(FILE *stream, const char *keyword,
		double value, const char *comment);
static void FITS_PrintString(FILE *stream, const char *keyword,
		const char *value, const char *comment);
static void FITS_PrintLogical(FILE *stream, const char *keyword,
		int value, const char *comment);
static void FITS_PrintCommentString(FILE *stream, const char *keyword,
		const char *comment);
static void FITS_PrintComment(FILE *stream, const char *comment);
static void FITS_PrintHistory(FILE *stream, const char *comment);
static void FITS_PrintEmpty(FILE *stream);
static void FITS_PrintEnd(FILE *stream);


static char *StringCopy(char *dst, const char *src, long len);
/*	Copy string SRC into DST.  If LEN is positive (or zero), it is the
	maximum number of characters that get copied; otherwise, the actual
	length of SRC is used.  DST must have enough space to store the result
 	(at least LEN+1 bytes if LEN>=0).  The resulting DST is always zero
	terminated.  SRC can be NULL (empty string) and DST is returned. */

static char *StringCopy(char *dst, const char *src, long len)
{
    long i;
    
    if (src==NULL) {
	dst[0]= 0;
    } else if (len>=0) {
	for (i=0 ; i<len ; i++) {
	    if ((dst[i]= src[i]) == '\0') {
		return dst;
	    }
	}
	dst[len]= 0;
    } else {
	for (i=0 ; (dst[i]= src[i])!='\0' ; i++) {
	    ;
	}
    }
    return dst;
}

static const char *ClipComment(const char *comment);
static const char *ClipComment(const char *comment)
{
    static char string[48];
    
    return (const char *) StringCopy(string, comment, 47);
}

static void FITS_PrintInteger(FILE *stream, const char *keyword, long value,
	const char *comment)
{
    fprintf(stream, "%-8s= %20ld / %-47s", keyword, value,
	    ClipComment(comment));
    fflush(stream);
}

static void FITS_PrintReal(FILE *stream, const char *keyword, double value,
	const char *comment)
{
    fprintf(stream, "%-8s= %#20.13G / %-47s", keyword, value,
	    ClipComment(comment));
    fflush(stream);
}

static void FITS_PrintString(FILE *stream, const char *keyword,
	const char *value, const char *comment)
{
    char string[19];

    fprintf(stream, "%-8s= '%-18s' / %-47s", keyword,
	    StringCopy(string, value, 18),
	    ClipComment(comment));
    fflush(stream);
}

static void FITS_PrintLogical(FILE *stream, const char *keyword,
	int value, const char *comment)
{
    fprintf(stream, "%-8s= %20s / %-47s", keyword, (value ? "T" : "F"),
	    ClipComment(comment));
    fflush(stream);
}

static void FITS_PrintCommentString(FILE *stream, const char *keyword,
	const char *comment)
{
    char string[72];

    fprintf(stream, "%-8s %-71s", keyword, StringCopy(string, comment, 71));
    fflush(stream);
}

static void FITS_PrintComment(FILE *stream, const char *comment)
{
    FITS_PrintCommentString(stream, "COMMENT", comment);
}

static void FITS_PrintHistory(FILE *stream, const char *comment)
{
    FITS_PrintCommentString(stream, "HISTORY", comment);
}

static void FITS_PrintEnd(FILE *stream)
{
    FITS_PrintCommentString(stream, "END", NULL);
}

static void FITS_PrintEmpty(FILE *stream)
{
    FITS_PrintCommentString(stream, "", NULL);
}

/*---------------------------------------------------------------------------*/

#define WIDTH  384
#define HEIGHT 288
#define BUFFER 1024

#define XDIM1 WIDTH
#define YDIM1 HEIGHT
#define X1(p) (((p)>> 2)&0x1ff)
#define Y1(p) (((p)>>14)&0x1ff)

#define XDIM2 (WIDTH*2)
#define YDIM2 (HEIGHT*2)
#define X2(p) (((p)>> 1)&0x3ff)
#define Y2(p) (((p)>>13)&0x3ff)

#define XDIM4 (WIDTH*4)
#define YDIM4 (HEIGHT*4)
#define X4(p) ( (p)     &0x7ff)
#define Y4(p) (((p)>>12)&0x7ff)

const char *help= "\n\
Name:\n\
  cp40dump -- dump CP40 photons as a cube of 4 long exposure images.\n\
\n\
Usage:\n\
  cp40dump [OPTIONS] SRC DST\n\
\n\
Arguments:\n\
  SRC         Source CP40 data file.\n\
  DST ...     Destination FITS file.\n\
\n\
Options:\n\
  -h, --help  Print this help.\n\
  -o, --old   Old CP40 header (512 bytes) and raw frames.\n\
  -r, --raw   Raw frames (no frame header).\n\n";

static int GetOption(const char *arg, int abbrev, const char *opt);
static int GetOption(const char *arg, int abbrev, const char *opt)
{
    int c;
    
    if (arg[0]!='-') return 0;
    if (arg[1]!='-') {
	while ((c= *++arg)) {
	    if (c==abbrev) return 1;
	}
    } else if (arg[2]==opt[0] && strcmp(arg+3, opt+1)==0) {
	return 2;
    }
    return 0;
}

int main(int argc, const char *argv[])
{
    FILE *file;
    RawFrame frame;
    const char *in= NULL, *out= NULL;
    long *img, *buf, nelem, i, p, x, y;
    int swap= (ByteOrder()==LITTLE_ENDIAN);
    long skip= 16384;
    int raw= 0, opt, flags;

    if (sizeof(long)!=4) {
	fprintf(stderr, "sizeof(long) != 4\n");
	exit(1);
    }
    if (argc==1) {
syntax:
	fputs(help, stderr);
	exit(1);
    }
    for (i=1 ; i<argc ; i++) {
	flags= 1;
	if (argv[i][0]=='-') {
	    if ((opt= GetOption(argv[i], 'o', "old"))) {
		flags= 0;
		skip= 512;
		raw= 1;
		if (opt==2) continue;
	    }
	    if ((opt= GetOption(argv[i], 'r', "raw"))) {
		flags= 0;
		raw= 1;
		if (opt==2) continue;
	    }
	    if (GetOption(argv[i], 'h', "help")) {
		fputs(help, stderr);
		exit(0);
	    }
	}
	if (flags) {
	    if (in==NULL) in= argv[i];
	    else if (out==NULL) out= argv[i];
	    else goto syntax;
	}
    }
    if (out==NULL) {
	goto syntax;
    }
    flags= 0;
    if (swap) flags |= FLAG_SWAP;
    if (raw) flags |= FLAG_RAW;

    if ((file= fopen(in, "r"))==NULL) {
	fprintf(stderr, "can't open file for reading \"%s\"\n", in);
	exit(1);
    }
    if ((img= malloc(XDIM1*YDIM1*4*sizeof(long))) == NULL) {
	fprintf(stderr, "can't alloc memory\n");
	exit(1);
    }
    memset(img, 0, WIDTH*HEIGHT*4*sizeof(long));
    if (fseek(file, skip, SEEK_SET)!=0) {
	fprintf(stderr, "can't skip header\n");
	exit(1);
    }
    memset(&frame, 0, sizeof(frame));
    while (NextFrame(&frame, file, flags)) {
	if ((nelem= frame.flux)<=0) continue;
	buf= frame.buffer;
	for (i=0 ; i<nelem ; i++) {
	    p= buf[i];
	    if ((x= X1(p)) >= XDIM1 || (y= Y1(p)) >= YDIM1) {
		continue;
	    }
	    switch (p & 0x800800) {
	    case 0x000000: img[x+XDIM1*y]+= 1; break;
	    case 0x000800: img[x+XDIM1*y+(XDIM1*YDIM1)]+= 1; break;
	    case 0x800000: img[x+XDIM1*y+(XDIM1*YDIM1*2)]+= 1; break;
	    default:       img[x+XDIM1*y+(XDIM1*YDIM1*3)]+= 1; break;
	    }
	}	
    }
    fclose(file);

    /* Write FITS image. */
    if (swap) SwapByte((char *)img, sizeof(long), (XDIM1*YDIM1*4));
    if ((file= fopen(out, "w"))==NULL) {
	fprintf(stderr, "can't open file for writing \"%s\"\n", out);
	exit(1);
    }
    FITS_PrintLogical(file, "SIMPLE", 1, "true FITS file (no extensions)");
    FITS_PrintInteger(file, "BITPIX",32, "number of bits per pixel");
    FITS_PrintInteger(file, "NAXIS",  3, "number of dimensions");
    FITS_PrintInteger(file, "NAXIS1",XDIM1, "width");
    FITS_PrintInteger(file, "NAXIS2",YDIM1, "height");
    FITS_PrintInteger(file, "NAXIS3", 4, "depth");
    for (i=7 ; i<36 ; i++) FITS_PrintEmpty(file);
    FITS_PrintEnd(file);
    if (ftell(file)%2880 != 0) {
	fprintf(stderr, "error writing FITS header\n");
	exit(1);
    }
    if (fwrite(img, sizeof(long), (XDIM1*YDIM1*4), file) != (XDIM1*YDIM1*4)) {
	fprintf(stderr, "error writing FITS image\n");
	exit(1);
    }
    fclose(file);
    return 0;
}

#define CHUNK 1024
static int NextFrame(RawFrame *framePtr, FILE *stream, int flags)
{
    long header[2], nread;


    clearerr(stream);
	
    if (flags & FLAG_RAW) {
	
	framePtr->flux= CHUNK;
	
    } else { 
    
	/* Read frame header. */
	if ((nread= fread(header, sizeof(long), 2, stream)) != 2) {
	    if (feof(stream)) {
		if (nread!=0) {
		    fputs("warning: premature end of data file\n", stderr);
		}
	    } else {
		fputs("warning: read error\n", stderr);
	    }
	    framePtr->flux= 0;
	    return 0;
	}
	if (flags & FLAG_SWAP) SwapByte((char *)header, sizeof(long), 2);
	framePtr->number= header[0];
	framePtr->flux= header[1];
    }

    /* (Re)allocate photon buffer. */
    if (framePtr->flux > framePtr->size) {
	if (framePtr->size < CHUNK) framePtr->size= CHUNK;
	while (framePtr->size < framePtr->flux) {
#if 1
	    framePtr->size+= CHUNK;
#else
	    framePtr->size*= 2;
#endif
	}
	if (framePtr->buffer) free(framePtr->buffer);
	if ((framePtr->buffer= malloc(framePtr->size*sizeof(long))) == NULL) {
	    fprintf(stderr, "can't alloc memory\n");
	    exit(1);
	}
    }

    /* Read photons. */
    nread= fread(framePtr->buffer, sizeof(long), framePtr->flux, stream);
    if (nread != framePtr->flux) {
	if (ferror(stream)) {
	    fputs("warning: read error\n", stderr);
	}
	if (! (flags & FLAG_RAW) && feof(stream)) {
	    fputs("warning: premature end of data file\n", stderr);
	}
	if (nread<0) nread= 0;
	framePtr->flux= nread;
    }
    if (nread) {
	if (flags & FLAG_SWAP) {
	    SwapByte((char *)framePtr->buffer, sizeof(long), nread);
	}
    } else if (flags & FLAG_RAW) {
	return 0;
    }
    return 1;
}
