/*
 * cp40cmd_simu.c --
 *
 *	Simulate photon server (can be compiled alone).
 *
 * Compilation:
 *	The following macros may be redefined at compile time:
 *	    -DHAVE_LGAMMA=0         if you don't have lgamma() in libm.a;
 *	    -DHAVE_GETHOSTBYNAME=0  if you don't have gethostbyname() in libc.
 *	Only the standard libraries are required: -lm (and -lc).
 *
 * Syntax:
 *	cp40simu portNumber dataFile flux
 *
 * History:
 *	$Log: cp40cmd_simu.c,v $
 *	Revision 2.0  1999/07/02 16:50:27  eric
 *	*** empty log message ***
 *
 *-----------------------------------------------------------------------------
 */

static const char *rcs_version= "$Id: cp40cmd_simu.c,v 2.0 1999/07/02 16:50:27 eric Exp $";

/* Macro to get rid of some GCC extensions when not compiling with GCC. */
#if ! (defined(__GNUC__) && __GNUC__ > 1)
# define __attribute__(x)
# define __inline__
#endif

#include <math.h>
#ifndef M_PI
# define M_PI 3.14159265358979323846
#endif
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>

#ifndef EXIT_SUCCESS
# define EXIT_SUCCESS 0
#endif
#ifndef EXIT_FAILURE
# define EXIT_FAILURE 1
#endif

#ifndef HAVE_LGAMMA
# define HAVE_LGAMMA 1
#endif
#ifndef HAVE_GETHOSTBYNAME
# define HAVE_GETHOSTBYNAME 1
#endif
/*
 * Private data and routines used in this file.
 */

int signalCode= 0; /* Value of signal intercepted. */

static double random_poisson(double xm, unsigned long *seed);
static int    read_poisson_data(const char *fileName, double flux);
static int   *make_photon_couting_image(int *countPtr);

/* ROUND_FLOOR(x,y)  returns largest multiple of Y less or equal X
   ROUND_CEIL(x,y)   returns smallest multiple of Y greater or equal X
   (X and Y are assumed to be positive integers). */
#define ROUND_FLOOR(x,y)  (((x)/(y))*(y))
#define ROUND_CEIL(x,y)    ROUND_FLOOR((x)+((y)-1), y)

/* Define the following macro if last photon in frame is positive
   (by default, first photon in frame is positive). */
/* #undef LAST_PHOTON_IS_POSITIVE */
#define PHOTON_SIZE    4 /* Size of one photon-event. */
#define BLOCK       1024 /* Alloc memory by a multiple of this value. */

/* Timeout settings. */
#define MICROSECONDS 1
#define MILLISECONDS (1000*MICROSECONDS)
#define SECONDS      (1000*MILLISECONDS)
#define MINUTES      (60*SECONDS)
static void	set_timeout(struct timeval *tvPtr, unsigned long value,
			unsigned long unit);
static int	timeout_expired(struct timeval *tvPtr);
static void	wait_timeout(struct timeval *tvPtr);
static void	delay(unsigned long sec, unsigned long usec);

static void	system_error(const char *message);


/* These bits may be OR'ed together into the "flags" argument of the
   create_socket() function.  */
#define SOCKET_ASYNC  (1<<0) /* Asynchronous client-socket connection. */
#define SOCKET_SERVER (1<<1) /* Create a server socket. */
#define SOCKET_CLIENT (1<<2) /* Create a client socket. */

/* Private definitions for sockets. */
#define SOCKET_BACKLOG   10	/* Maximum allowable length of the queue for
				   pending connections, see listen(2).  If
				   this is larger than SOMAXCONN, the actual
				   value used is silently truncated to
				   SOMAXCONN (5 for OS9). */
#define SOCKET_BUFSIZE 4096	/* Minimum size of socket internal buffers
				   (only used if SO_SNDBUF and SO_RCVBUF are
				   defined). */

static int	create_socket(int port, char *host, char *myaddr,
			int myport, int flags);
static int	init_socket_address(struct sockaddr_in *sockaddrPtr,
			char *host, int port);
static int	write_socket(int sock, const void *buffer, int nbytes);

/*_____________________________________________________________________________
 *
 * main --
 *
 *	Main routine.
 *
 * Return:
 *	Non-zero in case of error; otherwise (and hopefully!) zero.
 *
 * Side effects:
 *	See descrition.
 *_____________________________________________________________________________
 */

int main(int argc, char *argv[])
{
  struct timeval tv;
  int frame_date= 0;    /* Frame counter. */
  int listen_sock;      /* Server socket (used to listen for connection
                           requests). */
  int client_sock= -1;  /* Client socket (used to send photons to connected
                           client; -1 means not yet open). */
  int port;             /* Server port number. */
  socklen_t addrlen;
  struct sockaddr peeraddr_in;    
  int  count;         /* Number of events in frame. */
  char *buffer= NULL; /* Pending data buffer (the use of this buffer
                         is required to avoid, as far as possible,
                         frame truncation and always avoid photon
                         truncation (unless reconnection)). */
  int pending;        /* Number of pending bytes (i.e. those that have not
                         yet been written to socket). */
  int nwrite;         /* Number of bytes effectively written to socket. */
  double flux;
  char c;             /* Dummy character used to check that arguments are
                         correct. */

  /* Check syntax and parse arguments. */
  if (argc != 4) {
    fprintf(stderr, "\nName: %s\n", argv[0]);
    fprintf(stderr, "\nPurpose: simulate a photon server.\n");
  usage:
    fprintf(stderr, "\nSyntax: %s PORT FILE FLUX\n", argv[0]);
    fprintf(stderr, "     PORT = server port number\n");
    fprintf(stderr, "     FILE = data file image (384x288 bytes)\n");
    fprintf(stderr, "     FLUX = average number of photons per frame\n");
    fprintf(stderr, "\nVersion: %s\n\n", rcs_version);
    exit(EXIT_FAILURE);
  }
  if (sscanf(argv[1], "%d %c", &port, &c) != 1 || (port & 0xFFFF) != port) {
    fprintf(stderr, "%s: bad port number\n", argv[0]);
    goto usage;
  }
  if (sscanf(argv[3], "%lf %c", &flux, &c) != 1) {
    flux= -1.0;
  }
  if (! read_poisson_data(argv[2], flux)) {
    exit(EXIT_FAILURE);
  }

  /* Avoid abort if connection gets closed. */
  signal(SIGPIPE, SIG_IGN);
    
  /* Open listening (server) socket, wait until a client asks for
     connection then accept the connection.  Note: the call to accept()
     will block until a new connection arrives; an alternative solution
     would be to use select() to make sure that a client is actually
     asking for connection before calling accept(). */
  listen_sock= create_socket(port, NULL, NULL, 0, SOCKET_SERVER);
  if (listen_sock == -1) {
    system_error("cannot open server socket");
    exit(0);
  }
 reconnect:
  if (client_sock != -1) {
    close(client_sock);
    client_sock= -1;
  }
  do {
    fprintf(stderr, "waiting for connection\n");
    addrlen= sizeof(peeraddr_in);
    if ((client_sock= accept(listen_sock, &peeraddr_in, &addrlen)) == -1) {
      system_error("accept() fails");
    }
  } while (client_sock == -1);
  if (fcntl(client_sock, F_SETFL,
	    fcntl(client_sock, F_GETFL) | O_NONBLOCK) == -1) {
    system_error("cannot set non-blocking mode");
  }
  fprintf(stderr, "connection established\n");

    
  pending= 0;	/* Pending photon buffer is empty. */
  for (;;) {
	
    /* Make new photon-counting frame. */
    set_timeout(&tv, 20, MILLISECONDS);    
    if ((buffer= (char *)make_photon_couting_image(&count)) == NULL)
      exit(EXIT_FAILURE);
    frame_date++;
    pending= PHOTON_SIZE * count;
	    
    /* Try to flush pending photon buffer if any data remain. */
    while (pending > 0) {
      if ((nwrite= write_socket(client_sock, buffer, pending)) < 0)
	goto reconnect;
      if (nwrite >= pending) {
	/* Complete flush. */
	pending= 0;
      } else if (nwrite > 0) {
	/* Incomplete flush: move pending buffer
	   to beginning of remaining data. */
	pending -= nwrite;
	buffer += nwrite;
      }
    }
	
    /* Wait expiration of timeout before sending new frame. */
    wait_timeout(&tv);
  }

  /* Close connection with DMA device and socket connections if any. */
  if (client_sock != -1) close(client_sock);
  if (listen_sock != -1) close(listen_sock);
  return EXIT_SUCCESS;
}


/*
 * poisson.c --
 *
 *	Generate a photon-counting image.
 *
 */

#ifndef REAL
#   define REAL double
#endif

/* Macros to deal with raw photons. */
#define WIDTH       384			/* CCD width. */
#define HEIGHT      288			/* CCD height. */
#define RAW_CHN(p)  ((p)&0x00800800)	/* Mask for channel bits. */
#define GET_X(p)    (((p)>>2)&511)	/* Raw photon abscissa. */
#define GET_Y(p)    (((p)>>14)&511)	/* Raw photon ordinate. */

#define NUMBER (HEIGHT*WIDTH)
static double poisson_flux= 0.0;	/* Mean total flux. */
static REAL *poisson_data= NULL;



/* Private data for pseudo-random generator. */
#define RAND_M           1771875
#define RAND_A              2416
#define RAND_C            374441
#define RAND_INIT        (time(0)%RAND_M)
#define Rand(i)          ((i)=(RAND_A*(i)+RAND_C)%RAND_M)
#define Random(i)        ((REAL)Rand(i)/(REAL)RAND_M)
#define RandomN(i,n)     ((Rand(i)*(n))/RAND_M)

#if 0
int ramdom_bit(unsigned long *seed) {
    unsigned long s= *seed;
    int bit= ((s&131072)>>17)^((s&16)>>4)^((s&2)>>1)^(s&1);
    *seed= (s<<1) | bit;
    return bit;
}
#endif

static int read_poisson_data(const char *fileName, double flux)
{
  int fd;
  register unsigned char *image= NULL;
  register int i;
  register double sum, a;
  register REAL *data= NULL;

  image= (unsigned char *)malloc(NUMBER*sizeof(unsigned char));
  if (poisson_data==NULL) {
    poisson_data= (REAL*)malloc(NUMBER*sizeof(REAL));
  }
  if (poisson_data==NULL || image==NULL) {
    fprintf(stderr, "alloc error\n");
    goto error;
  }
  if (flux <= 0.0) {
    fprintf(stderr, "bad flux: %g\n", flux);
    goto error;
  }
  poisson_flux= flux;
  if ((fd= open(fileName, O_RDONLY)) < -1 
      || read(fd, image, NUMBER) != NUMBER) {
    if (fd >= 0) close(fd);
    fprintf(stderr, "cannot read image in file: %s\n", fileName);
    goto error;
  }
  close(fd);

  for (sum= 0.0, i=0; i<NUMBER; i++) {
    sum += image[i];
  }
  if (sum<=0.0) {
    fprintf(stderr, "blank image in file: %s\n", fileName);
    goto error;
  }
  a= 1.0 / sum;
  data= poisson_data;
  for (sum= 0.0, i=0; i<NUMBER; i++) {
    sum += image[i];
    data[i]= a*sum;
  }
  free(image);
  return 1;
 error:
  if (image) free(image);
  return 0;
}

/* Make a new photon-counting frame.  Return pointer to internally malloc'ed
   new frame. */
int *make_photon_couting_image(int *countPtr)
{
  static unsigned long seed1= 0;
  static unsigned long seed2= 0;
  static unsigned long seed3= 0;
  static int littleEndian =-1;
  int channel, count, i;
  static int *frame= NULL;
  static int size= 0;
  register REAL  random_value;
  register REAL *data= poisson_data;
  register int j, jmax, jmin;

  if (littleEndian == -1) {
    littleEndian= 1;
    littleEndian= *(char *)&littleEndian ? 1 : 0;
    fprintf(stderr, "This machine has %s endian byte order\n",
	    littleEndian ? "little" : "big");
    seed1=   time(0);
    seed2= 2*time(0);
    seed3= 3*time(0);
  }
    
  count= random_poisson(poisson_flux, &seed3);
  if (size < count) {
    for (size=2; size<count; size*=2)
      ;
    if (frame != NULL) free(frame);
    if ((frame= (int *)malloc(size * sizeof(int))) == NULL) {
      fprintf(stderr, "cannot allocate %d bytes\n", size);
      *countPtr= 0;
      return NULL;
    }
  }

  /* Use a simple bissection method to localize photons. */
  for (i=0; i<count; i++) {
    random_value= Random(seed1);
    jmin= 0;
    jmax= NUMBER-1;
    while (jmin < jmax) {
      j= (jmin+jmax)>>1;
      if (random_value < data[j]) {
	jmax= j;
      } else {
	jmin= j+1;
      }
    }
#if 1
    channel= (Rand(seed2)*4)/RAND_M;/* Highest bits give channel number. */
#else
    channel= Rand(seed2) % 4;	/* Lowest bits give channel number. */
#endif
    frame[i]= 0xff000000
      | ((channel<<22)&0x800000) | ((channel<<11)&0x800)
      | (((jmin/384)<<14)&0x7fc000) | (((jmin%384)<<2)&0x7fc);
  }
  *countPtr= count;

    /* Mark end of frame. */
#ifdef LAST_PHOTON_IS_POSITIVE
  frame[count-1] &= 0x7fffffff;
#else
  frame[0] &= 0x7fffffff;
#endif 
  if (littleEndian) {
    for (i=0; i<count; i++) {
      /* Swap bytes. */
      char c, *dst= (char *)frame;
      int n= count*sizeof(int);
      for (i=0; i<n; i+=4) {
	c= dst[i];   dst[i]= dst[i+3];   dst[i+3]= c;
	c= dst[i+1]; dst[i+1]= dst[i+2]; dst[i+2]= c;
      }
    }
  }
  return frame;
}

/*
 *-----------------------------------------------------------------------------
 *
 * random_poisson --
 *
 *	Random generator following Poisson's law.
 *
 * Return:
 *	A random number.
 *
 * Side effects:
 *	None.
 *
 *-----------------------------------------------------------------------------
 */
#if defined (HAVE_LGAMMA) && HAVE_LGAMMA
extern double lgamma(double xx);
#else
/*_____________________________________________________________________________
 *
 * lgamma --
 *
 *	The lgamma function returns the natural logarithm of the gamma
 *	function defined for a complex z by:
 *	  gamma(z) = \int_{0}^{+\infty} t^{z-1} e^{-t} dt
 *	For integer n, the gamma function is such that:
 *	  gamma(n+1) = n!
 *	The gamma function also satisfies the recurrence and reflection
 *	relations:
 *	  gamma(1+z) = z gamma(z)
 *	                      pi                    pi z
 *	  gamma(1-z) = ------------------ = --------------------
 *	               gamma(z) sin(pi z)   gamma(1+z) sin(pi z)
 *
 *	The lgamma function is defined for real x > 0, full accuracy is
 *	obtained for x > 1.  For x <= 1, the reflection formula can be used.
 *_____________________________________________________________________________
 */

static double lgamma(double xx);

static __inline__ double lgamma(register double x)
{
  register double xp= x+5.5;
  return -xp + (x+0.5)*log(xp)
    + log((2.5066282751072975e+00
	   + 1.9095517189307640e+02 / (x + 1.0)
	   - 2.1683668184372800e+02 / (x + 2.0)
	   + 6.0194417640233333e+01 / (x + 3.0)
	   - 3.0875132392854585e+00 / (x + 4.0)
	   + 3.0296387052532589e-03 / (x + 5.0) 
	   - 1.3523859590725960e-05 / (x + 6.0)) / x);
}
#endif /* HAVE_LGAMMA */

static double random_poisson(double xm, unsigned long *seed)
{
  static double sq, alxm, g, oldm= -1.0;
  register double em,t,y;
    
  if (xm < 12.0) {
    if (xm <= 0)
      return 0.0;
    if (xm != oldm) {
      oldm= xm;
      g= exp(-xm);
    }
    em= -1;
    t=1.0;
    do {
      em += 1.0;
      t *= Random(seed[0]);
    } while (t > g);
  } else {
    if (xm != oldm) {
      oldm= xm;
      sq= sqrt(2.0*xm);
      alxm= log(xm);
      g= xm*alxm-lgamma(xm+1.0);
    }
    do {
      do {
	y= tan(M_PI*Random(seed[0]));
	em= sq*y+xm;
      } while (em < 0.0);
      em= floor(em);
#ifdef _OSK
#define MAX_LOG  709.7 /* 709.78271289338399672769 */
#define MIN_LOG -708.3 /* -708.39641853226410612691 */
      t= em*alxm-lgamma(em+1.0)-g;
      if (t <= MIN_LOG) continue;
      if (t >= MAX_LOG) break;
      t= 0.9*(1.0+y*y)*exp(t);
#else
      t= 0.9*(1.0+y*y)*exp(em*alxm-lgamma(em+1.0)-g);
#endif
    } while (Random(seed[0]) > t);
  }
  return em;
}

static void set_timeout(struct timeval *tvPtr, unsigned long value,
			unsigned long unit)
{
  struct timeval now;
  unsigned long usec;

  gettimeofday(&now, NULL);
  switch (unit) {
  case MICROSECONDS:
    usec= now.tv_usec + value;
    tvPtr->tv_usec= usec % 1000000;
    tvPtr->tv_sec= now.tv_sec + usec / 1000000;
    return;
  case MILLISECONDS:
    usec= now.tv_usec + 1000 * (value % 1000);
    tvPtr->tv_usec= usec % 1000000;
    tvPtr->tv_sec= now.tv_sec + value / 1000 + usec / 1000000;
    break;
  case SECONDS:
    tvPtr->tv_sec= now.tv_sec + value;
    tvPtr->tv_usec= now.tv_usec;
    return;
  case MINUTES:
    tvPtr->tv_sec= now.tv_sec + 60 * value;
    tvPtr->tv_usec= now.tv_usec;
    return;
  }
}

static int timeout_expired(struct timeval *tvPtr)
{
  struct timeval now;
  
  gettimeofday(&now, NULL);
  return tvPtr->tv_sec < now.tv_sec
    || (tvPtr->tv_sec == now.tv_sec && tvPtr->tv_usec <= now.tv_usec);
}

static void wait_timeout(struct timeval *tvPtr)
{
  struct timeval tv;

  gettimeofday(&tv, NULL);
  if (tvPtr->tv_usec >= tv.tv_usec) {
    if (tvPtr->tv_sec < tv.tv_sec) return;
    tv.tv_sec= tvPtr->tv_sec - tv.tv_sec;
    tv.tv_usec= tvPtr->tv_usec - tv.tv_usec;
  } else {
    if (tvPtr->tv_sec <= tv.tv_sec) return;
    tv.tv_sec= tvPtr->tv_sec - tv.tv_sec - 1;
    tv.tv_usec= tvPtr->tv_usec + 1000000 - tv.tv_usec;
  }
  select(0,  NULL, NULL, NULL, &tv);
}

static void delay(unsigned long sec, unsigned long usec)
{
  struct timeval tv;
  tv.tv_sec= sec+usec/1000000;
  tv.tv_usec= usec%1000000;
  select(0,  NULL, NULL, NULL, &tv);
}    

/*_____________________________________________________________________________
 *
 * create_socket --
 *
 *	This function opens a new socket in client or server mode
 *	and initializes it.
 *
 * Results:
 *	Returns -1 if an error occurs; otherwise, returns a valid path
 *	descriptor referencing the socket.
 *
 * Side effects:
 *	Opens a socket and set its address and options.
 *_____________________________________________________________________________
 */

static int create_socket(port, host, myaddr, myport, flags)
     int   port;    /* Port number to open. */
     char *host;    /* Name of host on which to open port.  NULL
		       implies INADDR_ANY. */
     char *myaddr;  /* Optional client-side address. */
     int   myport;  /* Optional client-side port. */
     int   flags;   /* OR'ed bit flags.  If SOCKET_SERVER is set,
		       create a server socket; otherwise, create a client
		       socket.  For a client socket, if SOCKET_ASYNC
		       is set, attempt to do an asynchronous connect;
		       otherwise, do a synchronous connect or bind. */
{
  int value;                            /* Used to set/get socket options. */
  socklen_t length;                     /* Used to set/get socket options. */
  struct sockaddr_in  sockaddr;         /* Socket address. */
  struct sockaddr_in  mysockaddr;       /* Socket address for client. */
  struct sockaddr_in *bindsockaddr_ptr; /* Socket address for binding. */
  int async= flags & SOCKET_ASYNC;      /* Asynchronous connect? */
  int server= flags & SOCKET_SERVER;    /* Server? */
  int sock= -1;                         /* Socket file descriptor number. */
    
  /* Initializes socket address(es) and figure out sockect address to bind. */
  if (! init_socket_address(&sockaddr, host, port)) {
    goto error;
  }
  if (server) {
    bindsockaddr_ptr= &sockaddr;
  } else if (myaddr != NULL || myport != 0) {
    if (! init_socket_address(&mysockaddr, myaddr, myport)) {
      goto error;
    }
    bindsockaddr_ptr= &mysockaddr;
  } else {
    bindsockaddr_ptr= NULL;
  }
  
  
  /* Create socket and set socket options (minimum size of socket
     buffers). */
  
  if ((sock= socket(AF_INET, SOCK_STREAM, 0)) == -1) {
    goto error;
  }
#ifdef SO_SNDBUF
  length= sizeof(int);
  getsockopt(sock, SOL_SOCKET, SO_SNDBUF, &value, &length);
  if (value < SOCKET_BUFSIZE) {
    value= SOCKET_BUFSIZE;
    setsockopt(sock, SOL_SOCKET, SO_SNDBUF, &value, sizeof(int));
  }
#endif
#ifdef SO_RCVBUF
  length= sizeof(int);
  getsockopt(sock, SOL_SOCKET, SO_RCVBUF, &value, &length);
  if (value < SOCKET_BUFSIZE) {
    value= SOCKET_BUFSIZE;
    setsockopt(sock, SOL_SOCKET, SO_RCVBUF, &value, sizeof(int));
  }
#endif
  
  
  /* Eventually, bind to the specified port and set up to reuse addresses
     automatically. */
  
  if (bindsockaddr_ptr != NULL) {
#ifdef SO_REUSEADDR
    value= 1;
    setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &value, sizeof(int));
#endif
    if (bind(sock, (struct sockaddr *)bindsockaddr_ptr,
	     sizeof(struct sockaddr))) {
      goto error;
    }
  }
  
  if (server) {

    /* Server socket: start listening for connections. */
    
    if (listen(sock, SOCKET_BACKLOG) == -1) {
      goto error;
    }

  } else {
	
    /* Client socket: attempt to connect. The connect may fail at present
	   with an EINPROGRESS but at a later time it will complete.  The
	   caller will set up a file handler on the socket if she is
	   interested in being informed when the connect completes. */

    if (async && fcntl(sock, F_SETFL, fcntl(sock, F_GETFL) | O_NONBLOCK)
	== -1) {
      goto error;
    }
    if (connect(sock, (struct sockaddr *)&sockaddr, sizeof(sockaddr))) {
      if (async && errno == EINPROGRESS) {
	errno= 0;
      } else {
	goto error;
      }
    }
  }
    
  return sock;

 error:
  if (sock != -1) {
    close(sock);
  }
  return -1;	/* Error. */
}

/*_____________________________________________________________________________
 *
 * init_socket_address --
 *
 *	This function initializes a sockaddr structure for a host and port.
 *
 * Results:
 *	1 if the host was valid, 0 if the host could not be converted to
 *	an IP address.
 *
 * Side effects:
 *	Fills in the *sockaddrPtr structure.
 *_____________________________________________________________________________
 */

static int init_socket_address(sockaddrPtr, host, port)
     struct sockaddr_in *sockaddrPtr; /* Socket address */
     char *host;                      /* Host.  NULL implies INADDR_ANY */
     int port;                        /* Port number */
{
  struct in_addr addr; /* For 64/32 bit madness */

  memset(sockaddrPtr, '\0', sizeof(struct sockaddr_in));
  sockaddrPtr->sin_family= AF_INET;
  sockaddrPtr->sin_port= htons((unsigned short)(port & 0xFFFF));
  if (host == NULL) {
    addr.s_addr= INADDR_ANY;
  } else {
    addr.s_addr= inet_addr(host);
    if (addr.s_addr == -1) {
#if HAVE_GETHOSTBYNAME /* Use gethostbyname() to get host inet address. */
      struct hostent *hostent;		/* Host database entry */
	    
      hostent= gethostbyname(host);
      if (hostent != NULL) {
	memcpy(&addr, hostent->h_addr_list[0], hostent->h_length);
      } else {
#   ifdef EHOSTUNREACH
	errno= EHOSTUNREACH;
#   else
#	ifdef ENXIO
	errno= ENXIO;
#	endif
#   endif
	return 0;	/* error */
      }
#else /* Cannot use gethostbyname(). */
      return 0;	/* error */		
#endif
    }
  }
        
  /* NOTE: On 64 bit machines the assignment below is rumored to not do the
     right thing. Please report errors related to this if you observe
     incorrect behavior on 64 bit machines such as DEC Alphas.  Should we
     modify this code to do an explicit memcpy? */
  sockaddrPtr->sin_addr.s_addr= addr.s_addr;
  return 1;	/* Success. */
}

/* Write data to socket, return number of bytes actually written or
   -1 if the connection was closed. */
static int write_socket(int sock, const void *buffer, int nbytes)
{
  int nwrite;
    
  if ((nwrite= write(sock, buffer, nbytes)) >= 0)
    return nwrite;

  /* Write error. */
#ifdef EWOULDBLOCK
  if (errno == EWOULDBLOCK)
    return 0;
#endif
#ifdef ECONNRESET
  if (errno == ECONNRESET)
    return -1;
#endif
  system_error("write() error");
  return -1;
}

/*_____________________________________________________________________________
 *
 * system_error --
 *
 *	Print out error message corresponding to last system error (i.e. to
 *	current `errno` value).
 *_____________________________________________________________________________
 */

void system_error(const char *message)
{
  const char *err;
  switch (errno) {
#ifdef E2BIG
  case E2BIG:
    err= "argument list too long [E2BIG]";
    break;
#endif
#ifdef EACCES
  case EACCES:
    err= "permission denied [EACCES]";
    break;
#endif
#ifdef EADDRINUSE
  case EADDRINUSE:
    err= "address already in use [EADDRINUSE]";
    break;
#endif
#ifdef EADDRNOTAVAIL
  case EADDRNOTAVAIL:
    err= "can't assign requested address [EADDRNOTAVAIL]";
    break;
#endif
#ifdef EADV
  case EADV:
    err= "advertise error [EADV]";
    break;
#endif
#ifdef EAFNOSUPPORT
  case EAFNOSUPPORT:
    err= "address family not supported by protocol family [EAFNOSUPPORT]";
    break;
#endif
#ifdef EAGAIN
  case EAGAIN:
    err= "resource temporarily unavailable [EAGAIN]";
    break;
#endif
#ifdef EALIGN
  case EALIGN:
    err= "EALIGN [EALIGN]";
    break;
#endif
#if defined(EALREADY) && (!defined(EBUSY) || (EALREADY != EBUSY ))
  case EALREADY:
    err= "operation already in progress [EALREADY]";
    break;
#endif
#ifdef EBADE
  case EBADE:
    err= "bad exchange descriptor [EBADE]";
    break;
#endif
#ifdef EBADF
  case EBADF:
    err= "bad file number [EBADF]";
    break;
#endif
#ifdef EBADFD
  case EBADFD:
    err= "file descriptor in bad state [EBADFD]";
    break;
#endif
#ifdef EBADMSG
  case EBADMSG:
    err= "not a data message [EBADMSG]";
    break;
#endif
#ifdef EBADR
  case EBADR:
    err= "bad request descriptor [EBADR]";
    break;
#endif
#ifdef EBADRPC
  case EBADRPC:
    err= "RPC structure is bad [EBADRPC]";
    break;
#endif
#ifdef EBADRQC
  case EBADRQC:
    err= "bad request code [EBADRQC]";
    break;
#endif
#ifdef EBADSLT
  case EBADSLT:
    err= "invalid slot [EBADSLT]";
    break;
#endif
#ifdef EBFONT
  case EBFONT:
    err= "bad font file format [EBFONT]";
    break;
#endif
#ifdef EBUSY
  case EBUSY:
    err= "mount device busy [EBUSY]";
    break;
#endif
#ifdef ECHILD
  case ECHILD:
    err= "no children [ECHILD]";
    break;
#endif
#ifdef ECHRNG
  case ECHRNG:
    err= "channel number out of range [ECHRNG]";
    break;
#endif
#ifdef ECOMM
  case ECOMM:
    err= "communication error on send [ECOMM]";
    break;
#endif
#ifdef ECONNABORTED
  case ECONNABORTED:
    err= "software caused connection abort [ECONNABORTED]";
    break;
#endif
#ifdef ECONNREFUSED
  case ECONNREFUSED:
    err= "connection refused [ECONNREFUSED]";
    break;
#endif
#ifdef ECONNRESET
  case ECONNRESET:
    err= "connection reset by peer [ECONNRESET]";
    break;
#endif
#if defined(EDEADLK) && (!defined(EWOULDBLOCK) || (EDEADLK != EWOULDBLOCK))
  case EDEADLK:
    err= "resource deadlock avoided [EDEADLK]";
    break;
#endif
#if defined(EDEADLOCK) && (!defined(EDEADLK) || (EDEADLOCK != EDEADLK))
  case EDEADLOCK:
    err= "resource deadlock avoided [EDEADLOCK]";
    break;
#endif
#ifdef EDESTADDRREQ
  case EDESTADDRREQ:
    err= "destination address required [EDESTADDRREQ]";
    break;
#endif
#ifdef EDIRTY
  case EDIRTY:
    err= "mounting a dirty fs w/o force [EDIRTY]";
    break;
#endif
#ifdef EDOM
  case EDOM:
    err= "math argument out of range [EDOM]";
    break;
#endif
#ifdef EDOTDOT
  case EDOTDOT:
    err= "cross mount point [EDOTDOT]";
    break;
#endif
#ifdef EDQUOT
  case EDQUOT:
    err= "disk quota exceeded [EDQUOT]";
    break;
#endif
#ifdef EDUPPKG
  case EDUPPKG:
    err= "duplicate package name [EDUPPKG]";
    break;
#endif
#ifdef EEXIST
  case EEXIST:
    err= "file already exists [EEXIST]";
    break;
#endif
#ifdef EFAULT
  case EFAULT:
    err= "bad address in system call argument [EFAULT]";
    break;
#endif
#ifdef EFBIG
  case EFBIG:
    err= "file too large [EFBIG]";
    break;
#endif
#ifdef EHOSTDOWN
  case EHOSTDOWN:
    err= "host is down [EHOSTDOWN]";
    break;
#endif
#ifdef EHOSTUNREACH
  case EHOSTUNREACH:
    err= "host is unreachable [EHOSTUNREACH]";
    break;
#endif
#if defined(EIDRM) && (!defined(EINPROGRESS) || (EIDRM != EINPROGRESS))
  case EIDRM:
    err= "identifier removed [EIDRM]";
    break;
#endif
#ifdef EINIT
  case EINIT:
    err= "initialization error [EINIT]";
    break;
#endif
#ifdef EINPROGRESS
  case EINPROGRESS:
    err= "operation now in progress [EINPROGRESS]";
    break;
#endif
#ifdef EINTR
  case EINTR:
    err= "interrupted system call [EINTR]";
    break;
#endif
#ifdef EINVAL
  case EINVAL:
    err= "invalid argument [EINVAL]";
    break;
#endif
#ifdef EIO
  case EIO:
    err= "I/O error [EIO]";
    break;
#endif
#ifdef EISCONN
  case EISCONN:
    err= "socket is already connected [EISCONN]";
    break;
#endif
#ifdef EISDIR
  case EISDIR:
    err= "illegal operation on a directory [EISDIR]";
    break;
#endif
#ifdef EISNAME
  case EISNAM:
    err= "is a name file [EISNAM]";
    break;
#endif
#ifdef ELBIN
  case ELBIN:
    err= "ELBIN [ELBIN]";
    break;
#endif
#ifdef EL2HLT
  case EL2HLT:
    err= "level 2 halted [EL2HLT]";
    break;
#endif
#ifdef EL2NSYNC
  case EL2NSYNC:
    err= "level 2 not synchronized [EL2NSYNC]";
    break;
#endif
#ifdef EL3HLT
  case EL3HLT:
    err= "level 3 halted [EL3HLT]";
    break;
#endif
#ifdef EL3RST
  case EL3RST:
    err= "level 3 reset [EL3RST]";
    break;
#endif
#ifdef ELIBACC
  case ELIBACC:
    err= "can not access a needed shared library [ELIBACC]";
    break;
#endif
#ifdef ELIBBAD
  case ELIBBAD:
    err= "accessing a corrupted shared library [ELIBBAD]";
    break;
#endif
#ifdef ELIBEXEC
  case ELIBEXEC:
    err= "can not exec a shared library directly [ELIBEXEC]";
    break;
#endif
#ifdef ELIBMAX
  case ELIBMAX:
    err= 	    "attempting to link in more shared libraries than system limit [ELIBMAX]";
    break;
#endif
#ifdef ELIBSCN
  case ELIBSCN:
    err= ".lib section in a.out corrupted [ELIBSCN]";
    break;
#endif
#ifdef ELNRNG
  case ELNRNG:
    err= "link number out of range [ELNRNG]";
    break;
#endif
#if defined(ELOOP) && (!defined(ENOENT) || (ELOOP != ENOENT))
  case ELOOP:
    err= "too many levels of symbolic links [ELOOP]";
    break;
#endif
#ifdef EMFILE
  case EMFILE:
    err= "too many open files [EMFILE]";
    break;
#endif
#ifdef EMLINK
  case EMLINK:
    err= "too many links [EMLINK]";
    break;
#endif
#ifdef EMSGSIZE
  case EMSGSIZE:
    err= "message too long [EMSGSIZE]";
    break;
#endif
#ifdef EMULTIHOP
  case EMULTIHOP:
    err= "multihop attempted [EMULTIHOP]";
    break;
#endif
#ifdef ENAMETOOLONG
  case ENAMETOOLONG:
    err= "file name too long [ENAMETOOLONG]";
    break;
#endif
#ifdef ENAVAIL
  case ENAVAIL:
    err= "not available [ENAVAIL]";
    break;
#endif
#ifdef ENET
  case ENET:
    err= "ENET [ENET]";
    break;
#endif
#ifdef ENETDOWN
  case ENETDOWN:
    err= "network is down [ENETDOWN]";
    break;
#endif
#ifdef ENETRESET
  case ENETRESET:
    err= "network dropped connection on reset [ENETRESET]";
    break;
#endif
#ifdef ENETUNREACH
  case ENETUNREACH:
    err= "network is unreachable [ENETUNREACH]";
    break;
#endif
#ifdef ENFILE
  case ENFILE:
    err= "file table overflow [ENFILE]";
    break;
#endif
#ifdef ENOANO
  case ENOANO:
    err= "anode table overflow [ENOANO]";
    break;
#endif
#if defined(ENOBUFS) && (!defined(ENOSR) || (ENOBUFS != ENOSR))
  case ENOBUFS:
    err= "no buffer space available [ENOBUFS]";
    break;
#endif
#ifdef ENOCSI
  case ENOCSI:
    err= "no CSI structure available [ENOCSI]";
    break;
#endif
#if defined(ENODATA) && (!defined(ECONNREFUSED) || (ENODATA != ECONNREFUSED))
  case ENODATA:
    err= "no data available [ENODATA]";
    break;
#endif
#ifdef ENODEV
  case ENODEV:
    err= "no such device [ENODEV]";
    break;
#endif
#ifdef ENOENT
  case ENOENT:
    err= "no such file or directory [ENOENT]";
    break;
#endif
#ifdef ENOEXEC
  case ENOEXEC:
    err= "exec format error [ENOEXEC]";
    break;
#endif
#ifdef ENOLCK
  case ENOLCK:
    err= "no locks available [ENOLCK]";
    break;
#endif
#ifdef ENOLINK
  case ENOLINK:
    err= "link has be severed [ENOLINK]";
    break;
#endif
#ifdef ENOMEM
  case ENOMEM:
    err= "not enough memory [ENOMEM]";
    break;
#endif
#ifdef ENOMSG
  case ENOMSG:
    err= "no message of desired type [ENOMSG]";
    break;
#endif
#ifdef ENONET
  case ENONET:
    err= "machine is not on the network [ENONET]";
    break;
#endif
#ifdef ENOPKG
  case ENOPKG:
    err= "package not installed [ENOPKG]";
    break;
#endif
#ifdef ENOPROTOOPT
  case ENOPROTOOPT:
    err= "bad proocol option [ENOPROTOOPT]";
    break;
#endif
#ifdef ENOSPC
  case ENOSPC:
    err= "no space left on device [ENOSPC]";
    break;
#endif
#if defined(ENOSR) && (!defined(ENAMETOOLONG) || (ENAMETOOLONG != ENOSR))
  case ENOSR:
    err= "out of stream resources [ENOSR]";
    break;
#endif
#if defined(ENOSTR) && (!defined(ENOTTY) || (ENOTTY != ENOSTR))
  case ENOSTR:
    err= "not a stream device [ENOSTR]";
    break;
#endif
#ifdef ENOSYM
  case ENOSYM:
    err= "unresolved symbol name [ENOSYM]";
    break;
#endif
#ifdef ENOSYS
  case ENOSYS:
    err= "function not implemented [ENOSYS]";
    break;
#endif
#ifdef ENOTBLK
  case ENOTBLK:
    err= "block device required [ENOTBLK]";
    break;
#endif
#ifdef ENOTCONN
  case ENOTCONN:
    err= "socket is not connected [ENOTCONN]";
    break;
#endif
#ifdef ENOTDIR
  case ENOTDIR:
    err= "not a directory [ENOTDIR]";
    break;
#endif
#if defined(ENOTEMPTY) && (!defined(EEXIST) || (ENOTEMPTY != EEXIST))
  case ENOTEMPTY:
    err= "directory not empty [ENOTEMPTY]";
    break;
#endif
#ifdef ENOTNAM
  case ENOTNAM:
    err= "not a name file [ENOTNAM]";
    break;
#endif
#ifdef ENOTSOCK
  case ENOTSOCK:
    err= "socket operation on non-socket [ENOTSOCK]";
    break;
#endif
#ifdef ENOTSUP
  case ENOTSUP:
    err= "operation not supported [ENOTSUP]";
    break;
#endif
#ifdef ENOTTY
  case ENOTTY:
    err= "inappropriate device for ioctl [ENOTTY]";
    break;
#endif
#ifdef ENOTUNIQ
  case ENOTUNIQ:
    err= "name not unique on network [ENOTUNIQ]";
    break;
#endif
#ifdef ENXIO
  case ENXIO:
    err= "no such device or address [ENXIO]";
    break;
#endif
#if defined(EOPNOTSUPP) && (!defined(ENOTSUP) || (ENOTSUP != EOPNOTSUPP))
  case EOPNOTSUPP:
    err= "operation not supported on socket [EOPNOTSUPP]";
    break;
#endif
#ifdef EPERM
  case EPERM:
    err= "not owner [EPERM]";
    break;
#endif
#if defined(EPFNOSUPPORT) && (!defined(ENOLCK) || (ENOLCK != EPFNOSUPPORT))
  case EPFNOSUPPORT:
    err= "protocol family not supported [EPFNOSUPPORT]";
    break;
#endif
#ifdef EPIPE
  case EPIPE:
    err= "broken pipe [EPIPE]";
    break;
#endif
#ifdef EPROCLIM
  case EPROCLIM:
    err= "too many processes [EPROCLIM]";
    break;
#endif
#ifdef EPROCUNAVAIL
  case EPROCUNAVAIL:
    err= "bad procedure for program [EPROCUNAVAIL]";
    break;
#endif
#ifdef EPROGMISMATCH
  case EPROGMISMATCH:
    err= "program version wrong [EPROGMISMATCH]";
    break;
#endif
#ifdef EPROGUNAVAIL
  case EPROGUNAVAIL:
    err= "RPC program not available [EPROGUNAVAIL]";
    break;
#endif
#ifdef EPROTO
  case EPROTO:
    err= "protocol error [EPROTO]";
    break;
#endif
#ifdef EPROTONOSUPPORT
  case EPROTONOSUPPORT:
    err= "protocol not suppored [EPROTONOSUPPORT]";
    break;
#endif
#ifdef EPROTOTYPE
  case EPROTOTYPE:
    err= "protocol wrong type for socket [EPROTOTYPE]";
    break;
#endif
#ifdef ERANGE
  case ERANGE:
    err= "math result unrepresentable [ERANGE]";
    break;
#endif
#if defined(EREFUSED) && (!defined(ECONNREFUSED) || (EREFUSED != ECONNREFUSED))
  case EREFUSED:
    err= "EREFUSED [EREFUSED]";
    break;
#endif
#ifdef EREMCHG
  case EREMCHG:
    err= "remote address changed [EREMCHG]";
    break;
#endif
#ifdef EREMDEV
  case EREMDEV:
    err= "remote device [EREMDEV]";
    break;
#endif
#ifdef EREMOTE
  case EREMOTE:
    err= "pathname hit remote file system [EREMOTE]";
    break;
#endif
#ifdef EREMOTEIO
  case EREMOTEIO:
    err= "remote i/o error [EREMOTEIO]";
    break;
#endif
#ifdef EREMOTERELEASE
  case EREMOTERELEASE:
    err= "EREMOTERELEASE [EREMOTERELEASE]";
    break;
#endif
#ifdef EROFS
  case EROFS:
    err= "read-only file system [EROFS]";
    break;
#endif
#ifdef ERPCMISMATCH
  case ERPCMISMATCH:
    err= "RPC version is wrong [ERPCMISMATCH]";
    break;
#endif
#ifdef ERREMOTE
  case ERREMOTE:
    err= "object is remote [ERREMOTE]";
    break;
#endif
#ifdef ESHUTDOWN
  case ESHUTDOWN:
    err= "can't send afer socket shutdown [ESHUTDOWN]";
    break;
#endif
#ifdef ESOCKTNOSUPPORT
  case ESOCKTNOSUPPORT:
    err= "socket type not supported [ESOCKTNOSUPPORT]";
    break;
#endif
#ifdef ESPIPE
  case ESPIPE:
    err= "invalid seek [ESPIPE]";
    break;
#endif
#ifdef ESRCH
  case ESRCH:
    err= "no such process [ESRCH]";
    break;
#endif
#ifdef ESRMNT
  case ESRMNT:
    err= "srmount error [ESRMNT]";
    break;
#endif
#ifdef ESTALE
  case ESTALE:
    err= "stale remote file handle [ESTALE]";
    break;
#endif
#ifdef ESUCCESS
  case ESUCCESS:
    err= "Error 0 [ESUCCESS]";
    break;
#endif
#if defined(ETIME) && (!defined(ELOOP) || (ETIME != ELOOP))
  case ETIME:
    err= "timer expired [ETIME]";
    break;
#endif
#if defined(ETIMEDOUT) && (!defined(ENOSTR) || (ETIMEDOUT != ENOSTR))
  case ETIMEDOUT:
    err= "connection timed out [ETIMEDOUT]";
    break;
#endif
#ifdef ETOOMANYREFS
  case ETOOMANYREFS:
    err= "too many references: can't splice [ETOOMANYREFS]";
    break;
#endif
#ifdef ETXTBSY
  case ETXTBSY:
    err= "text file or pseudo-device busy [ETXTBSY]";
    break;
#endif
#ifdef EUCLEAN
  case EUCLEAN:
    err= "structure needs cleaning [EUCLEAN]";
    break;
#endif
#ifdef EUNATCH
  case EUNATCH:
    err= "protocol driver not attached [EUNATCH]";
    break;
#endif
#ifdef EUSERS
  case EUSERS:
    err= "too many users [EUSERS]";
    break;
#endif
#ifdef EVERSION
  case EVERSION:
    err= "version mismatch [EVERSION]";
    break;
#endif
#if defined(EWOULDBLOCK) && (!defined(EAGAIN) || (EWOULDBLOCK != EAGAIN))
  case EWOULDBLOCK:
    err= "operation would block [EWOULDBLOCK]";
    break;
#endif
#ifdef EXDEV
  case EXDEV:
    err= "cross-domain link [EXDEV]";
    break;
#endif
#ifdef EXFULL
  case EXFULL:
    err= "message tables full [EXFULL]";
    break;
#endif
  default:
    if (message) {
      fprintf(stderr, "%s: unknown error [errno=%d]\n",
	      message, errno);
    } else {
      fprintf(stderr, "unknown error [errno=%d]\n", errno);
    }
    return;
  }
  if (message) {
    fprintf(stderr, "%s: %s\n", message, err);
  } else {
    fprintf(stderr, "%s\n", err);
  }
}
/*---------------------------------------------------------------------------*/
