#include "cp40tk.h"
#include "gpl_ops0.h"

/*---------------------------------------------------------------------------*/
/* CACHE MANAGEMENT */
static char small_cache[16384]; /* Small buffer cache for i/o or conversion,
				   if we need more we will alloc and free
				   a bigger cache. */
static void *get_cache(unsigned size);
static void  release_cache(void *cache);
static void *get_cache(unsigned size)
{
  void *cache;
  if (size <= sizeof small_cache) return small_cache;
  if ((cache= gpl_alloc(size)) != NULL) return cache;
  gpl_error("insufficient memory for %u bytes cache", size);
  return NULL;
}
static void  release_cache(void *cache)
{
  if (cache && cache!=small_cache) gpl_free(cache);
}

/*---------------------------------------------------------------------------*/
/* COLOR TABLE */

#define NLEVELS 1000  /* maximum number of levels for fast colorization */
#define NCOLORS  256  /* maximum number of colors */
#define RGB_MAX  255  /* maximum color intensity */
static unsigned char rgb_lut[3*NCOLORS];
static int rgb_ncolors= 0;

static void lut_gray(int ncolors);
static void lut_gray(int ncolors)
{
  int i, a, b;
  gpl_byte_t level;
  if (ncolors < 2) ncolors= 2;
  if (ncolors > NCOLORS) ncolors= NCOLORS;
  a= ncolors-1;
  b= a/2;
  for (i=0 ; i<ncolors ; i++) {
    level= (i*RGB_MAX + b) / a;
    rgb_lut[3*i]= rgb_lut[3*i+1]= rgb_lut[3*i+2]= level;
  }
  rgb_ncolors= ncolors;
}

/*---------------------------------------------------------------------------*/
static int task_image_string_match(char *string, char *format,
				   int *width_ptr, int *height_ptr);
static int task_image_string_read(Tcl_Interp *interp, char *string,
				  char *format, Tk_PhotoHandle handle,
				  int dst_x, int dst_y,
				  int width, int height,
				  int src_x, int src_y);

static Tk_PhotoImageFormat cp40_photo= {
  "CP40",
  NULL,
  task_image_string_match,
  NULL,
  task_image_string_read,
  NULL,
  NULL,
};

extern int Cp40tk_Init(Tcl_Interp *interp)
{
  Tk_CreatePhotoImageFormat(&cp40_photo);
  return TCL_OK;
}

/*---------------------------------------------------------------------------*/

static int task_image_string_match(char *string, char *format,
				   int *width_ptr, int *height_ptr)
{
  cp40_task_t *task;
  if (! (task= cp40tcl_get_task(string))) return 0;
  *width_ptr= task->width;
  *height_ptr= task->height;
  return 1;
}

/*---------------------------------------------------------------------------*/
/*	Colorize image block:
	  DST       - destination block (at least 3*NLINES*WIDTH elements)
	  SRC_PTR   - base address of block source
	  WIDTH     - width of block to colorize
	  NLINES    - height of block to colorize
	  SRC_PITCH - full width of source image
	  A, B      - linear scaling parameters: INTENSITY= (int)(A * SRC + B)
	  LUT=        color lookup table (a 3*NCOLORS array)
	                red_intensity= LUT[3*i]
	                green_intensity= LUT[3*i+1]
			blue_intensity= LUT[3*i+2]
	  NCOLORS=    number of RGB colors in LUT
*/

typedef void colorize_block_linear_op_t(unsigned char dst[], void *src_ptr,
					int width, int nlines, int src_pitch,
					double a, double b,
					unsigned char lut[], int ncolors);

#undef FUNCTION
#define FUNCTION(NAME, CTYPE)						\
static colorize_block_linear_op_t NAME;					\
static void NAME(unsigned char dst[], void *src_ptr,			\
		 int width, int nlines, int src_pitch,			\
		 double a, double b, unsigned char lut[], int ncolors)	\
{									\
  int i, j, k, kmax= 3*(ncolors-1), dst_pitch= 3*width;			\
  CTYPE *src= src_ptr;							\
  while (--nlines >= 0) {						\
    for (i=j=0 ; i<width ; i++, j+=3) {					\
      if ((k= (int)(a*src[i]+b)) < 0) k= 0;				\
      else if ((k*= 3) > kmax) k= kmax;					\
      dst[j]= lut[k]; dst[j+1]= lut[k+1]; dst[j+2]= lut[k+2];		\
    }									\
    dst+= dst_pitch;							\
    src+= src_pitch;							\
  }									\
}
FUNCTION(colorize_block_linear_b, gpl_byte_t)
FUNCTION(colorize_block_linear_s, short)
FUNCTION(colorize_block_linear_i, int)
FUNCTION(colorize_block_linear_l, long)
FUNCTION(colorize_block_linear_f, float)
FUNCTION(colorize_block_linear_d, double)

#undef FUNCTION
#define FUNCTION(NAME, CTYPE)						\
static colorize_block_linear_op_t NAME;					\
static void NAME(unsigned char dst[], void *src_ptr,			\
		 int width, int nlines, int src_pitch,			\
		 double a, double b, unsigned char lut[], int ncolors)	\
{									      \
  int i, j, k, kmax= 3*(ncolors-1), kone, dst_pitch= 3*width;		      \
  CTYPE *src= src_ptr;							      \
  if ((kone= kmax/2)<1) kone= kmax;					      \
  while (--nlines >= 0) {						      \
    for (i=j=0 ; i<width ; i++, j+=3) {					      \
      if ((k= src[i]) <= 0) k= 0;					      \
      else if (k==1) k= kone;						      \
      else k= kmax;							      \
      dst[j]= lut[k]; dst[j+1]= lut[k+1]; dst[j+2]= lut[k+2];		      \
    }									      \
    dst+= dst_pitch;							      \
    src+= src_pitch;							      \
  }									      \
}
FUNCTION(colorize_fov_b, gpl_byte_t)
FUNCTION(colorize_fov_s, short)
FUNCTION(colorize_fov_i, int)
FUNCTION(colorize_fov_l, long)
FUNCTION(colorize_fov_f, float)
FUNCTION(colorize_fov_d, double)

/*---------------------------------------------------------------------------*/

static int task_image_string_read(Tcl_Interp *interp, char *string,
				  char *format, Tk_PhotoHandle handle,
				  int dst_x, int dst_y,
				  int width, int height,
				  int src_x, int src_y)
{
  cp40_task_t *task;
  tim_array_t *array;
  int is_fov;
  char *src;
  void *cache;
  int nlines, src_pitch, src_width, src_height, src_pixel_size;
  Tk_PhotoImageBlock block;
  double a;
  colorize_block_linear_op_t *colorize_block;

  /* Get source. */
  if (! (task= cp40tcl_get_task(string))) return 0;
  is_fov= strcmp(task->token, "fov")==0;
  src_width= task->width;
  src_height= task->height;

  /* Fix dimensions. */
  if (width > src_width - src_x) width= src_width - src_x;
  if (height > src_height - src_y) height= src_height - src_y;
  if (width<=0 || height<=0 || src_x>=src_width || src_y>=src_height) {
    return TCL_OK;
  }
 
  /* Get maximum of image data and compute scale. */
  array= task->image->array;
  switch (array->id) {
  case TIM_ID_BYTE:
    a= (double)gpl_max_b(array->data, array->number);
    colorize_block= is_fov ? colorize_fov_b : colorize_block_linear_b;
    break;
  case TIM_ID_SHORT:
    a= (double)gpl_max_s(array->data, array->number);
    colorize_block= is_fov ? colorize_fov_s : colorize_block_linear_s;
    break;
  case TIM_ID_INT:
    a= (double)gpl_max_i(array->data, array->number);
    colorize_block= is_fov ? colorize_fov_i : colorize_block_linear_i;
    break;
  case TIM_ID_LONG:
    a= (double)gpl_max_l(array->data, array->number);
    colorize_block= is_fov ? colorize_fov_l : colorize_block_linear_l;
    break;
  case TIM_ID_FLOAT:
    a= (double)gpl_max_f(array->data, array->number);
    colorize_block= is_fov ? colorize_fov_f : colorize_block_linear_f;
    break;
  case TIM_ID_DOUBLE:
    a= (double)gpl_max_d(array->data, array->number);
    colorize_block= is_fov ? colorize_fov_d : colorize_block_linear_d;
    break;
  default:
    Tcl_AppendResult(interp, "bad pixel type in \"", string, "\"", NULL);
    return TCL_ERROR;
  }
  if (rgb_ncolors < 2) lut_gray(NCOLORS);
  if (a > 0.0) {
    a= (rgb_ncolors-1.0)/a;
  } else {
    a= 0.0;
  }

  /* Instanciate image block. */
  block.pixelSize= 3;
  block.offset[0]= 0;
  block.offset[1]= 1;
  block.offset[2]= 2;
  block.width= width;
  block.pitch= block.pixelSize * width; /* bytes per block line */
  
  /* Allocate block memory (cache). */
  nlines= sizeof(small_cache) / block.pitch;
  if (nlines < 1) nlines= 1;
  else if (nlines > height) nlines= height;
  cache= get_cache(nlines * block.pitch);
  block.pixelPtr= cache;
  block.height= nlines;

  /* Colorize image by blocks. */
  src_pixel_size= tim_size_of(array->id);
  src_pitch= src_width * src_pixel_size;
  src= (char *)array->data + src_pixel_size * (src_x + src_y * src_width);
  Tk_PhotoExpand(handle, dst_x + width, dst_y + height);
  while (height > 0) {
    if (nlines > height) block.height= (nlines= height);
    colorize_block(cache, src, width, nlines, src_width, a, 0.5,
		   rgb_lut, rgb_ncolors);
    Tk_PhotoPutBlock(handle, &block, dst_x, dst_y, width, nlines);
    dst_y+= nlines;
    src+= nlines * src_pitch;
    height-= nlines;
  }
  release_cache(cache);
  return TCL_OK;
}
/*---------------------------------------------------------------------------*/
#define UNDER_CONSTRUCTION 0
#if UNDER_CONSTRUCTION
#undef zmin
#define zmin 0 /* we already know that datamin=0 in CP40 images */
void foo()
{
  int i, number= array->number;
  CTYPE *data= array->data, zmax= data[0];

  for (i=1 ; i<number ; i++) if (data[i]>zmax) zmax= data[i];
  if ((datamax-datamin) < NLEVELS) {
  }
}

foo()
{
  if (fast) {
    for (i=0 ; i<number ; i++) {
      intensity= lut_intensity[data[i]];
      dst[i].r= lut_color[intensity].r;
      dst[i].g= lut_color[intensity].g;
      dst[i].b= lut_color[intensity].b;
    }
  }
}


void colorize(rgb_t *dst, const CTYPE *src, int number, double a, double b)
{
  int i, intensity;
  for (i=0 ; i<number ; i++) {
    intensity= (int)(a * src[i] + b);
    if (intensity < 0) intensity= 0;
    else if (intensity > intensity_max) intensity= intensity_max;
    dst[i].r= lut_color[intensity].r;
    dst[i].g= lut_color[intensity].g;
    dst[i].b= lut_color[intensity].b;
  }
}
#endif /* UNDER_CONSTRUCTION */
