/* NOTES:
   - Resize by window manager has no effect -> patch PGPLOT source (or my 
     window manager settings) to set the 'no resize' property.

   - In Yorick, two different pointers on the same array have the same value
     (and comparison operations == and != work on pointers).

   - Window hierarchy, e.g.:

       TopLevel (PGPLOT device, width, height)
         |
         +-- Panel 1 (used for widgets)
	 |     |
	 |     +- Widget 1
	 |     |
	 |     +- Widget 2
	 |     |
	 |     +- ...
	 |     |
	 |     `- Widget N
	 |
	 +-- Panel (used for plot)
	 |
	 +-- Panel (used for plot)
	 |
	 +-- Panel (used for plot)

   - Due to the way PGPLOT handles coordinates 

   - Optimization: turn off clipping for some operations.

   - PGplot only draws to devices and never remember what was drawn
*/

/* A TopLevel widget is a PGPLOT X11-window (one device). */
struct W_TopLevel {
  int id;               /* PGPLOT identifier */
  string device;        /* PGPLOT device name+file */
  float dpi;            /* pixel per inches (approximative) */
  float width, height;  /* view surface in device units (pixels) */
  pointer next;         /* chained list of panels */
}

/* Widget operation table. */
struct W_Ops {
  string ident;
  /* Methods (must be strings because Yorick does not allow
     pointers to functions). */
  string paint;    /* void paint(widget *self); */
  string resize;   /* void resize(widget *self, float x, float y, float width,
		                  float height); */
  string invoke;   /* void invoke(); */
}



/*---------------------------------------------------------------------------*/
/* LABEL WIDGET CLASS */

/*---------------------------------------------------------------------------*/
/* BUTTON WIDGET CLASS */

local _w_button_ops;
/* DOCUMENT local _w_button_ops;     operation table for `button' widgets */
_w_button_ops= W_Ops(ident="button", paint="_w_button_paint",
		     resize="_w_button_resize", invoke="_w_button_invoke");

struct W_Button {
  pointer ops;
  pointer panel;
  pointer next;
  string label;
  int fg, bg;
  float x, y, width, height, relief;
  string callback;
}

func w_button(p, label=, fg=, bg=, x=, y=, width=, height=, relief=, callback=)
{
  /* Set defaults. */
  w_default, relief, 2.0; /* default 2 pixels raised border */
  w_default, fg, W_GUI_FG;
  w_default, bg, W_GUI_BG;

  /* Create widget record. */
  w= &W_Button(panel=p, ops=&_w_button_ops, label=label, fg=fg, bg=bg,
	       x=x, y=y, width=width, height=height, relief=relief,
	       callback=callback);

  /* Insert widget in list of parent children. */
  w->next= p->next;
  p->next= w;
  return w;
}

func _w_button_invoke(w, event)
{
  if (w->callback) return symbol_def(w->callback)(w, event);
}

/* paint methods: assume the caller have already set the viewport and window
   (in pixel) and should be encapsulated between  __pgsave, __pgunsa calls */
func _w_button_paint(w)
{
  _w_paint_3d_rectangle(w->x, w->y, w->width, w->height,
			w->relief, w->bg);
  __pgsci, w->fg;
  //__pgscf, w->font;
  //__pgsch, w->size;
  if (strlen(w->label)) {
    __pgstbg, W_TRANSPARENT;
    __pgptxt, w->x+w->width/2.0F, w->y+w->height/2.0F, 0.0F, 0.5F, w->label;
  }
}

/*---------------------------------------------------------------------------*/
/* ENTRY WIDGET CLASS */

/*---------------------------------------------------------------------------*/
/* HIGH LEVEL ROUTINES */

struct W_Event {
  float x, y;
  char key;
};

func w_paint(w)
/* DOCUMENT: w_paint, w;
     (Re)paint widget pointed by W.
*/
{
  __pgsave;
  __pgbbuf;
  _w_panel_select, w->panel;
  symbol_def(w->ops->paint)(w);
  __pgebuf;
  __pgunsa;
}

func w_invoke(w, event)
/* DOCUMENT: result= w_invoke(w, event);
       -or-  w_invoke, w, event;
     Invoke widget pointed by W with mouse event EVENT.
*/
{
  __pgsave;
  __pgbbuf;
  _w_panel_select, w->panel;
  result= symbol_def(w->ops->invoke)(w, event);
  __pgebuf;
  __pgunsa;
  return result;
}

/*---------------------------------------------------------------------------*/
/* LOW LEVEL UTILITY ROUTINES */

/*
  LOW LEVEL ROUTINES:
    These routines (1) assume that the target PGPLOT device is already selected
    (i.e., by PGSLCT) and that the viewport and window are correctly set
    (e.g., by PGSVP and PGSWIN); (2) may modify attribute settings (i.e.,
    they do not call PGSAVE, PGUNSA); (3) act under the current buffering
    policy.  Their name start by the prefix "_w_..."

  HIGH LEVEL ROUTINES:
    These routines should left PGPLOT attributes unchanged and buffer
    PGPLOT calls for efficiency.
*/

func w_default(&sym, defval)
/* DOCUMENT: w_default, sym, defval;
       -or-  value= w_default(sym, defval);
     If variable SYM is nil, set variable SYM to default value DEFVAL
     (when called as a subroutine) or return DEFVAL.
*/
{
  if (! am_subroutine()) return (is_void(sym) ? defval : sym);
  if (is_void(sym)) sym= defval;
}

func _w_paint_3d_rectangle(x0, y0, width, height, relief, bg)
/* DOCUMENT: _w_paint_3d_rectangle, x0, y0, width, height, relief, bg;

     Low level routine that draw a solid rectangle of coordinates
     [X0,X0+WIDTH,Y0,Y0+HEIGHT] with color index BG
     If RELIEF=0 no border gets drawn; otherwise abs(RELIEF) is the
     thickness of the interior border to draw: the rectangle will appear
     sunken if RELIF<0, and raised if RELIEF>0.  */
{
  {extern W_GUI_LO, W_GUI_HI;}
  __pgsfs, 1; /* solid fill */
  x1= x0+width;
  y1= y0+height;
  /* Paint borders if any. */
  if (relief) {
    if (relief>0) {
      /* Rectangle will appear raised. */
      s= relief;
      ul= W_GUI_HI;
      ll= W_GUI_LO;
    } else {
      /* Rectangle will appear sunken. */
      s= -relief;
      ul= W_GUI_LO;
      ll= W_GUI_HI;
    }
    /* Paint upper-left sides. */
    __pgsci, ul;
    __pgpoly, 7, [x0, x0+s, x0+s, x1-s, x1, x0, x0],
                 [y0, y0+s, y1-s, y1-s, y1, y1, y0]; 
    /* Paint lower-left sides. */
    __pgsci, ll;
    __pgpoly, 7, [x0, x0+s, x1-s, x1-s, x1, x1, x0],
                 [y0, y0+s, y0+s, y1-s, y1, y0, y0];
    /* Paint interior of rectangle. */
    __pgsci, bg;
    __pgrect, x0+s, x1-s, y0+s, y1-s;
  } else {
    /* Paint interior of rectangle. */
    __pgsci, bg;
    __pgrect, x0, x1, y0, y1;
  }
}


/*---------------------------------------------------------------------------*/
/* WRAPPERS AROUND LOW LEVEL PGPLOT ROUTINES */
_w_pgslct= __pgslct;
func _w_pgsvp(box) { __pgsvp, box(1), box(2), box(3), box(3); }
func _w_pgswin(box) { __pgswin, box(1), box(2), box(3), box(3); }
PG_ZERO= 0.0F;
PG_ONE= 1.0F;

/*---------------------------------------------------------------------------*/
func w_toplevel_new(device, width, height)
/* DOCUMENT w= w_toplevel_new();
       -or- w= w_toplevel_new(device);
       -or- w= w_toplevel_new(device, width, height);
     Create a new PGPLOT page for DEVICE (default, "/xwin") optionally
     with dimensions WIDTH and HEIGHT (in pixels).
 */
{
  /* Open PGPLOT device. */
  if (is_void(device)) device= "/xwin";
  else if (structof(device)!=string) error, "bad type for DEVICE";
  if (! (id= __pgopen(device)))
    error, swrite("unable to open PGPLOT device \"%s\"", device);
  __pgask, 0; /* I don't want new-page prompting */
  if (catch(0x1f)) {
    __pgclos;
    error, catch_message;
  }

  /* Get page dimensions. */
  x0= x1= y0= y1= 0.0F;
  __pgqvsz, W_UNITS_PIXEL, x0, x1, y0, y1;
  cur_w= x1-x0;
  cur_h= y1-y0;
  __pgqvsz, W_UNITS_INCH, x0, x1, y0, y1;
  dpi_x= cur_w/(x1-x0);
  dpi_y= cur_h/(y1-y0);
  if (! is_void(width) && (width!=cur_w || height!=cur_h)) {
    /* Try to change the page dimensions. */
    __pgpap, width/dpi_x, (height/dpi_y)/(width/dpi_x);
    __pgpage;
    __pgqvsz, W_UNITS_PIXEL, x0, x1, y0, y1;
    width= x1-x0;
    height= y1-y0;
  } else {
    width= cur_w;
    height= cur_h;
  }
  dpi= sqrt(dpi_x * dpi_y);
  if (74.0<dpi && dpi<76.0) dpi= 75.0;
  else if (99.0<dpi && dpi<101.0) dpi= 100.0;
  return &W_TopLevel(id=id, device=device, dpi=dpi,
		     width=width, height=height);
}

func w_toplevel_resize(w, width, height)
{
  old_id= 0;
  __pgqid, old_id;
  if (old_id && w->id!=old_id) __pgslct, w->id;
  x0= x1= y0= y1= 0.0F;
  __pgqvsz, W_UNITS_PIXEL, x0, x1, y0, y1;
  w->width= x1-x0;
  w->height= y1-y0;
  if ((! is_void(width) && width!=w->width)
      || (! is_void(height) && height!=w->height)) {
    /* Try to change the page dimensions. */
    __pgqvsz, W_UNITS_INCH, x0, x1, y0, y1; /* 1 -> inches */
    dpi_x= w->width/(x1-x0);
    dpi_y= w->height/(y1-y0);
    __pgpap, width/dpi_x, (height/dpi_y)/(width/dpi_x);
    __pgpage;
    __pgqvsz, W_UNITS_PIXEL, x0, x1, y0, y1;
    w->width= x1-x0;
    w->height= y1-y0;
  }
  if (old_id && w->id!=old_id) __pgslct, old_id;
}

func w_panel_select(w, erase)
{
  __pgslct, w->id;
  a= w->box;
  if (erase) {
    x0= a(1); x1= a(2); y0= a(3); y1= a(4);
    __pgsvp, x0, x1, y0, y1;
    __pgswin, x0, x1, y0, y1;
    old_fs= new_fs= W_FILL_SOLID;
    old_ci= new_ci= w->bg;
    __pggfs, old_fs;
    __pgpci, old_ci;
    change_fs= (old_fs!=new_fs);
    change_ci= (old_ci!=new_ci);
    if (change_fs) __pgsfs, new_fs;
    if (change_ci) __pgsci, new_ci;
    __pgrect, x0, x1, y0, y1;
    if (change_fs) __pgsfs, old_fs;
    if (change_ci) __pgsci, old_ci;
  }

  /* Setup PGplot viewport. */
  if (anyof((margin= w->margin))) {
    a+= [1,-1,1,-1]*margin;
    __pgsvp, a(1), a(2), a(3), a(4);
  } else if (!erase) {
    __pgsvp, a(1), a(2), a(3), a(4);
  }

  /* Setup PGplot window. */
  a= w->world;
  __pgswin, a(1), a(2), a(3), a(4);
}

/*---------------------------------------------------------------------------*/
/* PANEL WIDGET CLASS */

/* A Panel widget is a container for one plot or for other widgets which
   defines the relationship between the device and the world coordinates. */
struct W_Panel {
  int id;
  int bg;
  float area(4);  /* NDC coordinates [XMIN,XMAX,YMIN,YMAX] of the total area
		     covered by the panel */
  float ndc(4);   /* NDC coordinates [XMIN,XMAX,YMIN,YMAX] of the viewport */
  float world(4); /* world coordinates [XMIN,XMAX,YMIN,YMAX] of the viewport */
}


func _w_panel_select(w)
{
  _w_pgslct, w->id;
  _w_pgsvp, w->ndc;
  _w_pgswin, w->world;
}

func _w_panel_paint(w)
{
  _w_pgslct, w->id;
  _w_pgsvp, w->area;
  __pgsci, w->bg;
  __pgsfs, PG_FILL_SOLID;
  __pgswin, 0.0F, 1.0F, 0.0F, 1.0F;
  __pgrect, 0.0F, 1.0F, 0.0F, 1.0F;
  _w_pgsvp, w->ndc;
  _w_pgswin, w->world;
}

func _w_panel_redraw(w)
{
  _w_panel_paint, w;
  while ((w= w->next))
    symbol_def(w->ops->paint)(w);
}

func _w_panel_destroy(w)
{
  /* no-op */
}

func w_panel(id=, bg=, margin=, world=, area=)
{
  old_id= 0;
  __pgqid, old_id;

  if (is_void(id)) {
    if (! old_id) old_id= __pgopen("/xwin");
    id= old_id;
  }

  if (is_void(bg)) bg= 0;
  if (is_void(margin)) margin= array(0.0F, 4);
  if (is_void(area)) area= [0.0F, 1.0F, 0.0F, 1.0F];
  if (is_void(world)) {
    /* World will be in pixel units. */
    x0= x1= y0= y1= 0.0F;
    __pgslct, id;
    __pgqvsz, W_UNITS_PIXEL, x0, x1, y0, y1;
    if (id!=old_id) __pgslct, old_id;
    ax= x1-x0;
    ay= y1-y0;
    world= [x0+ax*(area(1)+margin(1)), x0+ax*(area(2)-margin(2)),
            y0+ay*(area(3)+margin(3)), y0+ay*(area(4)-margin(4))];
  }
  return &W_Panel(id=id, bg=bg, area=area, margin=margin, world=world);
}

#if 0
{
  /* Setup window in pixel units. */
  nrows= 3;
  if (is_void(width)) width= 50.0F; /* Default WIDTH in pixels. */
  if (is_void(height)) height= 20.0F; /* Default HEIGHT in pixels. */
  x0= x1= y0= y1= 0.0F;
  __pgqvsz, 3, x0, x1, y0, y1;
  height= min(height, (y1-y0)/nrows/3);

  vp_width= 1.0F;
  vp_height= 0.15F;
  __pgsvp, 0.0F, vp_width, 1.0F-vp_height, 1.0F;
  vp_height*= (y1-y0);
  vp_width*= (x1-x0);
}

#endif
/*---------------------------------------------------------------------------*/
/* INITIALIZATION */
func w_init
/* DOCUMENT w_init;
     Initializes some constants needed by widget interface build on top
     of PGplot routines. */ 
{
  /* Setup colors. */
  {extern W_TRANSPARENT;}
  {extern W_BG, W_FG, W_WHITE, W_BLACK, W_RED, W_GREEN, W_BLUE;}
  {extern W_CYAN, W_MAGENTA, W_YELLOW, W_ORANGE;}
  {extern W_GUI_BG, W_GUI_FG, W_GUI_HI, W_GUI_LO, W_GUI_DISABLE;}
  W_TRANSPARENT= -1;
  __pgscr, (W_BG=           0), 0.0F, 0.0F, 0.0F;
  __pgscr, (W_FG=           1), 1.0F, 1.0F, 1.0F;
  __pgscr, (W_RED=          2), 1.0F, 0.0F, 0.0F;
  __pgscr, (W_GREEN=        3), 0.0F, 1.0F, 0.0F;
  __pgscr, (W_BLUE=         4), 0.0F, 0.0F, 1.0F;
  __pgscr, (W_CYAN=         5), 0.0F, 1.0F, 1.0F;
  __pgscr, (W_MAGENTA=      6), 1.0F, 0.0F, 1.0F;
  __pgscr, (W_YELLOW=       7), 1.0F, 1.0F, 0.0F;
  __pgscr, (W_ORANGE=       8), 1.0F, 0.5F, 0.0F;
  __pgscr, (W_BLACK=        9), 0.0F, 0.0F, 0.0F;
  __pgscr, (W_WHITE=       10), 1.0F, 1.0F, 1.0F;
  __pgscr, (W_GUI_BG=      11), 0.8F, 0.8F, 0.8F;
  __pgscr, (W_GUI_FG=      12), 0.0F, 0.0F, 0.0F;
  __pgscr, (W_GUI_HI=      13), 0.9F, 0.9F, 0.9F;
  __pgscr, (W_GUI_LO=      14), 0.4F, 0.4F, 0.4F;
  __pgscr, (W_GUI_DISABLE= 15), 0.5F, 0.5F, 0.5F;

  /* Font styles. */
  {extern W_FONT_SIMPLE, W_FONT_ROMAN, W_FONT_ITALIC, W_FONT_SCRIPT;}
  W_FONT_SIMPLE= 1;
  W_FONT_ROMAN=  2;
  W_FONT_ITALIC= 3;
  W_FONT_SCRIPT= 4;

  /* Fill-area styles. */
  {extern W_FILL_SOLID, W_FILL_OUTLINE, W_FILL_HATCHED, W_FILL_XHATCHED;}
  W_FILL_SOLID=    1;
  W_FILL_OUTLINE=  2;
  W_FILL_HATCHED=  3;
  W_FILL_XHATCHED= 4;

  /* Line styles. */
  {extern W_LINE_SOLID, W_LINE_DASHED, W_LINE_DASH_DOT, W_LINE_DOTTED;}
  {extern W_LINE_DASH_DOT3;}
  W_LINE_SOLID=     1;
  W_LINE_DASHED=    2;
  W_LINE_DASH_DOT=  3;
  W_LINE_DOTTED=    4;
  W_LINE_DASH_DOT3= 5;

  /* Length units (for PGQVP, PGQVSZ, PGLEN, PGQCS, ...). */
  W_UNITS_NDC=      0; /* normalized device coordinates */
  W_UNITS_INCH=     1; /* inches */
  W_UNITS_MM=       2; /* millimeters */
  W_UNITS_PIXEL=    3; /* absolute device coordinates: dots/pixels*/
  W_UNITS_WORLD=    4; /* world coordinates (only for PGLEN or PGQCS) */
  W_UNITS_RELATIVE= 5; /* fraction of the current viewport size (only for
			  PGLEN) */
}

local W_BG, W_FG, W_WHITE, W_BLACK, W_RED, W_GREEN, W_BLUE;
local W_CYAN, W_MAGENTA, W_YELLOW, W_ORANGE;
local W_GUI_BG, W_GUI_FG, W_GUI_HI, W_GUI_LO, W_GUI_DISABLE;
/* DOCUMENT W_BG            // background
            W_FG            // foreground
	    W_WHITE, W_BLACK, W_RED, W_GREEN, W_BLUE
            W_CYAN, W_MAGENTA, W_YELLOW, W_ORANGE
            W_GUI_BG, W_GUI_FG, W_GUI_HI, W_GUI_LO, W_GUI_DISABLE
     Indices of PGplot (predefined) colors.  The GUI ones should be
     considered as reserved for the widgets. */

if (W_BG!=1) w_init;

/*---------------------------------------------------------------------------*/
