/*
 * gpl_fft.h --
 *
 *	Definitions for Fast Fourier Transform in GPLib library.
 *
 * Copyright (c) 1999 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id$
 *	$Log$
 */
#ifndef _GPL_FFT_H
#define _GPL_FFT_H 1
#include "gpl_base.h"

/*---------------------------------------------------------------------------*/

extern int gpl_fft_f(float *re, float *im, int ntot, int n, int nspan,
		   int isn, void *ws);
extern int gpl_fft_d(double *re, double *im, int ntot, int n, int nspan,
		   int isn, void *ws);
extern int gpl_fft_c(gpl_fcomplex_t *z, int ntot, int n, int nspan,
		   int sign, void *ws);
extern int gpl_fft_z(gpl_dcomplex_t *z, int ntot, int n, int nspan,
		   int sign, void *ws);
/*	Compute multivariate complex fourier transform in place using
	mixed-radix fast fourier transform algorithm.  Apart from WS which
	must be NULL or a value returned by gpl_fft_setup, the other arguments
	have the same meaning as for FTMR (see FTMR below). */

extern void *gpl_fft_setup(int n, size_t size_of_real);
/*	Allocate and initializes workspace for further call to gpl_ftmr.
	N is the dimension of the current variable and SIZE_OF_REAL is
	sizeof(float) for gpl_ftmr_f and sizeof(double) for gpl_ftmr_d. In
	case of error NULL is returned and the error message can be
	retrieved with gpl_error_info.  The initialization mainly consists
	in the prime factor decomposition of N.  The same workspace can
	be used for any strides, as far as N and SIZE_OF_REAL remain the
	same. */

extern void gpl_fft_cleanup(void);
/*	Force freeing internal workspace possibly allocated by one of
	the gpl_fft_{f,d,c,z} routines when WS is passed as NULL. */

extern int gpl_fft_best_dim(int dim);
/*	Return the closest integer greater or equal DIM which is a factor
	of smal primes, i.e., 2, 3 and 5 only. */

extern void gpl_roll_b(gpl_byte_t dst[], const gpl_byte_t src[],
		     int stride, int len, int number, int offset);
extern void gpl_roll_s(short dst[], const short src[],
		     int stride, int len, int number, int offset);
extern void gpl_roll_i(int dst[], const int src[],
		     int stride, int len, int number, int offset);
extern void gpl_roll_l(long dst[], const long src[],
		     int stride, int len, int number, int offset);
extern void gpl_roll_f(float dst[], const float src[],
		     int stride, int len, int number, int offset);
extern void gpl_roll_d(double dst[], const double src[],
		     int stride, int len, int number, int offset);
extern void gpl_roll_c(gpl_fcomplex_t dst[], const gpl_fcomplex_t src[],
		     int stride, int len, int number, int offset);
extern void gpl_roll_z(gpl_dcomplex_t dst[], const gpl_dcomplex_t src[],
		     int stride, int len, int number, int offset);
/*	Apply a circular permutation by OFFSET elements from SRC to DST
	along one dimension of length LEN; STRIDE is the spacing of the
	elements along that dimension; NUMBER is the total number of
	elements in SRC anf DST.  SRC and DST must be different.
	Nothing is done for invalid arguments (e.g., if NUMBER is not
	a multiple of LEN*STRIDE). */

/*---------------------------------------------------------------------------*/
/* FTMR: MIXED-RADIX  FAST  FOURIER  TRANSFORM */

#define GPL_FTMR_NFACTORS 20  /* Number of elements in FTMR factor array (see
			       gpl_ftmr_factorize and gpl_ftmr) should be greater
			       or equal 5 + maximum number of prime factors
			       in any dimension. */

extern int gpl_ftmr_f(float *re, float *im, int ntot, int n, int nspan,
		    int isn, int *factor, float *at);
extern int gpl_ftmr_d(double *re, double *im, int ntot, int n, int nspan,
		    int isn, int *factor, double *at);
/* Multivariate  complex  fourier  transform,  computed  in  place  using
 * mixed-radix  fast  fourier  transform  algorithm by  R.  C.  Singleton
 * (Stanford Research Institute, Sept. 1968). The arguments are:
 *
 *   A (input/output - real array)
 *        Real part of the data  and, on return, of the resulting Fourier
 *        transform.
 *   B (input/output - real array)
 *        Imaginary part  of the  data and, on  return, of  the resulting
 *        Fourier transform.
 *   NTOT (input - integer)
 *        The total number of complex data values.
 *   N (input - integer)
 *        The dimension of the current variable.
 *   NSPAN (input - integer)
 *        Stride: NSPAN/N is the spacing of consecutive data values while
 *        indexing the current variable.
 *   ISN (input - integer)
 *        The sign of ISN determines the sign of the complex exponential,
 *        and the magnitude  of ISN is the spacing  of consecutive values
 *        in arrays  A and B.  Usually,  if the real  and imaginary parts
 *        are stored in different arrays ISN  is +/-1; or if the real and
 *        imaginary parts are consecutively stored in the same array then
 *        ISN is +/-2 and B=A+1.
 *   FACTOR (input - GPL_FFT_NFACTORS integer array)
 *        Factorization of N as returned by gpl_ftmr_factorize.
 *   WS (workspace array)
 *        Workspace array, must be large enough to store:
 *          4 * MAXF reals
 *        or
 *          2 * MAXF reals + MAXP integers
 *        whichever is the largest.  Beware  that  WS  must  be correctly
 *        aligned. Note: MAXF=FACTOR[3] and MAXP=FACTOR[4].
 *
 * Multivariate data  is indexed according  to the FORTRAN  array element
 * successor function,  without limit on  the number of  implied multiple
 * subscripts.   The subroutine  is called  once for  each  variate.  The
 * calls for a multivariate transform may be in any order.  A tri-variate
 * transform with A(N1,N2,N3), B(N1,N2,N3) is computed by:
 * 
 *     ftmr(A, B, N1*N2*N3, N1, N1, 1);
 *     ftmr(A, B, N1*N2*N3, N2, N1*N2, 1);
 *     ftmr(A, B, N1*N2*N3, N3, N1*N2*N3, 1);
 * 
 * For a single-variate transform, NTOT =  N = NSPAN = (number of complex
 * data values), e.g.
 * 
 *     ftmr(A, B, N, N, N, 1);
 * 
 * The data  can alternatively be stored  in a single complex  array C in
 * standard  FORTRAN  fashion,   i.e.   alternating  real  and  imaginary
 * parts.  Then, the  magnitude of  ISN changed  to two  to  give correct
 * indexing increment, and having, e.g.
 * 
 *     typedef float/double real_t;
 *     typedef struct {real_t re, im;} complex_t;
 *     complex_t Z[NTOT];
 *     real_t *A= (real_t *)Z;
 * 
 * the 2 following calls perform the same transform (on same data):
 * 
 *     ftmr(&Z[0].re, &Z[0].im, NTOT, n, nspan, 2);
 *     ftmr(A, A+1, NTOT, n, nspan, 2);
 */

extern void gpl_ftmr_factorize(int n, int factor[]);
/* Factorize  N in  primes for  FTMR algorithm  and store  them  in array
 * FACTOR which must have at least GPL_FFT_NFACTORS elements. On output:
 *     FACTOR[0]= N          Dimension for the transform.
 *     FACTOR[1]= M          The number of initialized factors in NFAC.
 *     FACTOR[2]= KT         (see FTMR).
 *     FACTOR[3]= MAXF       Maximum prime factor of N.
 *     FACTOR[4]= MAXP       Must be greater or equal the number of prime
 *                           factors   of  N.    In   addition,  if   the
 *                           square-free portion  K of N has  two or more
 *                           prime factors, then  MAXP must be greater or
 *                           equal K-1.  Actually MAXP=N-1 (See FTMR).
 *     FACTOR[5..M+4]= NFAC  Prime factors.
 */

/*---------------------------------------------------------------------------*/
/* REAL 2D FFT */
extern int gpl_forward_rfft2_f(gpl_fcomplex_t *dst, const float *src,
			     int nx, int ny, float *ws);
extern int gpl_forward_rfft2_d(gpl_dcomplex_t *dst, const double *src,
			     int nx, int ny, double *ws);
extern int gpl_backward_rfft2_f(float *dst, const gpl_fcomplex_t *src,
			      int nx, int ny);
extern int gpl_backward_rfft2_d(double *dst, const gpl_dcomplex_t *src,
			      int nx, int ny);


#endif /* _GPL_FFT_H ----------------------------------------------------------*/
