/*
 * gpl_hfile.c --
 *
 *	Implementation of simple editable hash-table files.  Another
 *	improvement with respect to actual implementation of management of
 *	simple hash table is that I tried to do the necessary housekeeping
 *	to avoid problems if some of these routines get interrupted,
 *	e.g. by user typing ^C.  This may seem too paranoid but is really
 *	needed if the GPLib library is linked with some interactive language,
 *	e.g. Yorick or IDL, which allow such interruptions.
 *
 * Copyright and warranty:
 *	Copyright (c) 1997 Eric THIEBAUT.
 *	
 *	This program is  free  software;  you  can  redistribute  it and/or
 *	modify it under the terms of  the  GNU  General  Public  License as
 *	published by the delete Software Foundation; either version 2  of the
 *	License, or (at your option) any later version.
 *	
 *	This program is distributed in the hope that it will be useful, but
 *	WITHOUT  ANY  WARRANTY;  without  even  the  implied   warranty  of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.   See  the GNU
 *	General Public License for more details.
 *	
 *	You should have received a copy of the GNU  General  Public License
 *	along with this program (see file ``LICENSE'' at the  top directory
 *	of  this  distribution);  if  not,  write  to  the   delete  Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *	
 *	To contact me, mail at: Eric THIEBAUT
 *				Observatoire de Lyon
 *				9, avenue Charles Andre
 *				F-69561 Saint Genis Laval Cedex
 *	or e-mail at:           thiebaut@obs.univ-lyon1.fr
 *
 * History:
 *	$Id$
 *	$Log$
 *-----------------------------------------------------------------------------
 */

#include <unistd.h>
#include <stdlib.h>
#include "gpl_base.h"
#include "gpl_string.h"
#include "gpl_hash.h"

/* Private routines and macros. */

/* Private data. */
static const char *empty= "";
static FILE *db_file= NULL; /* Database file stream. */

/*_____________________________________________________________________________
 *
 * gpl_hash_restore --
 *
 *	Read editable key-value file and store its contents in a hash table.
 *
 * Return:
 *	Standard result: GPL_OK or GPL_ERROR.
 *
 * Side effects:
 *	Output gpl_hash_table_t is assumed to have been correctly initialized
 *	by gpl_hash_init_table(table, GPL_HASH_STRING_KEYS) and its contents may
 *	be overwriten.
 *_____________________________________________________________________________
 */

int gpl_hash_restore(gpl_hash_table_t *table, const char *filename)
{
  static gpl_dstr_t *linePtr= NULL;
  static gpl_dstr_t *keyPtr= NULL;
  static gpl_dstr_t *valuePtr= NULL;
  static int init= 0;
  char *ptr, *string;
  int i, c, status= GPL_OK, eof=0, new;
  int length, line_count= 0;
  gpl_hash_entry_t *entry;
  
  if (! init) {
#undef  CANNOT_ALLOC
#define CANNOT_ALLOC(ptr, type) (ptr==NULL && (ptr= GPL_ALLOC_ONE(type))==NULL)
    GPL_ASSERT(! GPL_IS_SPACE('\0')); /* The following code relies on the fact
				     that '\0' is not a space. */
    if (   CANNOT_ALLOC(linePtr, gpl_dstr_t)
	|| CANNOT_ALLOC(keyPtr, gpl_dstr_t)
	|| CANNOT_ALLOC(valuePtr, gpl_dstr_t)) {
      gpl_sys_error("cannot create hash table strings");
      goto error;
    }
    gpl_dstr_init(linePtr);
    gpl_dstr_init(keyPtr);
    gpl_dstr_init(valuePtr);
    init= 1;
  } else {
    gpl_dstr_reset(linePtr);
    gpl_dstr_reset(keyPtr);
    gpl_dstr_reset(valuePtr);
  }
  if (table->key_type != GPL_HASH_STRING_KEYS) {
    return gpl_error("hash table key type must be `GPL_HASH_STRING_KEYS`");
  }
  if (gpl_open_static_file(&db_file, filename, "r") != GPL_OK) {
    return GPL_ERROR;
  }

  for (;;) {
    /* Read next line and get first non-space character. */
    if (! eof && gpl_dstr_read_line(linePtr, db_file, &eof, &line_count)!=GPL_OK) {
      goto error;
    }
    for (ptr=gpl_dstr_value(linePtr) ; GPL_IS_SPACE(ptr[0]) ; ptr++)
      ;
    c= ptr[0];
    
    /* Possibly append continuation line (append a newline if continuation
       line is filled with spaces). */
    if (c == '>') {
      if (gpl_dstr_length(keyPtr) <= 0) {
	/* Empty key string: nowhere to append anything... */
	goto unexpected_char;
      }
      for (ptr+=1, i=0 ; GPL_IS_SPACE(ptr[i]) ; i++)
	;
      if ((ptr[i]==0 ? gpl_dstr_append_char(valuePtr, '\n')
	   : gpl_dstr_append(valuePtr, ptr, -1)) != GPL_OK) {
	goto error;
      }
      continue;
    }
    
    /* Save previously parsed key-value pair if any and reset the
       corresponding dynamic strings to be empty. */
    if (gpl_dstr_length(keyPtr) > 0) {
      entry= gpl_hash_create_entry(table,gpl_dstr_value(keyPtr), &new);
      if (entry == NULL || (string= gpl_dstr_steal(valuePtr)) == NULL) {
	goto error;
      }
      gpl_hash_set_value(entry, string, GPL_HASH_DYNAMIC);
      gpl_dstr_reset(keyPtr);
    }
    
    /* Skip empty line and comments. */
    if (c == 0 || c == '#') {
      if (eof) break;
      continue;
    }
    
    /* Store key and value strings. */
    for (length=i=0 ; ; i++) {
      if ((c= ptr[i]) == '=') {
	length= i;
	break;
      }
      if (GPL_IS_SPACE(c)) {
	length= i;
	for (i+=1 ; GPL_IS_SPACE(ptr[i]) ; i++)
	  ;
	c= ptr[i];
	break;
      }
      if (c==0 || c=='>') {
	break;
      }
    }
    if (length <= 0 || c != '=') {
      goto unexpected_char;
    }
    if (gpl_dstr_append(keyPtr, ptr, length) != GPL_OK) {
      goto error;
    }
    for (ptr+=i+1 ; GPL_IS_SPACE(ptr[0]) ; ptr++)
      ;
    if (ptr[0] != 0 && gpl_dstr_append(valuePtr, ptr, -1) != GPL_OK) {
      goto error;
    }
  }
  
 done:
  gpl_close_static_file(&db_file);
  gpl_dstr_free(linePtr);
  gpl_dstr_free(keyPtr);
  gpl_dstr_free(valuePtr);
  return status;
  
 unexpected_char:
  if (c >= ' ' && c <= '~') {
    gpl_error("unexpected char '%c' in hash file at line %d", c, line_count);
  } else if (c==0) {
    gpl_error("unexpected end-of-line in hash file at line %d", line_count);
  } else {
    gpl_error("unexpected char 0x%02x in hash file at line %d", (unsigned)c,
	    line_count);
  }
 error:
  status= GPL_ERROR;
  goto done;
}

/*_____________________________________________________________________________
 *
 * gpl_hash_save --
 *
 *	Simple driver to save hash table into editable file.
 *
 * Return:
 *	Standard result: GPL_OK or GPL_ERROR.
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

int gpl_hash_save(gpl_hash_table_t *table, const char *filename, int overwrite)
{
  gpl_hash_entry_t *entry;
  gpl_hash_search_t search;
  
  if (table->key_type != GPL_HASH_STRING_KEYS) {
    return gpl_error("hash table key type must be `GPL_HASH_STRING_KEYS'");
  }
  if (filename == NULL || filename[0] == 0) {
    return gpl_error("bad file name for saving hash-table");
  }
  if (access(filename, F_OK) == 0 && ! overwrite) {
    return gpl_error("cannot save hash-table: file already exists");
  }
  if (gpl_open_static_file(&db_file, filename, "w") != GPL_OK) {
    return GPL_ERROR;
  }
  
  /* Loops over every hash entries. */
  for (entry= gpl_hash_first_entry(table, &search);
       entry != NULL;
       entry= gpl_hash_next_entry(&search)) {
    if (gpl_hash_write_entry(db_file, entry->key.string,
		       (char *)entry->client_data, 0) != GPL_OK) {
      gpl_close_static_file(&db_file);
      return GPL_ERROR;
    }
  }
  gpl_close_static_file(&db_file);
  return GPL_OK;
}

/*_____________________________________________________________________________
 *
 * gpl_hash_write_comment --
 *
 *	Print out a comment line in a database file (STRING must not contain
 *	any newline; this is not checked).
 *
 * Return:
 *	None.
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

int gpl_hash_write_comment(FILE *file, const char *string)
{
  if ((string!=NULL && (fputs("# ", file) < 0 || fputs(string, file) < 0))
      || putc('\n', file) < 0) {
    return gpl_sys_error("error writing database comment");
  }
  return GPL_OK;
}

/*_____________________________________________________________________________
 *
 * gpl_hash_write_entry --
 *
 *	Print a hash entry key-value pair taking care of splitting value
 *	if it is too long (if COMMENT is non-zero the resulting line(s) get
 *	commented).
 *
 * Return:
 *	None.
 *
 * Side effects:
 *	See description.
 *_____________________________________________________________________________
 */

#define MAX_LENGTH 50

int gpl_hash_write_entry(FILE *file, const char *key, const char *value, int comment)
{
  int c;
  int print= 0;
  char *sign= "= "; /* Equal sign followed by a space. */
  long i0= 0;       /* Begin of string to print. */
  long i1= -1;      /* Current position in value string. */
  long j= -1;       /* Position of last space. */
  const char *format= comment ? "# %18s %s" : "%20s %s";
  
  if (value==NULL) value= empty;
  if (comment) {
    if (key==NULL) {
      key= empty;
    }
  } else {
    /* Check key. */
    int i= -1;
    if (key!=NULL) {
      while ((c= key[++i]) != 0) {
	if (GPL_IS_SPACE(c) || c=='=' || c=='>') {
	  i= 0;
	  break;
	}
      }
    }
    if (i<=0) {
      gpl_warning("skipping hash table entry (bad key)");
      return GPL_OK;
    }
  }
  
  /* Write hash entry key-value. */
  for (;;) {
    c= value[++i1];
    if (c==0 || GPL_IS_SPACE(c)) {
      j= i1;
      print= c==0 || c=='\n';
    }
    if (print || (i1 - i0) >= MAX_LENGTH) {
      if (j <= i0) {
	j= i1;
      }
      if (j >= i0) {
	if (fprintf(file, format, key, sign) <= 0
	    || (j > i0 && fwrite(value+i0, 1, j-i0, file) != j-i0)
	    || putc('\n', file) < 0) {
	  break;
	}
      }
      if ((c= value[j]) == 0) {
	return GPL_OK;
      }
      if (c == '\n') {
	if (fprintf(file, format, empty, ">\n") <= 0) {
	  break;
	}
	i0= j+1;
      } else {
	i0= j;
      }
      key= empty;
      sign= ">"; /* Continuation sign followed by NO space. */
      print= 0;
    }
  }
  return gpl_sys_error("error writing hash table entry");
}
/*---------------------------------------------------------------------------*/
