/*
 * gpl_num.h --
 *
 *	Definitions for numerical routines from various FORTRAN public
 *	domain libraries (LAPACK, SLATEC, LINPACK, EISPACK, MINPACK, ...)
 *	implemented in GPLib library.
 *
 *	I have tried to keep the code portable and easy to use without
 *	GPLib.   To that end, the only external definitions imported from
 *	GPLib are:
 *	  o type definition:
 *	     - gpl_fcomplex_t
 *	     - gpl_dcomplex_t
 *	  o routines:
 *	     - gpl_error
 *	     - gpl_fatal_error
 *
 * Copyright (c) 1999 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id$
 *	$Log$
 *-----------------------------------------------------------------------------
 */
#ifndef _GPL_NUM_H
#define _GPL_NUM_H 1

/* Import some GPLib definitions (see note in header above). */
#include "gpl_base.h"

/* C++ needs to know that types and declarations are C, not C++. */
#ifdef  __cplusplus
# define _GPL_NUM_BEGIN_DECLS  extern "C" {
# define _GPL_NUM_END_DECLS    }
#else
# define _GPL_NUM_BEGIN_DECLS  /* empty */
# define _GPL_NUM_END_DECLS    /* empty */
#endif
_GPL_NUM_BEGIN_DECLS

/*---------------------------------------------------------------------------*/
/* ROUTINES TO QUERY MACHINE CONSTANTS */

/* Determine single/double precision machine parameters:

   CMACH   result
   ------  ----------------------------------------------------------
   'E' 'e'  EPS   = relative machine precision
   'S' 's'  SFMIN = safe minimum, such that 1/SFMIN does not overflow
   'B' 'b'  BASE  = base of the machine
   'P' 'p'  PREC  = EPS*BASE
   'N' 'n'  T     = number of (BASE) digits in the mantissa
   'R' 'r'  RND   = 1 when rounding occurs in addition, 0 otherwise
   'M' 'm'  EMIN  = minimum exponent before (gradual) underflow
   'U' 'u'  RMIN  = underflow threshold - BASE**(EMIN-1)
   'L' 'l'  EMAX  = largest exponent before overflow
   'O' 'o'  RMAX  = overflow threshold - (BASE**EMAX)*(1-EPS)
   ------  ----------------------------------------------------------

   Note: These constants are dynamically computed the first time the
	 routine is called.  Since this computation may be time consuming,
	 you may explicitly call gpl_num_slamch and/or gpl_num_dlamch
	 before timing parts of your code. */
extern float  gpl_num_slamch(char cmach);
extern double gpl_num_dlamch(char cmach);

/* Return integer machine data representation constants:

   call          C-type        result                                 
   ------------  ------------  ---------------------------------------
   gpl_num_i1mach( 5)     int       number of bits
   gpl_num_i1mach( 6)     int       number of characters
   gpl_num_i1mach( 7)     int       A, the base
   gpl_num_i1mach( 8)     int       S, number of base-A digits per integer
   gpl_num_i1mach( 9)     int       A**S - 1, the largest magnitude
   gpl_num_i1mach(10) float/double  B, the base for floating-point numbers
   gpl_num_i1mach(11)    float      T, the number of base-B digits
   gpl_num_i1mach(12)    float      EMIN, the smallest exponent E
   gpl_num_i1mach(13)    float      EMAX, the largest exponent E
   gpl_num_i1mach(14)    double     T, the number of base-B digits
   gpl_num_i1mach(15)    double     EMIN, the smallest exponent E
   gpl_num_i1mach(16)    double     EMAX, the largest exponent E

   Notes:
     - It is assumed that INTEGERS are represented in the S-digit,
       base-A form:
          SIGN ( X(S-1)*A**(S-1) + ... + X(1)*A + X(0) )
       where 0<=X(i)<A for i=0..S-1.

    - It is assumed that floating-point numbers are represented in the
      T-digit, base-B form:
          SIGN (B**E)*( (X(1)/B) + ... + (X(T)/B**T) )
      where 0<=X(i)<B for i=1..T  and 0<X(1) and EMIN<=E<=EMAX.

   - Invalid WHICH argument results in a fatal error (see gpl_fatal_error). */
extern int gpl_num_i1mach(int which);

/* Return machine data representation constants for floating point numbers
   (gpl_num_r1mach is for C-type `float' and gpl_num_d1mach is for C-type
   `double'):

   call          result                                 
   ------------  --------------------------------------------
   gpl_num_r1mach(1)  B**(EMIN-1), the smallest positive magnitude
   gpl_num_r1mach(2)  B**EMAX*(1 - B**(-T)), the largest magnitude
   gpl_num_r1mach(3)  B**(-T), the smallest relative spacing
   gpl_num_r1mach(4)  B**(1-T), the largest relative spacing
   gpl_num_r1mach(5)  LOG10(B)
   
   Notes:
       - gpl_num_r1mach(4) is the so-called machine EPSILON value; usually
	 used to define machine precision.

      - Invalid WHICH argument results in a fatal error (see gpl_fatal_error). */
extern float gpl_num_r1mach(int which);
extern double gpl_num_d1mach(int which);

/*---------------------------------------------------------------------------*/
/* MISCELLANEOUS */

/* Return true (non-zero) if relative difference between A and B is smaller or
   equal tolerance TOL (TOL >= 0). */
extern int gpl_num_almost_equal(double a, double b, double tol);

/*---------------------------------------------------------------------------*/
/* BASIC LINEAR ALGEBRA ROUTINES (e.g. BLAS) */

/* Find the smallest index of that component of a vector having the maximum
   magnitude.  N is the number of elements to consider, X is the input array
   and INCX the spacing of elements. */
extern int gpl_num_iamax_f(int n, const float x[], int incx);
extern int gpl_num_iamax_d(int n, const double x[], int incx);

/* Find the smallest index of the component of a complex vector having the
   maximum sum of magnitudes of real and imaginary parts.  N is the number of
   elements to consider, Z is the input array and INCZ the spacing of
   elements. */
extern int gpl_num_iamax_c(int n, const gpl_fcomplex_t z[], int incz);
extern int gpl_num_iamax_z(int n, const gpl_dcomplex_t z[], int incz);

/* Return sum of absolute value of elements of X (sum of absolute value of
   real part and imaginary part elements of Z).  INCR is the increment
   (possibly negative) and NUMBER is the number of elements to sum. */
extern float  gpl_num_asum_f(int number, const float x[], int incr);
extern double gpl_num_asum_d(int number, const double x[], int incr);
extern float  gpl_num_asum_c(int number, const gpl_fcomplex_t z[], int incr);
extern double gpl_num_asum_z(int number, const gpl_dcomplex_t z[], int incr);

/* Returns the euclidean norm of vector X taking care of under/overflows.
   Result is sqrt( X'.X ).  N is the number of elements to sum and INCX is
   the increment between elements. */
extern double gpl_num_enorm_d(int n, const double x[]);
extern float  gpl_num_enorm_f(int n, const float x[]);
extern double gpl_num_nrm2_d(int n, const double x[], int incx);
extern float  gpl_num_nrm2_f(int n, const float x[], int incx);

/* Return dot product between X and Y.  N is the number of elements in the
   dot product. INCX and INCY are the spacing of elements in arrays X and Y
   respectively.  The routine gpl_num_ddot use double precision arihmetic. */
extern float  gpl_num_dot_f(int n, const float x[], int incx,
			           const float y[], int incy);
extern double gpl_num_ddot_f(int n, const float x[], int incx,
			            const float y[], int incy);
extern double gpl_num_dot_d(int n, const double x[], int incx,
			           const double y[], int incy);

/* Store result of elementwise A*X + Y into Y. N is the number of elements to
   consider. INCX and INCY are the spacing of elements in arrays X and Y
   respectively. */
extern void gpl_num_axpy_f(int n, float a, const float x[], int incx,
			                         float y[], int incy);
extern void gpl_num_axpy_d(int n, double a, const double x[], int incx,
			                          double y[], int incy);

/* In-place rescaling of an array.
   Multiply N elements of X by A every INCX elements. */
extern void gpl_num_scal_f(int n, float a, float x[], int incx);
extern void gpl_num_scal_d(int n, double a, double x[], int incx);

/* Vector copy.
   Copy N elements from array X (every INCX elements) to array Y (every INCY
   elements). */
extern void gpl_num_copy_f(int n, const float  x[], int incx,
			                float  y[], int incy);
extern void gpl_num_copy_d(int n, const double x[], int incx,
			                double y[], int incy);

/* Swap vector elements.
   Exchange N elements between array X (every INCX elements) and
   array Y (every INCY elements). */
extern void gpl_num_swap_f(int n, float x[], int incx, float y[], int incy);
extern void gpl_num_swap_d(int n, double x[], int incx, double y[], int incy);

/* Apply a plane Givens rotation. */
extern void gpl_num_rot_f(int n, float x[], int incx, float y[], int incy,
			  float cs, float sn);
extern void gpl_num_rot_d(int n, double x[], int incx, double y[], int incy,
			  double cs, double sn);

/* Construct a plane Givens rotation. */
extern void gpl_num_rotg_f(float *a, float *b, float *cs, float *sn);
extern void gpl_num_rotg_d(double *a, double *b, double *cs, double *sn);


/* Solve tridiagonal system. */
extern void gpl_num_gtsl_f(int n, float c[], float d[], float e[],
			   float b[], int *info);
extern void gpl_num_gtsl_d(int n, double c[], double d[], double e[],
			   double b[], int *info);

extern void gpl_num_pofa_f(float a[], int lda, int n, int *info);
extern void gpl_num_pofa_d(double a[], int lda, int n, int *info);
extern void gpl_num_posl_f(const float a[], int lda, int n, float *b);
extern void gpl_num_posl_d(const double a[], int lda, int n, double *b);

/*---------------------------------------------------------------------------*/
/* SOLVE GENERAL LINEAR SYSTEM BY LU FACTORIZATION */

/* Factor a matrix using Gaussian elimination. */
extern void gpl_num_gefa_f(float a[], int lda, int n, int ipvt[], int *info);
extern void gpl_num_gefa_d(double a[], int lda, int n, int ipvt[], int *info);
/*    GPL_NUM_GEFA is usually called by GPL_NUM_GECO, but it can be called
      directly with a saving in time if RCOND is not needed.  (Time for
      GPL_NUM_GECO) = (1 + 9/N)*(Time for GPL_NUM_GEFA).

      On Entry:
        A     - The LDA by N matrix to be factored.
        LDA   - The leading dimension of the array  A.
        N     - The order of the matrix  A.

      On Return:
        A     - An upper triangular matrix and the multipliers which were
	        used to obtain it.   The factorization can be written
		A = L*U , where L  is a product of permutation and unit lower
		triangular matrices and  U  is upper triangular.
        IPVT  - An integer N-vector of pivot indices.
        INFO  - = 0  normal value.
                = K  if  U(K,K)==0.0 .  This is not an error condition for
		     this subroutine, but it does indicate that SGESL or
		     SGEDI will divide by zero if called.  Use RCOND in
		     SGECO for a reliable indication of singularity.

   REFERENCES  J. J. Dongarra, J. R. Bunch, C. B. Moler, and G. W.
                 Stewart, LINPACK Users' Guide, SIAM, 1979.
   ROUTINES CALLED  ISAMAX, SAXPY, SSCAL
   REVISION HISTORY  (YYMMDD)
   780814  DATE WRITTEN
   890831  Modified array declarations.  (WRB)
   890831  REVISION DATE from Version 3.2
   891214  Prologue converted to Version 4.0 format.  (BAB)
   900326  Removed duplicate information from DESCRIPTION section.
           (WRB)
   920501  Reformatted the REFERENCES section.  (WRB)
   99????  FORTRAN -> C (ET) */

/* Factor a matrix using Gaussian elimination and estimate
   the condition number of the matrix. */
extern void gpl_num_geco_f(float a[], int lda, int n, int ipvt[],
			   float *rcond, float z[]);
extern void gpl_num_geco_d(double a[], int lda, int n, int ipvt[],
			   double *rcond, double z[]);

/* Compute the determinant and inverse of a matrix using the factors computed
   by GPL_NUM_GECO or GPL_NUM_GEFA. */
extern void gpl_num_gedi_f(float a[], int lda, int n, int ipvt[], 
			   float det[], float work[], int job);
extern void gpl_num_gedi_d(double a[], int lda, int n, int ipvt[], 
			   double det[], double work[], int job);

/* GPL_NUM_GESL Solve the real system A*X=B or TRANS(A)*X=B using the
   factors of GPL_NUM_GECO or GPL_NUM_GEFA. */
extern void gpl_num_gesl_f(const float a[], int lda, int n,
			   const int ipvt[], float b[], int job);
extern void gpl_num_gesl_d(const double a[], int lda, int n,
			   const int ipvt[], double b[], int job);

/* GPL_NUM_GEFS (DRIVER ROUTINE)
            Solve a general system of linear equations. */
extern int gpl_num_gefs_f(float a[], int lda, int n, float v[], 
			  int task, int *ind, float work[], int ipvt[]);
extern int gpl_num_gefs_d(double a[], int lda, int n, double v[], 
			  int task, int *ind, double work[], int ipvt[]);

/* GPL_NUM_GEIR (DRIVER ROUTINE)
            Solve a general system of linear equations.  Iterative
	    refinement is used to obtain an error estimate. */
extern int gpl_num_geir_f(float a[], int lda, int n, float v[], 
			  int task, int *ind, float work[], int ipvt[]);
extern int gpl_num_geir_d(double a[], int lda, int n, double v[], 
			  int task, int *ind, double work[], int ipvt[]);

/*---------------------------------------------------------------------------*/
/* QR FACTORIZATION
   (e.g. to find least square solution of an overdetermined linear system) */

extern void gpl_num_qrdc_f(float x[], int ldx, int n, int p, 
			   float qraux[], int jpvt[], float work[], int job);
extern void gpl_num_qrdc_d(double x[], int ldx, int n, int p, 
			   double qraux[], int jpvt[], double work[], int job);

extern void gpl_num_qrsl_f(const float x[], int ldx, int n, int k,
			   const float qraux[], const float y[],
			   float qy[], float qty[], float b[],
			   float rsd[], float xb[], int *info);
extern void gpl_num_qrsl_d(const double x[], int ldx, int n, int k,
			   const double qraux[], const double y[],
			   double qy[], double qty[], double b[],
			   double rsd[], double xb[], int *info);

extern void gpl_num_rwupdt_f(int n, float r[], int ldr, const float w[], 
			     float b[], float *alpha,
			     float gcos[], float gsin[]);
extern void gpl_num_rwupdt_d(int n, double r[], int ldr, const double w[], 
			     double b[], double *alpha,
			     double gcos[], double gsin[]);

extern void gpl_num_qrfac_f(int m, int n, float a[], int lda, int pivot[],
			    float sigma[], float acnorm[], float wa[]);
extern void gpl_num_qrfac_d(int m, int n, double a[], int lda, int pivot[],
			    double sigma[], double acnorm[], double wa[]);

extern void gpl_num_qrsolv_f(int n, float r[], int ldr, const int pivot[],
			     float diag[], const float qtb[], float x[],
			     float sigma[], float wa[]);
extern void gpl_num_qrsolv_d(int n, double r[], int ldr, const int pivot[],
			     double diag[], const double qtb[], double x[],
			     double sigma[], double wa[]);
/*---------------------------------------------------------------------------*/
/* SINGULAR VALUE DECOMPOSITION */

/* Perform the singular value decomposition of a rectangular matrix. */
extern void gpl_num_svdc_f(float x[], int ldx, int m, int n, float s[],
			   float e[], float u[], int ldu, float v[],
			   int ldv, float work[], int job, int *info);
extern void gpl_num_svdc_d(double x[], int ldx, int m, int n, double s[],
			   double e[], double u[], int ldu, double v[],
			   int ldv, double work[], int job, int *info);

/* Check SVD factorization: given the Singular Value Decomposition of
   a matrix A, compute Y=A*X. */
extern void gpl_num_svchk_f(float y[], const float x[], int m, int n,
			    const float s[], const float u[], int ldu,
			    const float v[], int ldv, float ws[]);
extern void gpl_num_svchk_d(double y[], const double x[], int m, int n,
			    const double s[], const double u[], int ldu,
			    const double v[], int ldv, double ws[]);

/* Given the Singular Value Decomposition of a matrix A, solve the
   regularized problem: A*X=B and return the rank of A. */
extern int gpl_num_svslv_f(float x[], const float b[], int m, int n,
			   const float s[], const float u[], int ldu,
			   const float v[], int ldv, float r, float ws[]);
extern int gpl_num_svslv_d(double x[], const double b[], int m, int n,
			   const double s[], const double u[], int ldu,
			   const double v[], int ldv, double r, double ws[]);

/*---------------------------------------------------------------------------*/
/* Factor a real symmetric positive definite matrix and estimate the
   condition number of the matrix. */
extern void gpl_num_poco_f(float a[], int lda, int n, float *rcond,
			   float z[], int *info);
extern void gpl_num_poco_d(double a[], int lda, int n, double *rcond,
			   double z[], int *info);

/* Drivers routines (return GPL_OK or GPL_ERROR with error message that can be
   retrieved with gpl_error_info). */
extern int gpl_num_pofs_f(float a[], int lda, int n, float v[], int itask,
			  int *ind, float work[]);
extern int gpl_num_pofs_d(double a[], int lda, int n, double v[], int itask,
			  int *ind, double work[]);


/*---------------------------------------------------------------------------*/
/* OPERATOR ROUTINES TO MOCK OPTIMIZER SHORTCUTS */

/* These routines are intended to force A and B to be stored prior to doing
   the operation with A and B , for use in situations where optimizers might
   hold one of these in a register. */

/* Arithmetic Operators */
extern float  gpl_num_add_f( float a,  float b);
extern double gpl_num_add_d(double a, double b);
extern float  gpl_num_sub_f( float a,  float b);
extern double gpl_num_sub_d(double a, double b);
extern float  gpl_num_mul_f( float a,  float b);
extern double gpl_num_mul_d(double a, double b);
extern float  gpl_num_div_f( float a,  float b);
extern double gpl_num_div_d(double a, double b);

/* Comparison Operators
     gpl_num_eq_?(A,B)    A == B
     gpl_num_ne_?(A,B)    A != B
     gpl_num_ge_?(A,B)    A >= B
     gpl_num_gt_?(A,B)    A >  B
     gpl_num_le_?(A,B)    A <= B
     gpl_num_lt_?(A,B)    A <  B
     gpl_num_cmp_?(A,B)   -1, if A<B; 1, if A>B; 0, if A==B */
extern int gpl_num_eq_f(  float a,  float b);
extern int gpl_num_eq_d( double a, double b);
extern int gpl_num_ne_f(  float a,  float b);
extern int gpl_num_ne_d( double a, double b);
extern int gpl_num_ge_f(  float a,  float b);
extern int gpl_num_ge_d( double a, double b);
extern int gpl_num_gt_f(  float a,  float b);
extern int gpl_num_gt_d( double a, double b);
extern int gpl_num_le_f(  float a,  float b);
extern int gpl_num_le_d( double a, double b);
extern int gpl_num_lt_f(  float a,  float b);
extern int gpl_num_lt_d( double a, double b);
extern int gpl_num_cmp_f( float a,  float b);
extern int gpl_num_cmp_d(double a, double b);

/* Copy Operators */
extern void gpl_num_store_f( float *addr,  float value);
extern void gpl_num_store_d(double *addr, double value);

/*---------------------------------------------------------------------------*/
_GPL_NUM_END_DECLS
#endif /* _GPL_NUM_H --------------------------------------------------------*/

