/*
 * gpl_num_mach.c --
 *
 *	Determines constants for machine data representation.
 *	Translated from LAPACK and SLATEC routines.
 *
 * Copyright (c) 1999-2000 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id$
 *	$Log$
 *-----------------------------------------------------------------------------
 */

#ifndef _GPL_NUM_MACH_C
#define _GPL_NUM_MACH_C 1

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#undef PROTO
#if defined(NEED_PROTOTYPES) || defined(__STDC__) || defined(__cplusplus)
# define PROTO(x) x
#else
# define PROTO(x) ()
#endif

#undef  ABS_DIF
#define ABS_DIF(a,b) ((a)>=(b)?(a)-(b):(b)-(a))
#undef  MIN
#define MIN(a,b)     ((a)<=(b)?(a):(b))
#undef  MAX
#define MAX(a,b)     ((a)>=(b)?(a):(b))
#undef  BOOLEAN
#define BOOLEAN(expr) ((expr)?1:0)

/* FORTRAN types. */
#undef logical_t
#define logical_t int
#undef integer_t
#define integer_t long /* some machine may require `long' here */

/* Exported routines (these definitions must match those in "gpl_num.h"). */
extern float  gpl_num_slamch PROTO((char cmach));
extern double gpl_num_dlamch PROTO((char cmach));
extern int    gpl_num_i1mach PROTO((int i));
extern float  gpl_num_r1mach PROTO((int i));
extern double gpl_num_d1mach PROTO((int i));

/*---------------------------------------------------------------------------*/
/* PRIVATE DATA */
static int    need_init= 1;
static int    i1mach_data[12];
static float  r1mach_data[5];
static double d1mach_data[5];

#define I1MACH(i)  i1mach_data[(i)-5]
#define R1MACH(i)  r1mach_data[(i)-1]
#define D1MACH(i)  d1mach_data[(i)-1]

static void init_mach_data(void);

/*---------------------------------------------------------------------------*/

int gpl_num_i1mach(int i)
{
  if (need_init) init_mach_data();
  if (i<5 || i>16) {
    fprintf(stderr, "*** unexpected argument: GPL_NUM_I1MACH(%d)\n", i);
    abort();
  }
  return I1MACH(i);
}

float gpl_num_r1mach(int i)
{
  if (need_init) init_mach_data();
  if (i<1 || i>5) {
    fprintf(stderr, "*** unexpected argument: GPL_NUM_R1MACH(%d)\n", i);
    abort();
  }
  return R1MACH(i);
}

double gpl_num_d1mach(int i)
{
  if (need_init) init_mach_data();
  if (i<1 || i>5) {
    fprintf(stderr, "*** unexpected argument: GPL_NUM_d1mach(%d)\n", i);
    abort();
  }
  return D1MACH(i);
}

#ifndef ASSUME_IEEE
# include <limits.h>
#endif

static void init_mach_data(void)
{
#ifdef ASSUME_IEEE
#  warning "assume IEEE"
  int j, emin, emax, t;

  /* INTEGERS
   *   Assume integers are represented in the S-digit, base-A form:
   *       SIGN ( X(S-1)*A**(S-1) + ... + X(1)*A + X(0) )
   *   where 0<=X(i)<A for i=0..S-1. */
  I1MACH(5)=         32; /* number of bits */
  I1MACH(6)=	      4; /* number of characters */
  I1MACH(7)=	      2; /* A, the base */
  I1MACH(8)=	     31; /* S, number of base-A digits per integer */
  I1MACH(9)= 2147483647; /* A**S - 1, the largest magnitude */
  
  /* FLOATING POINT NUMBERS
   *   Assume floating-point numbers are represented in the T-digit,
   *   base-B form:
   *       SIGN (B**E)*( (X(1)/B) + ... + (X(T)/B**T) )
   *   where 0<=X(i)<B for i=1..T  and 0<X(1) and EMIN<=E<=EMAX. */
  I1MACH(10)=  2; /* B, the base */
  
  /* SINGLE-PRECISION */ {
    float one= 1.0f, b= (float)I1MACH(10);
    I1MACH(11)= (t=      24); /* T, the number of base-B digits */
    I1MACH(12)= (emin= -125); /* EMIN, the smallest exponent E */
    I1MACH(13)= (emax=  128); /* EMAX, the largest exponent E */

    /* R1MACH(4) = B**(1-T), the largest relative spacing. */
    for (R1MACH(4)=one, j=1 ; j<t ; j++) R1MACH(4)/= b;

    /* R1MACH(3) = B**(-T), the smallest relative spacing. */
    R1MACH(3)= R1MACH(4)/b;

    /* R1MACH(1) = B**(EMIN-1), the smallest positive magnitude. */
    for (R1MACH(1)=one, j=1 ; j<=1-emin ; j++) R1MACH(1)/= b;
    
    /* AMAX = B**EMAX*(1 - B**(-T)), the largest magnitude. */
    for (R1MACH(2)=one-R1MACH(3), j=1 ; j<=emax ; j++) R1MACH(2)*= b;

    R1MACH(5)= log10(b); /* LOG10(B) */
  }

  /* DOUBLE-PRECISION */ {
    double one= 1.0, b= (double)I1MACH(10);
    I1MACH(14)= (t=       53); /* T, the number of base-B digits */
    I1MACH(15)= (emin= -1021); /* EMIN, the smallest exponent E */
    I1MACH(16)= (emax=  1024); /* EMAX, the largest exponent E */

    /* D1MACH(4) = B**(1-T), the largest relative spacing. */
    for (D1MACH(4)=one, j=1 ; j<t ; j++) D1MACH(4)/= b;

    /* D1MACH(3) = B**(-T), the smallest relative spacing. */
    D1MACH(3)= D1MACH(4)/b;

    /* D1MACH(1) = B**(EMIN-1), the smallest positive magnitude. */
    for (D1MACH(1)=one, j=1 ; j<=1-emin ; j++) D1MACH(1)/= b;
    
    /* AMAX = B**EMAX*(1 - B**(-T)), the largest magnitude. */
    for (D1MACH(2)=one-D1MACH(3), j=1 ; j<=emax ; j++) D1MACH(2)*= b;

    D1MACH(5)= log10(b); /* LOG10(B) */
  }

#else /* ASSUME_IEEE not defined */

  /* INTEGERS */
#ifndef INT_BIT
# ifndef CHAR_BIT
#  warning "assume 8-bit char"
#  define CHAR_BIT 8
# endif
# define INT_BIT ((CHAR_BIT)*sizeof(int))
#endif
  I1MACH(5)= INT_BIT;     /* number of bits */
  I1MACH(6)= sizeof(int); /* number of characters */
  I1MACH(7)= 2;           /* A, the base */
  I1MACH(8)= INT_BIT - 1; /* S, number of base-A digits per integer */
  I1MACH(9)= INT_MAX;     /* A**S - 1, the largest magnitude */

  /* FLOATING POINT NUMBERS */
  I1MACH(10)= gpl_num_slamch('B'); /* B, the base */
  if (I1MACH(10) != gpl_num_dlamch('B')) {
    fprintf(stderr,
	    "*** Not same base radix for float's and double's: %d/%f\n",
	    I1MACH(10), gpl_num_dlamch('B'));
    abort();
  }

  /* SINGLE-PRECISION */
  I1MACH(11)= gpl_num_slamch('N'); /* T, mantissa (number of base-B digits) */
  I1MACH(12)= gpl_num_slamch('M'); /* EMIN, smallest exponent E */
  I1MACH(13)= gpl_num_slamch('L'); /* EMAX, largest exponent E */
  R1MACH(1)=  gpl_num_slamch('U'); /* B**(EMIN-1), smallest positive magnitude. */
  R1MACH(2)=  gpl_num_slamch('O'); /* B**EMAX*(1 - B**(-T)), largest magnitude. */
  R1MACH(3)=  gpl_num_slamch('E'); /* B**(-T), smallest relative spacing. */
  R1MACH(4)=  gpl_num_slamch('P'); /* B**(1-T), largest relative spacing. */
  R1MACH(5)= log10(gpl_num_slamch('B')); /*  LOG10(B) */ 

  /* DOUBLE-PRECISION */
  I1MACH(14)= gpl_num_dlamch('N'); /* T, mantissa (number of base-B digits) */
  I1MACH(15)= gpl_num_dlamch('M'); /* EMIN, smallest exponent E */
  I1MACH(16)= gpl_num_dlamch('L'); /* EMAX, largest exponent E */
  D1MACH(1)=  gpl_num_dlamch('U'); /* B**(EMIN-1), smallest positive magnitude. */
  D1MACH(2)=  gpl_num_dlamch('O'); /* B**EMAX*(1 - B**(-T)), largest magnitude. */
  D1MACH(3)=  gpl_num_dlamch('E'); /* B**(-T), smallest relative spacing. */
  D1MACH(4)=  gpl_num_dlamch('P'); /* B**(1-T), largest relative spacing. */
  D1MACH(5)= log10(gpl_num_dlamch('B')); /*  LOG10(B) */ 

#endif /* ASSUME_IEEE */

  /* Initialization complete. */
  need_init= 0;
}

/*---------------------------------------------------------------------------*/

#undef real_t
#define real_t float
#undef ZERO
#define ZERO   0.0f
#undef ONE
#define ONE    1.0f
#undef ADD
#define ADD    gpl_num_add_f
#undef SUB
#define SUB    gpl_num_sub_f
#undef MUL
#define MUL    gpl_num_mul_f
#undef DIV
#define DIV    gpl_num_div_f
#undef EQ
#define EQ     gpl_num_eq_f
#undef NE
#define NE     gpl_num_ne_f
#undef GE
#define GE     gpl_num_ge_f
#undef GT
#define GT     gpl_num_gt_f
#undef LE
#define LE     gpl_num_le_f
#undef LT
#define LT     gpl_num_lt_f
#undef CMP
#define CMP    gpl_num_cmp_f
#undef STORE
#define STORE  gpl_num_store_f
#undef SLAMCH
#define SLAMCH gpl_num_slamch
#undef POW
#define POW    ipow_f
#undef SLAMC1
#define SLAMC1 slamc1
#undef SLAMC2
#define SLAMC2 slamc2
#undef SLAMC4
#define SLAMC4 slamc4
#undef SLAMC5
#define SLAMC5 slamc5
#include __FILE__

#undef real_t
#define real_t double
#undef ZERO
#define ZERO   0.0
#undef ONE
#define ONE    1.0
#undef ADD
#define ADD    gpl_num_add_d
#undef SUB
#define SUB    gpl_num_sub_d
#undef MUL
#define MUL    gpl_num_mul_d
#undef DIV
#define DIV    gpl_num_div_d
#undef EQ
#define EQ     gpl_num_eq_d
#undef NE
#define NE     gpl_num_ne_d
#undef GE
#define GE     gpl_num_ge_d
#undef GT
#define GT     gpl_num_gt_d
#undef LE
#define LE     gpl_num_le_d
#undef LT
#define LT     gpl_num_lt_d
#undef CMP
#define CMP    gpl_num_cmp_d
#undef STORE
#define STORE  gpl_num_store_d
#undef SLAMCH
#define SLAMCH gpl_num_dlamch
#undef POW
#define POW    ipow_d
#undef SLAMC1
#define SLAMC1 dlamc1
#undef SLAMC2
#define SLAMC2 dlamc2
#undef SLAMC4
#define SLAMC4 dlamc4
#undef SLAMC5
#define SLAMC5 dlamc5
#include __FILE__

#else /* _GPL_NUM_MACH_C defined ----------------------------------------------*/

/* Imported routines (needed to force A and B to be stored prior to doing
   the operation with A and B, for use in situations where optimizers might
   hold one of these in a register). */
extern real_t ADD PROTO((real_t a, real_t b));
extern real_t SUB PROTO((real_t a, real_t b));
extern real_t MUL PROTO((real_t a, real_t b));
extern real_t DIV PROTO((real_t a, real_t b));
extern int     EQ PROTO((real_t a, real_t b));
extern int     NE PROTO((real_t a, real_t b));
extern int     GE PROTO((real_t a, real_t b));
extern int     GT PROTO((real_t a, real_t b));
extern int     LE PROTO((real_t a, real_t b));
extern int     LT PROTO((real_t a, real_t b));
extern int    CMP PROTO((real_t a, real_t b));
extern void STORE PROTO((real_t *addr, real_t value));

#undef ASSIGN
#define ASSIGN(var,expr) STORE(&var,expr)

/* Private routines. */
static real_t    POW    PROTO((real_t x, integer_t n));
static void      SLAMC1 PROTO((integer_t *beta, integer_t *t, logical_t *rnd,
			       logical_t *ieee1));
static void      SLAMC2 PROTO((integer_t *beta, integer_t *t, logical_t *rnd,
			       real_t *eps, integer_t *emin, real_t *rmin,
			       integer_t *emax, real_t *rmax));
static integer_t SLAMC4 PROTO((real_t start, integer_t base));
static void      SLAMC5 PROTO((integer_t beta, integer_t p, integer_t emin,
			       logical_t ieee, integer_t *emax, real_t *rmax));

/* Compute X**N by successive multiplications or divisions
   (N small integer). */
#if defined(NEED_PROTOTYPES) || defined(__STDC__) || defined(__cplusplus)
static real_t POW(real_t x, integer_t n)
#else
static real_t POW(x, n)
     real_t x;
     integer_t n;
#endif
{
  integer_t i;
  real_t xn;
  if (n>0) {
    for (xn=x, i=1 ; i<n ; ++i) xn*= x;
  } else if (n<0) {
    for (xn=ONE/x, i=-1 ; i>n ; --i) xn/= x;
  } else /* n==0 */ {
    xn= x?ONE:ZERO/ZERO;
  }
  return xn;
}

/*
 *  -- LAPACK auxiliary routine (version 3.0) --
 *     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
 *     Courant Institute, Argonne National Lab, and Rice University
 *     October 31, 1992
 *
 *
 *  Purpose
 *  =======
 *
 *  SLAMCH determines single precision machine parameters.
 *
 *  Arguments
 *  =========
 *
 *  CMACH   (input) CHARACTER*1
 *          Specifies the value to be returned by SLAMCH:
 *          = 'E' or 'e',   SLAMCH := eps
 *          = 'S' or 's ,   SLAMCH := sfmin
 *          = 'B' or 'b',   SLAMCH := base
 *          = 'P' or 'p',   SLAMCH := eps*base
 *          = 'N' or 'n',   SLAMCH := t
 *          = 'R' or 'r',   SLAMCH := rnd
 *          = 'M' or 'm',   SLAMCH := emin
 *          = 'U' or 'u',   SLAMCH := rmin
 *          = 'L' or 'l',   SLAMCH := emax
 *          = 'O' or 'o',   SLAMCH := rmax
 *
 *          where
 *
 *          eps   = relative machine precision
 *          sfmin = safe minimum, such that 1/sfmin does not overflow
 *          base  = base of the machine
 *          prec  = eps*base
 *          t     = number of (base) digits in the mantissa
 *          rnd   = 1.0 when rounding occurs in addition, 0.0 otherwise
 *          emin  = minimum exponent before (gradual) underflow
 *          rmin  = underflow threshold - base**(emin-1)
 *          emax  = largest exponent before overflow
 *          rmax  = overflow threshold  - (base**emax)*(1-eps)
 *
 * =====================================================================
 */

#if defined(NEED_PROTOTYPES) || defined(__STDC__) || defined(__cplusplus)
real_t SLAMCH(char cmach)
#else
real_t SLAMCH(cmach)
     char cmach;
#endif
{
  static logical_t first= 1;
  static real_t eps, prec, sfmin, base, t, rnd, emin, emax, rmin, rmax;

  if (first) {
    logical_t lrnd;
    integer_t beta, imax, imin, it;
    SLAMC2(&beta, &it, &lrnd, &eps, &imin, &rmin, &imax, &rmax);
    base= beta;
    t= it;
    if (lrnd) {
      rnd= ONE;
      eps= POW(base, (1-it)) / 2;
    } else {
      rnd= ZERO;
      eps= POW(base, (1-it));
    }
    prec= eps*base;
    emin= imin;
    emax= imax;
    sfmin= rmin;
    if (DIV(ONE,rmax) >= sfmin) {
      /* Use 1/RMAX plus a bit, to avoid the possibility of
	 rounding causing overflow when computing 1/SFMIN. */
      sfmin= DIV(ONE+eps,rmax);
    }
    first= 0;
  }

  switch (cmach) {
  case 'e': case 'E': return eps;
  case 's': case 'S': return sfmin;
  case 'b': case 'B': return base;
  case 'p': case 'P': return prec;
  case 'n': case 'N': return t;
  case 'r': case 'R': return rnd;
  case 'm': case 'M': return emin;
  case 'u': case 'U': return rmin;
  case 'l': case 'L': return emax;
  case 'o': case 'O': return rmax;
  }
  return ZERO;
}

/*
 *  -- LAPACK auxiliary routine (version 3.0) --
 *     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
 *     Courant Institute, Argonne National Lab, and Rice University
 *     October 31, 1992
 *
 *  Purpose
 *  =======
 *
 *  SLAMC1 determines the machine parameters given by BETA, T, RND, and
 *  IEEE1.
 *
 *  Arguments
 *  =========
 *
 *  BETA    (output) INTEGER
 *          The base of the machine.
 *
 *  T       (output) INTEGER
 *          The number of ( BETA ) digits in the mantissa.
 *
 *  RND     (output) LOGICAL
 *          Specifies whether proper rounding  ( RND = .TRUE. )  or
 *          chopping  ( RND = .FALSE. )  occurs in addition. This may not
 *          be a reliable guide to the way in which the machine performs
 *          its arithmetic.
 *
 *  IEEE1   (output) LOGICAL
 *          Specifies whether rounding appears to be done in the IEEE
 *          'round to nearest' style.
 *
 *  Further Details
 *  ===============
 *
 *  The routine is based on the routine  ENVRON  by Malcolm and
 *  incorporates suggestions by Gentleman and Marovich. See
 *
 *     Malcolm M. A. (1972) Algorithms to reveal properties of
 *        floating-point arithmetic. Comms. of the ACM, 15, 949-951.
 *
 *     Gentleman W. M. and Marovich S. B. (1974) More on algorithms
 *        that reveal properties of floating point arithmetic units.
 *        Comms. of the ACM, 17, 276-277.
 *
 * =====================================================================
 */
#if defined(NEED_PROTOTYPES) || defined(__STDC__) || defined(__cplusplus)
static void SLAMC1(integer_t *beta, integer_t *t,
		   logical_t *rnd, logical_t *ieee1)
#else
static void SLAMC1(beta, t, rnd, ieee1)
     integer_t *beta, *t;
     logical_t *ieee1, *rnd;
#endif
{
  /* LBETA, LIEEE1, LT and LRND are the local values of
     BETA, IEEE1, T and RND. */
  static logical_t first=1, lieee1, lrnd;
  static integer_t lbeta, lt;

  if (first) {
    /* Throughout this routine we use the function external routines
       ADD/SUB/... to ensure that relevant values are stored and not held
       in registers, or are not affected by optimizers. */
    real_t a, b, c, one=1, t1, t2;
    
    /* Compute  A = 2.0**M  with the  smallest positive integer M such
       that (numerically) (A + 1) == A. */
    ASSIGN(a,one);
    while ((ASSIGN(c,SUB(ADD(a,one),a)), c) == one) {
#ifdef PARANOIAC
      ASSIGN(a,MUL(a,2));
#else
      a*= 2;
#endif
    }

    /* Now compute  B=2.0**M  with the smallest positive integer M
       such that (numerically) (A+B) > A. */
    /*ASSIGN(b,one);*/
    ASSIGN(b,one);
    while ((ASSIGN(c,ADD(a,b)), c) <= a) {
#ifdef PARANOIAC
      ASSIGN(b,MUL(b,2));
#else
      b*= 2;
#endif
    }

    /* Now compute the base.  A and C  are neighbouring floating point
       numbers  in the  interval  (BETA**T, BETA**(T+1))  and so
       their difference is BETA. Adding 0.25 to C is to ensure that it
       is truncated to beta and not (BETA - 1). */
    lbeta = ADD(SUB(c,a),one/4);

    /* Now determine whether rounding or chopping occurs, by adding a
       bit less than BETA/2 and a bit more than BETA/2 to A. */
    ASSIGN(b, (real_t)lbeta);
#ifdef PARANOIAC
    ASSIGN(t1, ADD(a,SUB(b/2,b/100)));
    ASSIGN(t2, ADD(a,ADD(b/2,b/100)));
#else
    ASSIGN(t1, ADD(a,(b/2)-(b/100)));
    ASSIGN(t2, ADD(a,(b/2)+(b/100)));
#endif
    lrnd= t1==a && t2!=a;
    
    /* Try and decide whether rounding is done in the IEEE 'round to
       nearest' style. B/2 is half a unit in the last place of the two
       numbers A and C. Furthermore, A is even, i.e. has last bit zero,
       and C is odd. Thus adding B/2 to A should not change A, but
       adding B/2 to C should change C. */
    ASSIGN(t1, ADD(a,b/2));
    ASSIGN(t2, ADD(c,b/2));
    lieee1= lrnd && t1==a && t2>c;

    /* Now find  the  mantissa, T.  It should  be the  integer part of
       LOG to the base BETA of A,  however it is safer to determine  T
       by powering.  So we find T as the smallest positive integer for
       which (numerically) (BETA**T + 1.0) - BETA**T  == 1.0. */
    ASSIGN(a,one);
    while ((ASSIGN(c,SUB(ADD(a,one),a)), c) == one) {
      ++lt;
      a*=lbeta;
      /*ASSIGN(a,MUL(a,lbeta));*/
    }

    first= 0; /* OK all static variables updated. */
  }

  *beta= lbeta;
  *t= lt;
  *rnd= lrnd;
  *ieee1= lieee1;
}

/*
 *  -- LAPACK auxiliary routine (version 3.0) --
 *     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
 *     Courant Institute, Argonne National Lab, and Rice University
 *     October 31, 1992
 *
 *  Purpose
 *  =======
 *
 *  SLAMC2 determines the machine parameters specified in its argument
 *  list.
 *
 *  Arguments
 *  =========
 *
 *  BETA    (output) INTEGER
 *          The base of the machine.
 *
 *  T       (output) INTEGER
 *          The number of ( BETA ) digits in the mantissa.
 *
 *  RND     (output) LOGICAL
 *          Specifies whether proper rounding  ( RND = .TRUE. )  or
 *          chopping  ( RND = .FALSE. )  occurs in addition. This may not
 *          be a reliable guide to the way in which the machine performs
 *          its arithmetic.
 *
 *  EPS     (output) REAL
 *          The smallest positive number such that
 *
 *             fl( 1.0 - EPS ) .LT. 1.0,
 *
 *          where fl denotes the computed value.
 *
 *  EMIN    (output) INTEGER
 *          The minimum exponent before (gradual) underflow occurs.
 *
 *  RMIN    (output) REAL
 *          The smallest normalized number for the machine, given by
 *          BASE**( EMIN - 1 ), where  BASE  is the floating point value
 *          of BETA.
 *
 *  EMAX    (output) INTEGER
 *          The maximum exponent before overflow occurs.
 *
 *  RMAX    (output) REAL
 *          The largest positive number for the machine, given by
 *          BASE**EMAX * ( 1 - EPS ), where  BASE  is the floating point
 *          value of BETA.
 *
 *  Further Details
 *  ===============
 *
 *  The computation of  EPS  is based on a routine PARANOIA by
 *  W. Kahan of the University of California at Berkeley.
 *
 * =====================================================================
 */

#if defined(NEED_PROTOTYPES) || defined(__STDC__) || defined(__cplusplus)
static void SLAMC2(integer_t *beta, integer_t *t, logical_t *rnd, real_t *eps,
		   integer_t *emin, real_t *rmin,
		   integer_t *emax, real_t *rmax)
#else
static void SLAMC2(beta, t, rnd, eps, emin, rmin, emax, rmax)
     logical_t *rnd;
     integer_t *beta, *t, *emin, *emax;
     real_t    *eps, *rmax, *rmin;
#endif
{
  /* LBETA, LT, LRND, LEPS, LEMIN and LRMIN  are the local values of
     BETA, T, RND, EPS, EMIN and RMIN. */
  static logical_t first=1, iwarn=0, lrnd;
  static integer_t lbeta, lemax, lemin, lt;
  static real_t lrmax, lrmin, leps;

  if (first) {
    /* Throughout this routine  we use external routines ADD/SUB/...
       to ensure that relevant values are stored  and not held in
       registers,  or are not affected by optimizers. */
    logical_t   ieee, lieee1;
    integer_t  gnmin, gpmin, i, ngnmin, ngpmin;
    real_t a, b, c, half, rbase, sixth, third;
    real_t zero=0, one=1, two=2;

    /* SLAMC1 returns the parameters  LBETA, LT, LRND and LIEEE1. */
    SLAMC1(&lbeta, &lt, &lrnd, &lieee1);
#if 0
    fprintf(stderr, "          base:  BETA=%d\n", (int)lbeta);
    fprintf(stderr, "      mantissa:     T=%d\n", (int)lt);
    fprintf(stderr, "      rounding:   RND=%d\n", (int)lrnd);
    fprintf(stderr, "IEEE compliant: IEEE1=%d\n", (int)lieee1);
#endif

    /* Start to find EPS. */
    b= (real_t)lbeta;
    a= POW(b,-lt);
    leps= a;

    /* Try some tricks to see whether or not this is the correct EPS. */
    b= two / 3;
    half= one / 2;
    sixth= SUB(b, half);
    third= ADD(sixth, sixth);
    b= SUB(third, half);
    b= ADD(b, sixth);
    if (b<zero) b= -b;
    if (b<leps) b= leps;
    leps= 1;
    while ((leps>b) && (b>zero)) {
      ASSIGN(leps, b);
      ASSIGN(c, SUB(half, ADD(half*leps, POW(two,5)*POW(leps,2))));
      ASSIGN(b, ADD(half, c));
      ASSIGN(b, SUB(half, b));
      ASSIGN(b, ADD(half, c));
    }
    if (a<leps) leps= a;

    /* Computation of EPS complete. */

    /* Now find  EMIN.  Let A = + or - 1, and + or - (1 + BASE**(-3)).
       Keep dividing  A by BETA until (gradual) underflow occurs. This
       is detected when we cannot recover the previous A. */
    rbase= one / lbeta;
    a= ADD(one, POW(rbase, 3));
    ngpmin= SLAMC4( one, lbeta);
    ngnmin= SLAMC4(-one, lbeta);
    gpmin=  SLAMC4(   a, lbeta);
    gnmin=  SLAMC4(  -a, lbeta);
    ieee= 0;
    if (ngpmin==ngnmin && gpmin==gnmin) {
      if (ngpmin==gpmin) {
	/* Non twos-complement machines, no gradual underflow;
	   e.g., VAX. */
	lemin= ngpmin;
      } else if (gpmin-ngpmin==3) {
	/* Non twos-complement machines, with gradual underflow;
	   e.g., IEEE standard followers. */
	lemin= ngpmin - 1 + lt;
	ieee= 1;
      } else {
	/* A guess; no known machine. */
	lemin= MIN(ngpmin, gpmin);
	iwarn= 1;
      }
    } else if (ngpmin==gpmin && ngnmin==gnmin) {
      if (ABS_DIF(ngpmin,ngnmin)==1) {
	/* Twos-complement machines, no gradual underflow; e.g., CYBER 205. */
	lemin= MAX(ngpmin, ngnmin);
      } else {
	/* A guess; no known machine. */
	lemin= MIN(ngpmin, ngnmin);
	iwarn= 1;
      }
    } else if (gpmin==gnmin && ABS_DIF(ngpmin,ngnmin)==1) {
      if (gpmin-MIN(ngpmin, ngnmin)==3) {
	/* Twos-complement machines with gradual underflow;
	   no known machine. */
	lemin= MAX(ngpmin, ngnmin) - 1 + lt;
      } else {
	/* A guess; no known machine. */
	lemin= MIN(ngpmin, ngnmin);
	iwarn= 1;
      }
    } else {
      /* A guess; no known machine. */
      lemin= MIN(MIN(ngpmin, ngnmin), MIN(gpmin, gnmin));
      iwarn= 1;
    }

    /* Assume IEEE arithmetic if we found denormalised  numbers above,
       or if arithmetic seems to round in the  IEEE style,  determined
       in routine SLAMC1. A true IEEE machine should have both  things
       true; however, faulty machines may have one or the other. */
    ieee= ieee || lieee1;

    /* Compute  RMIN by successive division by  BETA. We could compute
       RMIN as BASE**(EMIN-1),  but some machines underflow during
       this computation. */
    for (lrmin=one, i=1-lemin ; i>=1 ; --i) lrmin= MUL(lrmin, rbase);

    /* Finally, call SLAMC5 to compute EMAX and RMAX. */
    SLAMC5(lbeta, lt, lemin, ieee, &lemax, &lrmax);

    /* OK all static variables updated. */
    first= 0;

    /* Comment out this if block if EMIN is ok*/
    if (iwarn) {
      fprintf(stderr, "***WARNING (%s: %d) EMIN=%ld may be incorrect\n",
	      __FILE__, __LINE__, (long)lemin);
      first= 1;
    }
  }

  *beta= lbeta;
  *t= lt;
  *rnd= lrnd;
  *eps= leps;
  *emin= lemin;
  *rmin= lrmin;
  *emax= lemax;
  *rmax= lrmax;
}

/*
 *  -- LAPACK auxiliary routine (version 3.0) --
 *     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
 *     Courant Institute, Argonne National Lab, and Rice University
 *     October 31, 1992
 *
 *  Purpose
 *  =======
 *
 *  SLAMC4 is a service routine for SLAMC2.
 *
 *  Arguments
 *  =========
 *
 *  EMIN    (output) EMIN
 *          The minimum exponent before (gradual) underflow, computed by
 *          setting A = START and dividing by BASE until the previous A
 *          can not be recovered.
 *
 *  START   (input) REAL
 *          The starting point for determining EMIN.
 *
 *  BASE    (input) INTEGER
 *          The base of the machine.
 *
 * =====================================================================
 */
#if defined(NEED_PROTOTYPES) || defined(__STDC__) || defined(__cplusplus)
static integer_t SLAMC4(real_t start, integer_t base)
#else
static integer_t SLAMC4(start, base)
     real_t start;
     integer_t base;
#endif
{
  integer_t  i, emin;
  real_t a, ap, b1, b2, rbase, zero=0;

  ASSIGN(rbase, (real_t)1/(real_t)base);
  ASSIGN(a, start);
  emin= 1;
  for (;;) {
    ASSIGN(b1, DIV(a,  base));
    ASSIGN(ap, MUL(b1, base));
    if (ap!=a) break;
    for (ASSIGN(ap,zero), i=1 ; i<=base ; ++i) ASSIGN(ap, ap+b1);
    if (ap!=a) break;
    ASSIGN(b2, MUL(a,  rbase));
    ASSIGN(ap, DIV(b2, rbase));
    if (ap!=a) break;
    for (ASSIGN(ap,zero), i=1 ; i<=base ; ++i) ASSIGN(ap, ap+b2);
    if (ap!=a) break;
    --emin;
    ASSIGN(a,  b1); /* A := START * BASE**(EMIN-1) */
  }
  return emin;
}

/*
 *  -- LAPACK auxiliary routine (version 3.0) --
 *     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
 *     Courant Institute, Argonne National Lab, and Rice University
 *     October 31, 1992
 *
 *  Purpose
 *  =======
 *
 *  SLAMC5 attempts to compute RMAX, the largest machine floating-point
 *  number, without overflow.  It assumes that EMAX + abs(EMIN) sum
 *  approximately to a power of 2.  It will fail on machines where this
 *  assumption does not hold, for example, the Cyber 205 (EMIN = -28625,
 *  EMAX = 28718).  It will also fail if the value supplied for EMIN is
 *  too large (i.e. too close to zero), probably with overflow.
 *
 *  Arguments
 *  =========
 *
 *  BETA    (input) INTEGER
 *          The base of floating-point arithmetic.
 *
 *  T       (input) INTEGER
 *          The number of base BETA digits in the mantissa of a
 *          floating-point value.
 *
 *  EMIN    (input) INTEGER
 *          The minimum exponent before (gradual) underflow.
 *
 *  IEEE    (input) LOGICAL
 *          A logical flag specifying whether or not the arithmetic
 *          system is thought to comply with the IEEE standard.
 *
 *  EMAX    (output) INTEGER
 *          The largest exponent before overflow
 *
 *  RMAX    (output) REAL
 *          The largest machine floating-point number.
 *
 * =====================================================================
 */
#if defined(NEED_PROTOTYPES) || defined(__STDC__) || defined(__cplusplus)
static void SLAMC5(integer_t beta, integer_t t, integer_t emin,
		   logical_t ieee, integer_t *emax, real_t *rmax)
#else
static void SLAMC5(beta, t, emin, ieee, emax, rmax)
     logical_t  ieee;
     integer_t  beta, emin, t, *emax;
     real_t    *rmax;
#endif
{
  real_t zero=0, one=1;
  integer_t exbits, expsum, i, lexp, nbits, uexp;
  real_t oldy, rbase, y, z;

  /* First compute LEXP and UEXP, two powers of 2 that bound abs(EMIN).
     We then assume that EMAX + abs(EMIN) will sum approximately to the
     bound that is closest to abs(EMIN). (EMAX is the exponent of the
     required number RMAX). */
  lexp= 1;
  exbits= 1;
  while (2*lexp <= (-emin)) {
    lexp*= 2;
    ++exbits;
  }
  if (lexp == -emin) {
    uexp= lexp;
  } else {
    uexp= 2*lexp;
    ++exbits;
  }

  /* Now -LEXP is less than or equal to EMIN, and -UEXP is greater
     than or equal to EMIN. EXBITS is the number of bits needed to
     store the exponent. */
  expsum= 2 * ((uexp+emin > -lexp-emin) ? lexp : uexp);

  /* EXPSUM is the exponent range, approximately equal to  EMAX-EMIN+1 . */
  (*emax)= expsum + emin - 1;

  /* NBITS is the total number of bits needed to store a
     floating-point number. */
  nbits= 1 + exbits + t;
  if ((nbits%2)==1 && beta==2) {
    /* Either there are an odd number of bits used to store a floating-point
       number, which is unlikely, or some bits are not used in the
       representation of numbers, which is possible, (e.g. Cray machines) or
       the mantissa has an implicit bit, (e.g. IEEE machines, Dec Vax
       machines), which is perhaps the most likely. We have to assume the last
       alternative.  If this is true, then we need to reduce EMAX by one
       because there must be some way of representing zero in an implicit-bit
       system. On machines like Cray, we are reducing EMAX by one
       unnecessarily. */
    --(*emax);
  }

  if (ieee) {
     /* Assume we are on an IEEE machine which reserves one exponent
        for infinity and NaN. */
    --(*emax);
  }

  /* Now create RMAX, the largest machine number, which should be equal
     to (1.0 - BETA**(-T)) * BETA**EMAX . */

  /* First compute 1.0 - BETA**(-T), being careful that the result is
     less than 1.0 . */
  rbase= one / beta;
  z= beta - one;
  oldy= y= zero; /* initialize OLDY to avoid compiler warning */
  for (i=1 ; i<=t ; ++i) {
    z*= rbase;
    if (y<one) oldy= y;
    y= ADD(y, z);
  }
  if (y>=one) y= oldy;

  /* Now multiply by BETA**EMAX to get RMAX. */
  for (i=1 ; i<=(*emax) ; ++i) y= MUL(y, beta);
  *rmax= y;
}
#endif /* _GPL_NUM_MACH_C -----------------------------------------------------*/
