/*
 * gpl_num_qr.c --
 *
 *	Least solution of an overdetermined linear system by QR factorization.
 *	Original FORTRAN routines are from SLATEC and LINPACK.
 *
 * Copyright (c) 1999 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id$
 *	$Log$
 */

#ifndef _GPL_NUM_QR_C
#define _GPL_NUM_QR_C 1

#include <math.h>

#include "gpl_num.h"

/* Compute Givens rotation (overflows and division by zero are avoided by
   using either tangent or arctangent of the angle of Givens rotation).
   X and Y should not be expressions (i.e. otherwise they are evaluated
   several times) GSIN and GCOS must be suitable LHS for assignation. */
#define GIVENS(GCOS,GSIN,X,Y)						      \
   if (ABS(X) >= ABS(Y)) {						      \
     real_t __gtan= (Y)/(X);						      \
     GSIN= (GCOS= ONE/SQRT(ONE+__gtan*__gtan)) * __gtan;		      \
   } else {								      \
     real_t __gcotan= (X)/(Y);						      \
     GCOS= (GSIN= ONE/SQRT(ONE+__cotan*__cotan)) * __gcotan;		      \
   }


#define GPL_NUM_SINGLE_PRECISION
#include "gpl_num_priv.h"
#include __FILE__
#undef GPL_NUM_SINGLE_PRECISION

#define GPL_NUM_DOUBLE_PRECISION
#include "gpl_num_priv.h"
#include __FILE__
#undef GPL_NUM_DOUBLE_PRECISION

#else /* _GPL_NUM_QR_C defined ------------------------------------------------*/

/* HTR1(N, X, Y) is the same as:
     A= -SDOT(N, X, 1, Y, 1) / X(1);
     SAXPY(N, A, X, 1, Y, 1);
   HTR2 is the same as HTR1 but with a avlue for X(1) specified
   separately. */
#undef HTR1
#undef HTR2
#ifdef GPL_NUM_SINGLE_PRECISION
# define HTR1 htr1_f
# define HTR2 htr2_f
#else
# define HTR1 htr1_d
# define HTR2 htr2_d
#endif
static void HTR1(integer_t n, const real_t x[], real_t y[]);
static void HTR1(integer_t n, const real_t x[], real_t y[])
{
  integer_t i;
  real_t a;
  if (n>0) {
    a= x[0]*y[0];
    for (i=1 ; i<n ; ++i) a+= x[i]*y[i];
    y[0]-= a;
    a/= x[0];
    for (i=1 ; i<n ; ++i) y[i]-= a*x[i];
  }
}

static void HTR2(integer_t n, real_t x0, const real_t x[], real_t y[]);
static void HTR2(integer_t n, real_t x0, const real_t x[], real_t y[])
{
  integer_t i;
  real_t a;
  if (n>0) {
    a= x0*y[0];
    for (i=1 ; i<n ; ++i) a+= x[i]*y[i];
    y[0]-= a;
    a/= x0;
    for (i=1 ; i<n ; ++i) y[i]-= a*x[i];
  }
}
/*---------------------------------------------------------------------------*/

void SQRDC(real_t *x, integer_t ldx, integer_t n, integer_t p, 
	   real_t *qraux, integer_t *jpvt, real_t *work, integer_t job)
{
  /* Local variables */
  integer_t j, jp, l, lup, pl, pu, maxj;
  real_t t, tt, nrmxl, maxnrm;

  /* Parameter adjustments to mimics FORTRAN indexing. */
#define x_(a_1,a_2) x[(a_2)*ldx + a_1]
  x-= ldx+1;
  --qraux;
  --jpvt;
  --work;
    
  /* Function Body */
  pl= 1;
  pu= 0;
  if (job) {
    /* Pivoting has been requested.
       Rearrange the columns according to JPVT. */
    for (j=1 ; j<=p ; ++j) {
      if (jpvt[j] > 0) {
	/* J-th column of X is an initial column: swap. */
	if (j!=pl) SSWAP(n, &x_(1,pl), 1, &x_(1,j), 1);
	jpvt[j]= jpvt[pl];
	jpvt[pl]= j;
	++pl;
      } else {
	/* J-th column of X is free/final. */
	jpvt[j]= (jpvt[j] < 0) ? -j : j;
      }
    }
    pu= p;
    for (j=p ; j>=1 ; --j) {
      if (jpvt[j] < 0) {
	jpvt[j]= -jpvt[j];
	if (j != pu) {
	  SSWAP(n, &x_(1,pu), 1, &x_(1,j), 1);
	  jp= jpvt[pu];
	  jpvt[pu]= jpvt[j];
	  jpvt[j]= jp;
	}
	--pu;
      }
    }
  }

  /* Compute the norms of the free columns. */
  for (j=pl ; j<=pu ; ++j) {
    qraux[j]= SNRM2(n, &x_(1,j), 1);
    work[j]= qraux[j];
  }

  /* Perform the Householder reduction of X. */
  lup= MIN(n,p);
  for (l=1 ; l<=lup ; ++l) {
    if (l>=pl && l<pu) {
      /* Locate the column of largest norm and
	 bring it into the pivot position. */
      maxnrm= qraux[l];
      maxj= l;
      for (j=l+1 ; j<=pu ; ++j) {
	if (qraux[j]>maxnrm) {
	  maxnrm= qraux[j];
	  maxj= j;
	}
      }
      if (maxj != l) {
	SSWAP(n, &x_(1,l), 1, &x_(1,maxj), 1);
	qraux[maxj]= qraux[l];
	work[maxj]= work[l];
	jp= jpvt[maxj];
	jpvt[maxj]= jpvt[l];
	jpvt[l]= jp;
      }
    }
    qraux[l]= ZERO;
    if (l == n) continue;

    /* Compute the Householder transformation for column L. */
    nrmxl= SNRM2(n-l+1, &x_(l,l), 1);
    if (! nrmxl) continue;
    if (x_(l,l)/* != ZERO */) nrmxl= SIGN(nrmxl, x_(l,l));
    SSCAL(n-l+1, ONE/nrmxl, &x_(l,l), 1);
    ++x_(l,l);
    
    /* Apply the transformation to the remaining
       columns, updating the norms. */
    for (j=l+1 ; j<=p ; ++j) {
#ifdef HTR1
      HTR1(n-l+1, &x_(l,l), &x_(l,j));
#else
      t= -SDOT(n-l+1, &x_(l,l), 1, &x_(l,j), 1) / x_(l,l);
      SAXPY(n-l+1, t, &x_(l,l), 1, &x_(l,j), 1);
#endif
      if (j<pl || j>pu || !qraux[j]) continue;
      t= x_(l,j)/qraux[j];
      if ((t= (ONE-t) * (ONE+t)) < ZERO) {
	t= ZERO;
	tt= ONE;
      } else {
	tt= qraux[j]/work[j];
#ifdef KN_SINGLE_PRECISION
	tt= ONE + 0.05F * t * (tt*tt);
#else
	tt= ONE + 0.05  * t * (tt*tt);
#endif
      }
      if (tt == ONE) {
	qraux[j]= SNRM2(n-l, &x_(l+1,j), 1);
	work[j]= qraux[j];
      } else {
	qraux[j]*= SQRT(t);
      }
    }
    
    /* Save the transformation. */    
    qraux[l]= x_(l,l);
    x_(l,l)= -nrmxl;
  }
}
#undef x_

/*---------------------------------------------------------------------------*/

/* DEPENDENCIES
     B   requires QTY
     XB  requires QTY
     RSD requires QTY
     
     in that case, if QTY is NULL, *INFO is set to -1 if one of the arrays
     B, XB, or RSD is non-NULL and QTY is NULL.

     outputs are: B, XB, QY, QTY, RSB
     inputs are: X, QRAUX, Y
 */

void SQRSL(const real_t *x, integer_t ldx, integer_t n, integer_t k,
	   const real_t *qraux, const real_t *y, real_t *qy, real_t *qty,
	   real_t *b, real_t *rsd, real_t *xb, integer_t *info)
{
  /* Local variables */
  int /*logical*/ cqty, cb, cr, cxb, cqy;
  integer_t i, j, ju;
#ifndef HTR2
  real_t t, temp;
#endif

  /* Parameter adjustments to mimics FORTRAN indexing. */
#define x_(a_1,a_2) x[(a_2)*ldx + a_1]
  x-= ldx+1;
  --qraux;
  --y;

  /* Determine what is to be computed. */
  if ((cqty= qty!=NULL)) --qty;
  if ((cqy=   qy!=NULL)) --qy;
  if ((cb=     b!=NULL)) --b;
  if ((cr=   rsd!=NULL)) --rsd;
  if ((cxb=   xb!=NULL)) --xb;
  if (!cqty && (cb || cr || cxb)) {
    *info= -1;
    return;
  }

  /* Initialization. */
  *info= 0;
  ju= MIN(k,n-1);

  /* Special action when N=1. */
  if (ju == 0) {
    if (cqy) qy[1]= y[1];
    if (cqty) qty[1]= y[1];
    if (cxb) xb[1]= y[1];
    if (cb) {
      if (! x_(1,1)) *info= 1;
      else b[1]= y[1] / x_(1,1);
    }
    if (cr) rsd[1]= ZERO;
    return;
  }

  /* Set up to compute QY or QTY.  (Note: order of copy/zero
     operations is important since some arrays may be identical).*/
  if (cqy && qy!=y) memcpy(&qy[1], &y[1], (n)*sizeof(y[1]));
  if (cqty && qty!=y) memcpy(&qty[1], &y[1], (n)*sizeof(y[1]));
  if (cqy) {
    /* Compute QY. */
    for (j=ju ; j>=1 ; --j) {
      if (qraux[j]/* != ZERO */) {
#ifdef HTR2
	HTR2(n-j+1, qraux[j], &x_(j,j), &qy[j]);
#else
	temp= x_(j,j);
	x_(j,j)= qraux[j];
	t= -SDOT(n-j+1, &x_(j,j), 1, &qy[j], 1) / x_(j,j);
	SAXPY(n-j+1, t, &x_(j,j), 1, &qy[j], 1);
	x_(j,j)= temp;
#endif
      }
    }
  }
  if (cqty) {
    /* Compute TRANS(Q)*Y. */
    for (j=1 ; j<=ju ; ++j) {
      if (qraux[j]/* != ZERO */) {
#ifdef HTR2
	HTR2(n-j+1, qraux[j], &x_(j,j), &qty[j]);
#else
	temp= x_(j,j);
	x_(j,j)= qraux[j];
	t= -SDOT(n-j+1, &x_(j,j), 1, &qty[j], 1) / x_(j,j);
	SAXPY(n-j+1, t, &x_(j,j), 1, &qty[j], 1);
	x_(j,j)= temp;
#endif
      }
    }
  }

  /* Set up to compute B, RSD, or XB.  (Note: order of copy/zero
     operations is important since some arrays may be identical). */
  if (cb && b!=qty) memcpy(&b[1], &qty[1], (k)*sizeof(qty[1]));
  if (cxb && xb!=qty) memcpy(&xb[1], &qty[1], (k)*sizeof(qty[1]));
  if (cr && n>k && rsd!=qty) memcpy(&rsd[k+1], &qty[k+1],
				    (n-k)*sizeof(qty[k+1]));
  if (cxb && k<n) for (i=k+1 ; i<=n ; ++i) xb[i]= ZERO;
  if (cr) for (i=1 ; i<=k ; ++i) rsd[i]= ZERO;
  
  if (cb) {
    /* Compute B. */
    for (j=k ; j>=1 ; --j) {
      if (! x_(j,j)) {
	*info = j;
	break;
      }
      b[j]/= x_(j,j);
      if (j>1) SAXPY(j-1, -b[j], &x_(1,j), 1, &b[1], 1);
    }
  }

  if (cr || cxb) {
    /* Compute RSD or XB as required. */
    for (j=ju ; j>=1 ; --j) {
      if (! qraux[j]) continue;
#ifdef HTR2
      if (cr)  HTR2(n-j+1, qraux[j], &x_(j,j), &rsd[j]);
      if (cxb) HTR2(n-j+1, qraux[j], &x_(j,j), &xb[j]);
#else
      temp= x_(j,j);
      x_(j,j)= qraux[j];
      if (cr) {
	t= -SDOT(n-j+1, &x_(j,j), 1, &rsd[j], 1) / x_(j,j);
	SAXPY(n-j+1, t, &x_(j,j), 1, &rsd[j], 1);
      }
      if (cxb) {
	t= -SDOT(n-j+1, &x_(j,j), 1, &xb[j], 1) / x_(j,j);
	SAXPY(n-j+1, t, &x_(j,j), 1, &xb[j], 1);
      }
      x_(j,j)= temp;
#endif
    }
  }
}
#undef x_

/*---------------------------------------------------------------------------*/

void RWUPDT(integer_t n, real_t r[], integer_t ldr, const real_t w[], 
	    real_t b[], real_t *alpha, real_t gcos[], real_t gsin[])
{
  real_t temp, rowj;
  integer_t i, j;

#define r_(a1,a2) r[(a2)*ldr + (a1)]

  for (j=0 ; j<n ; ++j) {
    rowj= w[j];

    /* Apply the previous transformations to R(I,J), I=0..J-1, and to W(J). */
    for (i=0 ; i<j ; ++i) {
      temp= gcos[i] * r_(i,j) + gsin[i] * rowj;
      rowj= gcos[i] * rowj    - gsin[i] * r_(i,j);
      r_(i,j)= temp;
    }

    /* Determine a Givens rotation which eliminates W(J) (overflows and
       division by zero are avoided by using either tangent or arctangent of
       the angle of Givens rotation). */
    if (rowj) {
      real_t gsinj, gcosj;
      temp= r_(j,j);
      if (ABS(temp) >= ABS(rowj)) {
	temp= rowj / temp; /* TEMP= tan(ANGLE) */
	gsinj= (gcosj= ONE / SQRT(ONE + temp*temp)) * temp;
      } else {
	temp/= rowj; /* TEMP= cotan(ANGLE) */
	gcosj= (gsinj= ONE / SQRT(ONE + temp*temp)) * temp;
      }
      /* Apply the current transformation to R(J,J), B(J), and ALPHA. */
      r_(j,j)=  gcosj * r_(j,j)      + gsinj * rowj;
      b[j]=     gcosj * (temp= b[j]) + gsinj * (*alpha);
      (*alpha)= gcosj * (*alpha)     - gsinj * temp;
      gsin[j]= gsinj;
      gcos[j]= gcosj;
    } else {
      /* No rotation. */
      gcos[j]= ONE;
      gsin[j]= ZERO;
    }
  }
}
#undef r_

/*---------------------------------------------------------------------------*/

void QRFAC(integer_t m, integer_t n, real_t a[], integer_t lda, 
	   integer_t pivot[], real_t sigma[], real_t acnorm[], real_t wa[])
{
#define a_(a1,a2) a[(a2)*lda + a1]
  integer_t kmax, i, j, jp1, k, minmn;
  real_t temp, small, ajnorm;

  /* Threshold to avoid division by a too small pivot. */
  small= 20 * R1MACH(4); /* R1MACH(4) is machine precision EPSMACH */

  /* Compute the initial column norms and initialize several arrays. */
  for (j=0 ; j<n ; ++j) {
    wa[j]= (sigma[j]= (acnorm[j]= ENORM(m, &a_(0,j))));
  }
  if (pivot) {
    for (j=0 ; j<n ; ++j) {
      pivot[j]= j+1;  /* ### FIX INDEX ### */
    }
  }

  /* Reduce A to R with Householder transformations. */
  minmn= MIN(m,n);
  for (j=0 ; j<minmn ; ++j) {
    if (pivot) {
      /* Bring the column of largest norm into the pivot position. */
      for (kmax=k=j ; k<n ; ++k) {
	if (sigma[k] > sigma[kmax]) {
	  kmax= k;
	}
      }
      if (kmax != j) {
	for (i=0 ; i<m ; ++i) {
	  SWAP3(temp, a_(i,j), a_(i,kmax));
	}
	sigma[kmax]= sigma[j];
	wa[kmax]= wa[j];
	SWAP3(k, pivot[j], pivot[kmax]);
      }
    }

    /* Compute the Householder transformation to reduce the J-th
       column of A to a multiple of the J-th unit vector. */
    if (! (ajnorm= ENORM(m-j, &a_(j,j)))) {
      sigma[j]= ZERO;
      continue;
    }
    if ((temp= a_(j,j)) < ZERO) ajnorm= -ajnorm;
    a_(j,j)= ONE + temp/ajnorm;
    for (i=j+1 ; i<m ; ++i) a_(i,j)/= ajnorm;

    /* Apply the transformation to the remaining columns
       and update the norms. */
    jp1= j+1;
    for (k=jp1 ; k<n ; ++k) {
      temp= ZERO;
      for (i=j ; i<m ; ++i) temp+= a_(i,j) * a_(i,k);
      temp/= a_(j,j);
      for (i=j ; i<m ; ++i) a_(i,k)-= temp * a_(i,j);
      if (pivot && sigma[k]) {
	temp= a_(j,k) / sigma[k];
	temp= (ONE-temp) * (ONE+temp);
	temp= (temp <= ZERO) ? ZERO : (sigma[k]*= SQRT(temp)) / wa[k];
	if (temp*temp <= small) {
	  wa[k]= (sigma[k]= ENORM(m-jp1, &a_(jp1,k)));
	}
      }
    }
    sigma[j]= -ajnorm;
  }
}
#undef a_
/*---------------------------------------------------------------------------*/

void QRSOLV(integer_t n, real_t r[], integer_t ldr, const integer_t pivot[],
	    real_t diag[], const real_t qtb[], real_t x[], real_t sigma[],
	    real_t wa[])
{
#define r_(a1,a2) r[(a2)*ldr + a1]
  real_t temp, sigk, qtbpj, gsin, gcos;
  integer_t i, j, k, l, nsing;

  /* Prepare for computation of S: strict upper triangle of R is copied
     into strict lower triangle of R and diagonal of R is saved into X. */
  for (j=0 ; j<n ; ++j) {
    for (i=j ; i<n ; ++i) r_(i,j)= r_(j,i);
    x[j]= r_(j,j);
    wa[j]= qtb[j];
  }

  /* Eliminate the diagonal matrix D using a Givens rotation. */
  for (j=0 ; j<n ; ++j) {
    /* Prepare the row of D to be eliminated, locating the diagonal
       element using P from the QR factorization. */
    l= pivot ? (pivot[j]-1) : j; /* ### FIX INDEX ### */
    if (diag[l]) {
      sigma[j]= diag[l];
      for (k=j+1 ; k<n ; ++k) sigma[k]= ZERO;

      /* The transformations to eliminate the row of D modify
	 only a single element of (Q TRANSPOSE)*B beyond the first
	 N, which is initially zero. */
      qtbpj= ZERO;
      for (k=j ; k<n ; ++k) {
	/* Determine a Givens rotation which eliminates the appropriate
	   element in the current row of D (overflows and division by zero
	   are avoided by using either tangent or arctangent of the angle
	   of Givens rotation). */
	if (! (sigk= sigma[k])) continue;
	temp= r_(k,k);
	if (ABS(temp) > ABS(sigk)) {
	  temp= sigk / temp;  /* tan(ANGLE) */
	  gsin= (gcos= ONE / SQRT(ONE + temp*temp)) * temp;
	} else {
	  temp/= sigk;  /* cotan(ANGLE) */
	  gcos= (gsin= ONE / SQRT(ONE + temp*temp)) * temp;
	}
	
	/* Compute the modified diagonal element of R and
	   the modified element of ((Q TRANSPOSE)*B,0). */
	r_(k,k)= gcos * r_(k,k)       + gsin * sigk;
	wa[k]=   gcos * (temp= wa[k]) + gsin * qtbpj;
	qtbpj=   gcos * qtbpj         - gsin * temp;
	
	/* Accumulate the transformation in the row of S. */
	for (i=k+1 ; i<n ; ++i) {
	  r_(i,k)=  gcos * (temp= r_(i,k))  + gsin * sigma[i];
	  sigma[i]= gcos * sigma[i]         - gsin * temp;
	}
      }
    }

    /* Store the diagonal element of S and restore
       the corresponding diagonal element of R. */
    sigma[j]= r_(j,j);
    r_(j,j)= x[j];
  }

  /* Solve the triangular system for Z. If the system is
     singular, then obtain a least squares solution. */
  nsing= n;
  for (j=0 ; j<n ; ++j) {
    if (!sigma[j] && nsing==n) nsing= j;
    if (nsing<n) wa[j]= ZERO;
  }
  for (j=nsing-1 ; j>=0 ; --j) {
    for (temp=ZERO, i=j+1 ; i<nsing ; ++i) temp+= r_(i,j) * wa[i];
    wa[j]= (wa[j] - temp) / sigma[j];
  }

  /* Permute the components of Z back to components of X. */
  if (pivot) {
    for (j=0 ; j<n ; ++j) x[pivot[j]-1]= wa[j];  /* ### FIX INDEX ### */
  } else {
    for (j=0 ; j<n ; ++j) x[j]= wa[j];
  }
}
#undef r_

#endif /* _GPL_NUM_QR_C defined ---------------------------------------------*/
