/*
 * gpl_ops0.h --
 *
 *	Definitions for miscellaneous vector operations in GPLib.
 *
 * Copyright (c) 1999 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id$
 *	$Log$
 */
#ifndef GPL_OPS0_H
#define GPL_OPS0_H 1
#include "gpl_base.h"

#ifdef __cplusplus
extern "C" {
#endif

/*---------------------------------------------------------------------------*/
/* FILLING ARRAYS */

extern gpl_byte_t     *gpl_fill_b(gpl_byte_t     x[], int n, gpl_byte_t     value);
extern short        *gpl_fill_s(short        x[], int n, short        value);
extern int          *gpl_fill_i(int          x[], int n, int          value);
extern long         *gpl_fill_l(long         x[], int n, long         value);
extern float        *gpl_fill_f(float        x[], int n, float        value);
extern double       *gpl_fill_d(double       x[], int n, double       value);
extern gpl_fcomplex_t *gpl_fill_c(gpl_fcomplex_t x[], int n, gpl_fcomplex_t value);
extern gpl_dcomplex_t *gpl_fill_z(gpl_dcomplex_t x[], int n, gpl_dcomplex_t value);
/*	Fill N elements of array X with VALUE and return X.*/

/*---------------------------------------------------------------------------*/
/* SORTING ARRAYS */

extern int *gpl_sort_b(int *dst, const gpl_byte_t *src, int number);
extern int *gpl_sort_s(int *dst, const short    *src, int number);
extern int *gpl_sort_i(int *dst, const int      *src, int number);
extern int *gpl_sort_l(int *dst, const long     *src, int number);
extern int *gpl_sort_f(int *dst, const float    *src, int number);
extern int *gpl_sort_d(int *dst, const double   *src, int number);
/*	Store in DST index for array SRC sorted in ascending order.  Based
	on "heapsort" algorithm from "Numerical Recipes in C".*/

/*---------------------------------------------------------------------------*/
/* PSEUDO-RANDOM GENERATORS */

extern void gpl_random_seed(double seed);
/*	Initialize pseudo-random number generator with SEED.  If
	SEED<=0 or SEED>=1, a default SEED is used instead. */

extern void gpl_random_uniform_f(float x[], int number);
extern void gpl_random_uniform_d(double x[], int number);
/*	Fill array X (with NUMBER elements) with pseudo-random numbers
	following a uniform law (between 0 and 1). Uniform random generator
	is directly derived from the one implemented in Yorick by David
	Munro which, in turn, is based on Numerical Recipes in C. */

extern void gpl_random_normal_f(float x[], int number);
extern void gpl_random_normal_d(double x[], int number);
/*	Fill array X (with NUMBER elements) with pseudo-random numbers
	following a normal law (i.e. mean 0 and standard deviation 1). */

/*---------------------------------------------------------------------------*/
extern double gpl_euclidian_norm_f(const float  x[], int number);
extern double gpl_euclidian_norm_d(const double x[], int number);
/*	Compute Euclidian norm sqrt(sum(X*X)), avoiding overflows. */

extern double gpl_rms_b(const gpl_byte_t x[], int number);
extern double gpl_rms_s(const short    x[], int number);
extern double gpl_rms_i(const int      x[], int number);
extern double gpl_rms_l(const long     x[], int number);
extern double gpl_rms_f(const float  x[], int number);
extern double gpl_rms_d(const double x[], int number);
/*	Compute RMS value of X sqrt(sum(X*X)/NUMBER), avoiding overflows. */

extern double gpl_dot_product_f(const float  x[], const float  y[], int number);
extern double gpl_dot_product_d(const double x[], const double y[], int number);
/*	Compute sum(X*Y). */

extern float  *gpl_xpay_f(float xpay[], const float x[], float a,
			const float y[], int number);
extern double *gpl_xpay_d(double xpay[], const double x[], double a,
			const double y[], int number);
/*	Compute XAPY[i]= X[i] + A * Y[i] for i=0..NUMBER-1; return XAPY.
	Efficiently cope with A=0 (i.e. just copy X into XPAY) and A=1 (i.e.
	just add X plus Y into XPAY.  Operation can be made in-place: XPAY
	can be X or Y. */

extern float gpl_amax_f(const float x[], int number);
extern double gpl_amax_d(const double x[], int number);
/*	Return maximum absolute value of X (note: return 0 if NUMBER <= 0). */

extern gpl_byte_t gpl_min_b(const gpl_byte_t x[], int number);
extern short    gpl_min_s(const short    x[], int number);
extern int      gpl_min_i(const int      x[], int number);
extern long     gpl_min_l(const long     x[], int number);
extern float    gpl_min_f(const float    x[], int number);
extern double   gpl_min_d(const double   x[], int number);
extern gpl_byte_t gpl_max_b(const gpl_byte_t x[], int number);
extern short    gpl_max_s(const short    x[], int number);
extern int      gpl_max_i(const int      x[], int number);
extern long     gpl_max_l(const long     x[], int number);
extern float    gpl_max_f(const float    x[], int number);
extern double   gpl_max_d(const double   x[], int number);
/*	Return minimum/maximum value of X. */


extern gpl_byte_t gpl_sum_b(const gpl_byte_t x[], int number);
extern short    gpl_sum_s(const short    x[], int number);
extern int      gpl_sum_i(const int      x[], int number);
extern long     gpl_sum_l(const long     x[], int number);
extern float    gpl_sum_f(const float    x[], int number);
extern double   gpl_sum_d(const double   x[], int number);
extern void     gpl_sum_c(gpl_fcomplex_t *sum, const gpl_fcomplex_t z[],
			int number);
extern void     gpl_sum_z(gpl_dcomplex_t *sum, const gpl_dcomplex_t z[],
			int number);
/*	Return sum of values of X.  For complex array, the result is stored
	into SUM. */


extern double gpl_avg_b(const gpl_byte_t x[], int number);
extern double gpl_avg_s(const short    x[], int number);
extern double gpl_avg_i(const int      x[], int number);
extern double gpl_avg_l(const long     x[], int number);
extern double gpl_avg_f(const float    x[], int number);
extern double gpl_avg_d(const double   x[], int number);
extern void   gpl_avg_c(gpl_fcomplex_t *avg, const gpl_fcomplex_t z[], int number);
extern void   gpl_avg_z(gpl_dcomplex_t *avg, const gpl_dcomplex_t z[], int number);
/*	Return average of values of X.  For complex array, the result is stored
	into AVG. */
/*---------------------------------------------------------------------------*/
#ifdef __cplusplus
} /* C++ */
#endif
#endif /* GPL_OPS0_H ----------------------------------------------------------*/
