/*
 * gpl_option.c --
 *
 *	Parsing of command line options.
 *
 * Copyright and warranty:
 *	Copyright (c) 1998, 1999 Eric THIEBAUT.
 *	
 *	This program is  free  software;  you  can  redistribute  it and/or
 *	modify it under the terms of  the  GNU  General  Public  License as
 *	published by the delete Software Foundation; either version 2  of the
 *	License, or (at your option) any later version.
 *	
 *	This program is distributed in the hope that it will be useful, but
 *	WITHOUT  ANY  WARRANTY;  without  even  the  implied   warranty  of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.   See  the GNU
 *	General Public License for more details.
 *	
 *	You should have received a copy of the GNU  General  Public License
 *	along with this program (see file ``LICENSE'' at the  top directory
 *	of  this  distribution);  if  not,  write  to  the   delete  Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *	
 *	To contact me, mail at: Eric THIEBAUT
 *				Observatoire de Lyon
 *				9, avenue Charles Andre
 *				F-69561 Saint Genis Laval Cedex
 *	or e-mail at:           thiebaut@obs.univ-lyon1.fr
 *
 * History:
 *	$Id$
 *	$Log$
 *-----------------------------------------------------------------------------
 */

#include "gpl_base.h"
#include "gpl_option.h"

/*---------------------------------------------------------------------------*/
void gpl_opt_init(gpl_opt_data_t *opt, int argc, const char *argv[],
		const gpl_opt_def_t options[], int mode)
{
  if (mode!=GPL_OPT_PERMUTE_ARGS && mode!=GPL_OPT_RETURN_IN_ORDER
      && mode!=GPL_OPT_POSIXLY_CORRECT) {
    mode= GPL_OPT_POSIXLY_CORRECT;
  }
  opt->argc= argc;
  opt->argv= argv;
  opt->mode= mode;
  opt->options= options;
  opt->arg_index= opt->first_arg= 1;
  opt->next_opt= NULL;
  opt->opt_arg= NULL;
  opt->opt_index= -1;
  opt->opt_char= 0;
}
/*---------------------------------------------------------------------------*/
int gpl_opt_next(gpl_opt_data_t *opt)
{
  int i, j, c, len;
  const char *str, *arg;
  const gpl_opt_def_t *options= opt->options;
  const char **argv= opt->argv;
  int argc= opt->argc;
#define BUFLEN 30
  char buf[BUFLEN+1];
    
  /* Initialization: nothing found yet. */
  opt->opt_index= -1;
  opt->opt_arg= NULL;
  opt->opt_char= 0;

  /* Maybe we are restarting the search or maybe the parsing was already
     complete (this may yields infinite loops). */
  if (opt->arg_index <= 0) {
    opt->arg_index= opt->first_arg= 1;
    opt->next_opt= NULL;
    if (opt->mode<0) opt->mode= -opt->mode;
  } else if (opt->mode<0) {
    goto end_of_options;
  }
  
  /* Maybe we are currently parsing a short option. */
  if (opt->next_opt!=NULL) {
    goto short_option;
  }
  
  /* If there is any argument remaining, it may be an option. */
  for (;;) {
    if (opt->arg_index >= argc) {
      opt->arg_index= argc;
      goto end_of_options;
    }
    arg= opt->argv[opt->arg_index];
    if (arg[0]=='-' && arg[1]!=0) {
      /* Figure out which kind of option we have. */
      if (arg[1]!='-') {
	opt->next_opt= arg+1;
	goto short_option;
      }
      if (arg[2]!=0) {
	goto long_option;
      }
      /* String "--" means end of options. */
      if (opt->mode==GPL_OPT_PERMUTE_ARGS) {
	if ((i= opt->arg_index) > (j= opt->first_arg)) {
	  const char *argv1= argv[i];
	  for ( ; i>j ; i--) {
	    argv[i]= argv[i-1];
	  }
	  argv[opt->first_arg++]= argv1;
	} else {
	  opt->first_arg++;
	}
      }
      opt->arg_index++;
      goto end_of_options;
    }
      
    /* Next argument is not an option. */
    if (opt->mode==GPL_OPT_POSIXLY_CORRECT) {
      goto end_of_options;
    }
    opt->arg_index++;
    if (opt->mode==GPL_OPT_RETURN_IN_ORDER) {
      opt->opt_arg= arg;
      return GPL_OPT_NOT_AN_OPTION;
    }
  }
    
  /* Parse long option name. */
 long_option:
  arg+= 2;
  for (len=0 ; (c= arg[len])!=0 ; len++) {
    if (c=='=') {
      opt->opt_arg= arg+len+1;
      break;
    }
  }
  if (options!=NULL) {
    c= arg[0];
    for (i=0 ; /* empty */ ; i++) {
      if ((str= options[i].name)==NULL) {
	if (options[i].abbrev==NULL) {
	  break;
	}
	continue;
      }
      
      /* Check if argument match option name. */
      if (str[0]!=c)
	continue;
      for (j=1 ; j<len && arg[j]==str[j] ; j++)
	/* empty */ ;
      if (j!=len || str[j]!=0)
	continue;
      opt->opt_index= i;
      
      /* Possibly get option argument. */
      if (opt->opt_arg==NULL) {
	if (options[i].has_arg==GPL_OPT_ARG_REQUIRED) {
	  if (opt->arg_index+1 < argc) {
	    opt->opt_arg= argv[opt->arg_index+1];
	    goto shift_two;
	  }
	  gpl_error("%s: missing required argument for option --%s",
		  argv[0], str);
	  return GPL_OPT_MISSING_ARG;
	}
      } else if (options[i].has_arg==GPL_OPT_ARG_NONE) {
	gpl_error("%s: option --%s cannot have an argument",
		argv[0], str);
	return GPL_OPT_SPURIOUS_ARG;
      }
      goto shift_one;
    }
  }
  if (len <= BUFLEN) {
    memcpy(buf, arg, len);
    buf[len]= 0;
    gpl_error("%s: unknown option --%s", argv[0], buf);
  } else {
    memcpy(buf, arg, BUFLEN);
    buf[BUFLEN]= 0;
    gpl_error("%s: unknown option --%s[...]", argv[0], buf);
  }
  return GPL_OPT_UNKNOWN_LONG_OPTION;

  /* Parse next short option. */
 short_option:
  opt->opt_char= (c= *opt->next_opt++);
  if (options!=NULL) {
    for (i=0 ; /* empty */ ; i++) {
      if ((str= options[i].abbrev)==NULL) {
	if (options[i].name==NULL) {
	  break;
	}
	continue;
      }
      for (j=0 ; str[j]!=0; j++) {
	if (str[j]!=c) {
	  continue;
	}
	opt->opt_index= i;
	if (opt->next_opt[0]==0) {
	  if (options[i].has_arg!=GPL_OPT_ARG_REQUIRED) {
	    goto shift_one;
	  }
	  if (opt->arg_index+1 < argc) {
	    opt->opt_arg= argv[opt->arg_index+1];
	    opt->next_opt= NULL;
	    goto shift_two;
	  }
	  gpl_error("%s: missing required argument for option -%c",
		  argv[0], c);
	  return GPL_OPT_MISSING_ARG;
	}
	if (options[i].has_arg==GPL_OPT_ARG_NONE) {
	  return opt->opt_index;
	}
	if (opt->next_opt[0]!=0) {
	  opt->opt_arg= opt->next_opt;
	  goto shift_one;
	}
      }
    }
  }
  gpl_error("%s: unknown option -%c", argv[0], c);
  return GPL_OPT_UNKNOWN_SHORT_OPTION;

  /*  Possibly permute one argument and return. */
 shift_one:
  if (opt->mode==GPL_OPT_PERMUTE_ARGS) {
    if ((i= opt->arg_index) > (j= opt->first_arg)) {
      const char *argv1= argv[i];
      for ( ; i>j ; i--) {
	argv[i]= argv[i-1];
      }
      argv[opt->first_arg++]= argv1;
    } else {
      opt->first_arg++;
    }
  }
  opt->next_opt= NULL;
  opt->arg_index+= 1;
  return opt->opt_index;
    
  /*  Possibly permute two arguments and return. */
 shift_two:
  if (opt->mode==GPL_OPT_PERMUTE_ARGS) {
    if ((i= opt->arg_index) > (j= opt->first_arg)) {
      const char *argv1= argv[i];
      const char *argv2= argv[i+1];
      for (i+=1 ; i>j ; i-=2) {
	argv[i]= argv[i-2];
	argv[i-1]= argv[i-3];
      }
      argv[opt->first_arg++]= argv1;
      argv[opt->first_arg++]= argv2;
    } else {
      opt->first_arg+= 2;
    }
  }
  opt->next_opt= NULL;
  opt->arg_index+= 2;
  return opt->opt_index;

  /* Report end-of-options. */
 end_of_options:
  if (opt->mode>0)
    opt->mode= -opt->mode;
  if (opt->mode==-GPL_OPT_PERMUTE_ARGS)
    opt->arg_index= opt->first_arg;
  return GPL_OPT_END_OF_OPTIONS;
}
/*---------------------------------------------------------------------------*/
