/*
 * gpl_random.c --
 *
 *	Implementation of pseudo-random generators in GPLib.
 *
 *	Uniform random generator is directly derived from the one implemented
 *	in Yorick by David Munro which, in turn, is based on Numerical Recipes
 *	in C.
 *
 * Copyright (c) 1999 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *	
 * History:
 *	$Id$
 *	$Log$
 */
#include <math.h>
#include "gpl_ops0.h"

extern void gpl_random_seed(double seed);
/*	Initialize pseudo-random number generator with SEED.  If
	SEED<=0 or SEED>=1, a default SEED is used instead. */

extern void gpl_random_uniform_f(float x[], int number);
extern void gpl_random_uniform_d(double x[], int number);
/*	Fill array X (with NUMBER elements) with pseudo-random numbers
	following a uniform law (between 0 and 1). */

extern void gpl_random_normal_f(float x[], int number);
extern void gpl_random_normal_d(double x[], int number);
/*	Fill array X (with NUMBER elements) with pseudo-random numbers
	following a normal law (i.e. mean 0 and standard deviation 1). */

/*--------------------------------------------------------------------------*/
#ifndef GPL_RANDOM_C
#define GPL_RANDOM_C 1

/*---------------------------------------------------------------------------*/

/* Algorithm from Press and Teukolsky, Computers in Physics, 6, #5,
   Sep/Oct 1992, pp. 522-524.  They offer a $1000 reward to anyone
   who finds a statistical test this generator fails non-trivially.
   Based on a generator of L'Ecuyer with a shuffle algorithm of
   Bays-Durham and other improvements.
   The period of the generator is 2.3e18.  */

/* The long period is achieved by combining two sequences with
   nearly incomensurate periods.  */
static long idum= 0;       /* this is the primary seed */
static long idum2= 0;      /* this is the secondary seed */
#undef NTAB
#define NTAB 32
static long iv[NTAB], iy;  /* shuffle table (of idum) and previous result */

/* Choose two sets of linear congruential parameters-- the
   multiplier is IA, and the modulus is IM.  The IR and IQ are
   cunningly chosen so that (IA*x)%IM can be computed as
   IA*(x%IQ) - IR*(x/IQ), the latter expression having the
   advantage that the product will not overflow.  The IM values
   are near 2^31-1, the largest guaranteed positive long.
   The clever calculation of (IA*x)%IM is due to Schrage,
   ACM Trans. Mathem. Software 5, 132-138;
   IM= IA*IQ+IR with IR<IQ is the required relation.  */
#undef IM1
#define IM1 2147483563
#undef IA1
#define IA1 40014
#undef IQ1
#define IQ1 53668
#undef IR1
#define IR1 12211

#undef IM2
#define IM2 2147483399
#undef IA2
#define IA2 40692
#undef IQ2
#define IQ2 52774
#undef IR2
#define IR2 3791

#undef IMM1
#define IMM1 (IM1-1)
#undef NDIV
#define NDIV (1+IMM1/NTAB)


void gpl_random_seed(double seed)
{
  long j;

  /* translate seed to integer between 1 and IM2-1 inclusive */
  if (seed<=0.0 || seed>=1.0) seed= 0.6180339885;  /* default seed */
  idum= (long)(seed*(double)IM2);
  if (idum<1 || idum>=IM2) idum= 1;
  idum2= idum;

  /* do 8 warm-ups, then load shuffle table */
  for (j=NTAB+7 ; j>=0 ; j--) {
    idum= IA1*(idum%IQ1) - IR1*(idum/IQ1);
    if (idum<0) idum+= IM1;
    if (j<NTAB) iv[j]= idum;
  }
  iy= iv[0];
}

/*---------------------------------------------------------------------------*/

#define CTYPE           double
#define	RANDOM_UNIFORM  gpl_random_uniform_d
#define	RANDOM_NORMAL   gpl_random_normal_d
#define	ZERO            0.0
#define	ZERO_POINT_FIVE 0.5
#define	ONE             1.0
#define	TWO             2.0
#include __FILE__

#define CTYPE           float
#define	RANDOM_UNIFORM  gpl_random_uniform_f
#define	RANDOM_NORMAL   gpl_random_normal_f
#define	ZERO            0.0F
#define	ZERO_POINT_FIVE 0.5F
#define	ONE             1.0F
#define	TWO             2.0F
#include __FILE__

#else /* GPL_RANDOM_C defined */

/*---------------------------------------------------------------------------*/

void RANDOM_UNIFORM(CTYPE *random, int number)
{
  long j;
  if (idum<=0) gpl_random_seed(0.0);

  if (number <= 0) return;
  while (number--) {
    /* compute idum= (IA1*idum)%IM1 without overflow */
    idum= IA1*(idum%IQ1) - IR1*(idum/IQ1);
    if (idum<0) idum+= IM1;

    /* compute idum2= (IA2*idum2)%IM2 without overflow */
    idum2= IA2*(idum2%IQ2) - IR2*(idum2/IQ2);
    if (idum2<0) idum2+= IM2;

    /* previous result is used to determine which element of the shuffle
       table to use for this result */
    j= iy/NDIV;            /* in range 0..NTAB-1 */
    iy= iv[j]-idum2;
    iv[j]= idum;
    if (iy<1) iy+= IMM1;

    /* Really only IMM1 possible values can be returned, 1<=iy<=IMM1.
       Algorithm given by Press and Teukolsky has a slight bug.
       Here, the return values are centered in IMM1 equal bins.
       If 2.e9 distinct values are not enough, could use, say, idum2
       to scoot the points around randomly within bins...  */
    *random++= (ONE/(CTYPE)IM1)*(iy-ZERO_POINT_FIVE);
  }
}

void RANDOM_NORMAL(CTYPE *random, int number)
{
  static int    use_last= 0;     /* Last normal random value not yet used? */
  static double last_value= 0.0; /* Last normal random value, if not used. */
  double fac, r, x0, x1, x[2];
  int i= 0;
  
  if (number<=0) return;
  if (use_last) {
    *random++= last_value;
    number--;
    use_last= 0;
  }
  i= 0;
  while (i<number) {
    do {
      gpl_random_uniform_d(x, 2);
      x0= 2.0 * x[0] - 1.0;
      x1= 2.0 * x[1] - 1.0;
      r= x0*x0 + x1*x1;
    } while (r >= ONE);
    fac= sqrt(-2.0 * log(r) / r);
    random[i++]= fac * x0;
    if (i < number) {
      random[i++]= fac * x1;
    } else {
      last_value= fac * x1;
      use_last= 1;
      break;
    }
  }
}

/*---------------------------------------------------------------------------*/
#undef CTYPE
#undef RANDOM_UNIFORM
#undef RANDOM_NORMAL
#undef ZERO
#undef ZERO_POINT_FIVE
#undef ONE
#undef TWO
#endif /* GPL_RANDOM_C defined */
