/*
 * gpl_smooth.h --
 *
 *	Definitions for convolution and wavelet smoothing along
 *	one dimension in GPLib.
 *
 * Copyright (c) 1997, 1999, 2000 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id$
 *	$Log$
 *-----------------------------------------------------------------------------
 */
#ifndef _GPL_SMOOTH_H
#define _GPL_SMOOTH_H 1

#include "gpl_base.h"
#ifdef __cplusplus
extern "C" {
#endif

/*---------------------------------------------------------------------------*/
/* SMOOTHING AND WAVELET SMOOTHING ALONG ONE DIMENSION */

extern void gpl_smooth_f(float dst[], const float src[], int stride,
		       int n, int nafter, const float ker[], int w,
		       int scale, int border, float ws[]);
extern void gpl_smooth_d(double dst[], const double src[], int stride,
		       int n, int nafter, const double ker[], int w,
		       int scale, int border, double ws[]);
extern void gpl_smooth_c(gpl_fcomplex_t dst[],
		       const gpl_fcomplex_t src[], int stride,
		       int n, int nafter, const float ker[], int w,
		       int scale, int border, float ws[]);
extern void gpl_smooth_z(gpl_dcomplex_t dst[],
		       const gpl_dcomplex_t src[], int stride,
		       int n, int nafter, const double ker[], int w,
		       int scale, int border, double ws[]);
/*
 *   Convolve a multidimensional array by a given kernel along one dimension
 *   at a given scale.  These routines are mainly intended for smoothing
 *   operations (i.e. with a positive kernel).
 *
 * ARGUMENTS
 *   DST    - [output, N*STRIDE*NAFTER elements]
 *            Result of the convolution along the dimension of interest.
 *   SRC    - [input, N*STRIDE*NAFTER elements]
 *            Array to convolve.
 *   STRIDE - [input]
 *            The stride of the dimension of interest (i.e. the spacing of
 *            elements along this dimension), e.g.:
 *                STRIDE=1         to operate onto the first dimension of SRC;
 *                STRIDE=LEN1      where LEN1 is the length of the first
 *                                 dimension to operate onto the second
 *                                 dimension;
 *                STRIDE=LEN1*LEN2 to operate onto the third dimension, and
 *                                 so on.
 *   N      - [input, N]
 *            The length of the dimension of interest.
 *   NAFTER - [input]
 *            The product of all dimension lengths after the dimension of
 *            interest -- the operation of length N will be repeated
 *            a total of STRIDE*NAFTER times.
 *   KER    - [input, 2*W+1 elements]
 *            Kernel of convolution.
 *   W      - [input, W>=0]
 *            Half-width of KER (i.e. KER has 2*W+1 elements).
 *   SCALE  - [input, SCALE>=1]
 *            Wavelet scale for the ``a trou'' method (SCALE=1 for simple
 *            convolution).
 *
 *   BORDER - [input]
 *            Boundary conditions:
 *                BORDER=0  to extrapolate missing values by the
 *                          left/rightmost ones;
 *                BORDER=1  to extrapolate missing left values by zero and
 *                          missing right values by the rightmost one;
 *                BORDER=2  to extrapolate missing left values by the leftmost
 *                          one and missing right values by zero;
 *                BORDER=3  to extrapolate missing left/right values by zero.
 *                BORDER=4  to use periodic conditions;
 *                BORDER>4 or BORDER<0
 *                          to not extrapolate missing values but normalize
 *                          convolution product by sum of kernel weights taken
 *                          into account (assuming they are all positive).
 *   WS      - [workspace, 0, N, or 2*N elements]
 *             Workspace: 0 elements (i.e. unused) if STRIDE=1 and SRC is not
 *             the same as DST (overlapping is not checked); N elements, if
 *             STRIDE=1 and SRC=DST; otherwise, at least 2*N elements.
 *
 * DESCRIPTION
 *                  ___
 *                  \
 *   Store: DST[i]=  > KER[j+W] * SRC[i+j*SCALE]     along the direction of
 *                  /__                                           [interest.
 *             -W <= j <= +W
 * RETURN
 *   None.
 *
 * SIDE EFFECTS
 *   Contents of DST and WS are modified.
 */

#ifdef __cplusplus
} /* C++ */
#endif
#endif /* _GPL_SMOOTH_H -------------------------------------------------------*/
