/*
 * gpl_string.h --
 *
 *	Definitions for string utilities for GPLib library: provides conversion
 *	of human readable number to binary values, management of dynamic
 *	string, character conversion, ...
 *
 * Copyright (c) 1999 Eric THIEBAUT.
 *	See the file "LICENSE" for information on usage and redistribution
 *	of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * History:
 *	$Id$
 *	$Log$
 */
#ifndef GPL_STRING_H
#define GPL_STRING_H 1

#include <stdio.h>

/* C++ needs to know that types and declarations are C, not C++.  */
#ifdef  __cplusplus
#undef __BEGIN_DECLS
#undef __END_DECLS
# define __BEGIN_DECLS  extern "C" {
# define __END_DECLS    }
#else
# define __BEGIN_DECLS
# define __END_DECLS
#endif

__BEGIN_DECLS

/*---------------------------------------------------------------------------*/
/* CHARACTER ROUTINES */

extern int gpl_to_lower(int c);
extern int gpl_to_upper(int c);
/*----- These functions convert their argument to lower and upper case
	letter respectively (except that EOF (-1) remains EOF).  Assume
	ISO-Latin1 encoding for accentuated letters.  Also implemented as
	macros/inline functions.  Use GPL_TO_LOWER/GPL_TO_UPPER if you want
	to be sure to use the macro verison. */

extern int gpl_is_space(int c);
/*----- Check if character C is a space (any of '\t', '\n', '\v', '\f',
	'\r' or an ordinary space) or not.  Also implemented as
	macros/inline functions.  Use GPL_IS_SPACE if you want to be sure
	to use the macro verison.  Note: gpl_is_space(0) and
	GPL_IS_SPACE(0) are both false (as expected ;-). */

extern const short *_gpl_to_lower, *_gpl_to_upper;
extern const char  *_gpl_is_space;
/*----- These (private although globally accessible) pointers are used for
	fast character check/conversion and can be indexed in the range
	-128..255; so that their index -- the argument of the corresponding
	character macros/function -- can be either a signed or unsigned
	char.  Internally, static arrays of 384 elements are used. */

/* Character macros. */
#define GPL_IS_SPACE(c) (_gpl_is_space[(int)(c)])
#define GPL_TO_LOWER(c) (_gpl_to_lower[(int)(c)])
#define GPL_TO_UPPER(c) (_gpl_to_upper[(int)(c)])

/*---------------------------------------------------------------------------*/
/* STRING ROUTINES */

extern int gpl_str_get_integer(const char *string, long *long_ptr);
extern int gpl_str_get_real(const char *string, double *double_ptr);
/*----- Convert zero terminated input string STRING into a long integer (a
	double precision real) stored at LONG_PTR (resp. DOUBLE_PTR).  The
	string may contains leading/trailing spaces.  If the conversion
	succeeds, GPL_OK is returned; otherwise, GPL_ERROR is returned and
	the error message can be retrieved by gpl_error_info(). */

extern char *gpl_str_clone(const char *string);
/*----- Return dynamically allocated memory with a copy of STRING.  NULL is
	returned in case of error or if STRING is NULL. */

extern char *gpl_str_print_int(char *string, int value);
extern char *gpl_str_print_long(char *string, long value);
/*----- Print into STRING base 10 human readable representation of int/long
	VALUE.  About 3 times faster than `sprintf("%d", STRING,
	VALUE)'. */

extern char *gpl_str_upper(register char *dst, register const char *src);
extern char *gpl_str_lower(register char *dst, register const char *src);
/*----- Convert zero terminated input string SRC to lower/upper case
	letters (NULL string are the same as empty ones) and store the
	result in DST.  Return DST.  Operation can be done `in-place': with
	DST==SRC. */

extern char *gpl_str_trim(char *dst, const char *src, int flags);
#define GPL_STR_TRIM_LEFT  1
#define GPL_STR_TRIM_RIGHT 2
#define GPL_STR_TRIM_BOTH  (GPL_STR_TRIM_LEFT|GPL_STR_TRIM_RIGHT)
/*----- Copy zero-terminating string SRC into DST removing leading (bit
	GPL_STR_TRIM_LEFT of FLAGS is set) and/or trailing (bit
	GPL_STR_TRIM_RIGHT of FLAGS is set) spaces.  Returns DST. */

extern char *gpl_str_truncate(char *buf, const char *string, int maxlen,
			      const char *ellipsis);
/*----- Copy string with possible truncation.  If strlen(STRING) <= MAXLEN,
	copy STRING into BUF; otherwise, copy MAXLEN-strlen(ELLIPSIS)
	characters of STRING into BUF and append ELLIPSIS (or "[...]"  if
	ELLIPSIS is NULL).  Return BUF or NULL if STRING is NULL or MAXLEN
	<= strlen(ELLIPSIS) (or 5 if ELLIPSIS is NULL).  BUF must be at
	least MAXLEN+1 byte long.  */

extern const char *gpl_str_ordinal_suffix(int number);
/*----- Return static string "st" or "nd" or "rd" or "th". */

extern char *gpl_str_token(char *string, char **next, int *sval,
			 const char *slist);
/*----- Return next token from STRING or NULL if token list is exhausted.
	Tokens are separated by characters found in SLIST or spaces if
	SLIST is NULL (see GPL_IS_SPACE for the list of space characters).
	The end of the returned token is marked by a '\0' in STRING, if
	SVAL is not NULL the value of the separator that terminates the
	token is stored in *SVAL.  NEXT the address of a string pointer
	used to keep track of the position in STRING where to start
	searching next token: at first call *NEXT must be NULL and *NEXT
	must not be modified between calls to gpl_str_token. */

/*---------------------------------------------------------------------------*/
/* DYNAMIC STRING ROUTINES */

/* Structure to store a dynamic string.  Clients are not allowed to directly
   access, members of this structure (use the below defined macros instead). */
typedef struct _gpl_dstr gpl_dstr_t;
struct _gpl_dstr {
    char *string; /* Pointer to gpl_alloc'ed memory or to the internal static
		     buffer. */
    int   length; /* Actual string length. */
    int   maxlen; /* Maximum string length (it is guaranted that there is an
		     extra character left for the terminating zero). */
#define GPL_DSTR_STATIC_LENGTH 127
    char  buffer[GPL_DSTR_STATIC_LENGTH+1]; /* Internal static buffer. */
};

/* Clients should only use the following macros to access dynamic string
   members (DSP is the address of a valid -- initialized -- dynamic string
   structure). */
#define gpl_dstr_value(dsp)	((dsp)->string)
#define gpl_dstr_length(dsp)	((dsp)->length)
#define gpl_dstr_maxlen(dsp)	((dsp)->maxlen)

extern void gpl_dstr_init(gpl_dstr_t *dsp);
/*----- Initializes a dynamic string, discarding any previous contents of
	the string (gpl_dstr_free should have been called already if the
	dynamic string was previously in use). */

extern void gpl_dstr_free(gpl_dstr_t *dsp);
/*----- Frees up any memory allocated for the dynamic string and
	reinitializes the string to an empty state pointing to the internal
	buffer. */

extern gpl_dstr_t *gpl_dstr_reset(gpl_dstr_t *dsp);
/*----- Reset dynamic string to be empty (i.e. of zero length) and return
	its argument. */

extern int gpl_dstr_resize(gpl_dstr_t *dsp, int length);
/*----- Resizes dynamic string buffer so that it can store at least strings
	of LENGTH characters.  Returns standard result: GPL_OK or GPL_ERROR
	-- in that case, the error message can be retrieved with
	gpl_error_info(). */

extern int gpl_dstr_append_char(gpl_dstr_t *dsp, int c);
/*----- Append a single character C to the dynamic string DSP.  Returns
	standard result: GPL_OK or GPL_ERROR -- in that case, the error
	message can be retrieved with gpl_error_info(). */

extern int gpl_dstr_append(gpl_dstr_t *dsp, const char *string, int length);
/*----- Appends STRING to the dynamic string DSP.  If LENGTH is less than
	zero, strlen(STRING) is used instead; otherwise only LENGTH
	characters (including NULL's if any) are copied to the end of DSP
	(plus a terminating zero).  Returns standard result: GPL_OK or
	GPL_ERROR -- in that case, the error message can be retrieved with
	gpl_error_info(). */

extern int gpl_dstr_append_int(gpl_dstr_t *dsp, int value, int separator);
extern int gpl_dstr_append_ints(gpl_dstr_t *dsp, const int array[],
				int number, int separator);
extern int gpl_dstr_append_double(gpl_dstr_t *dsp, double value,
				  int separator);
extern int gpl_dstr_append_doubles(gpl_dstr_t *dsp, const double array[],
				   int number, int separator);
/*----- Convert value(s) into human readable ASCII string taking care that
	floating-point values do not look like integer and append them to
	DSP.  If SEPARATOR is not '\0' this character is used as a
	separator before each value but the first one (i.e. the one which
	starts the resulting string).  Return standard result: GPL_OK or
	GPL_ERROR -- in that case, the error message can be retrieved with
	gpl_error_info(). */


extern char *gpl_dstr_steal(gpl_dstr_t *dsp);
/*----- Steal contents of dynamic string pointed by DSP.  Return a dynamic
	string that must be gpl_free'd by the caller and reset input
	dynamic string.  In case of error, returns NULL, and error message
	can be retrieved with gpl_error_info(). */

extern int gpl_dstr_read_line(gpl_dstr_t *dsp, FILE *stream, int *eof_ptr,
			      int *line_count_ptr);
/*----- Reads next text line from STREAM and stores it in dynamic string
	DSP.  In actual files and devices the end of a line may be
	represented differently on different platforms, or even for
	different devices on the same platform. For example, under UNIX
	newlines (\n) are used in files, whereas carriage-return-linefeed
	(\r\n) sequences are normally used in network connections.  This
	routine recognizes, in that order, EOF, \r\n, \r or \n sequences as
	valid end-of-line and replaces them by a null char (\0).
	
	If EOF_PTR is non-NULL, its value is set to 1 upon end-of-file and
	to 0 otherwise.  If LINE_COUNT_PTR is non-NULL, its value is
	incremented if a new line is actually read.
	
	Returns standard result: GPL_OK or GPL_ERROR -- in that case, left
	an error message that can be retrieved with gpl_error_info(). */

extern void gpl_dstr_lower(gpl_dstr_t *dsp);
extern void gpl_dstr_upper(gpl_dstr_t *dsp);
/*----- Converts dynamic string DSP to lower/upper case characters. */

extern void gpl_dstr_trim(gpl_dstr_t *dsp, int flags);
/*----- Removes leading (bit GPL_STR_TRIM_LEFT of FLAGS is set) and/or
	trailing (bit GPL_STR_TRIM_RIGHT of FLAGS is set) spaces from
	DSP. */

extern void gpl_dstr_pack(gpl_dstr_t *dsp);
/*----- Replace all occurences of multiple space characters in DSP by a
	single ordinary space. */

/*---------------------------------------------------------------------------*/

#if defined(HAVE_INLINE) && HAVE_INLINE
extern inline int gpl_to_lower(int c) { return (int)GPL_TO_LOWER(c); }
extern inline int gpl_to_upper(int c) { return (int)GPL_TO_UPPER(c); }
extern inline int gpl_is_space(int c) { return (int)GPL_IS_SPACE(c); }
#else /* not HAVE_INLINE */
# define gpl_to_lower(c) ((int)GPL_TO_LOWER(c))
# define gpl_to_upper(c) ((int)GPL_TO_UPPER(c))
# define gpl_is_space(c) ((int)GPL_IS_SPACE(c))
#endif /* not HAVE_INLINE */

__END_DECLS
#endif /* GPL_STRING_H ------------------------------------------------------*/
