/*
 * gpl_workspace.h --
 *
 *	Portable routines to manage temporary workspaces that can be
 *	automatically deallocated.
 *
 *	These routines are intended to allow garbage collection after
 *	interrupts or non-local jumps (see documentation in that file).
 *
 *	I have tried to make this package (gpl_workspace.h + gpl_workspace.c)
 *	portable (pre-ANSI as well as ANSI C and C++ should be fine) and
 *	usable without GPLib (in which case you must define macro
 *	GPL_WS_NO_ELIB to any non-zero integer value, see below).
 *
 *	Please report bugs to: thiebaut@obs.univ-lyon1.fr
 *
 * Copyright (c) 1998-1999 Eric THIEBAUT.
 *	The handling of the stack direction and the function i00afunc() are
 *	borrowed from GNU `alloca.c' code by D. A. Gwyn and is therefore
 *	subject to the GPL (GNU Public License).
 *
 * History:
 *	$Id$
 *	$Log$
 *-----------------------------------------------------------------------------
 */

#ifndef _GPL_WORKSPACE_H
#define _GPL_WORKSPACE_H 1

/* Define macro GPL_WS_NO_ELIB to non-zero if you are using the workspace
   routines outside GPLib. */
#ifndef GPL_WS_NO_ELIB
# define GPL_WS_NO_ELIB 0
#endif

#if !GPL_WS_NO_ELIB
# include "gpl_base.h"
#endif

#if defined(__STDC__) || defined(__cplusplus)
# define GPL_WS_ANSI_C      1
# define GPL_WS_ARGS(x)     x
# define GPL_WS_PTR         void *
# define GPL_WS_CONST       const
#else
# define GPL_WS_ANSI_C      0
# define GPL_WS_ARGS(x)     ()
# define GPL_WS_PTR         char *
# define GPL_WS_CONST
#endif

#ifdef __cplusplus
extern "C" {
#endif

/*---------------------------------------------------------------------------*/
/* WORKSPACE
   ---------
	Worskpaces are pieces of data, usually dynamic memory, temporarily
	needed by a routine that can be either explicitly or automatically
	deallocated.  Automatic deallocated occurs when the workspace manager
	find a workspace attached at a deeper stack depth than the current
	depth.

	Worskpaces can be passed to other called routines (i.e. deeeper in
	the stack) but should never be returned to the caller.  Only
	detached worskpaces could be returned to the caller.

	Advantages over alloca() are:
	 - workspaces can be detached/deleted at any location in the code;
	 - explicit as well as implicit garbage collection;
	 - any client data can be attached to allow its deletion after
	   interrupts or non-local jumps;
         - workspace manager keep its information in a separate private area
	   (the GNU alloca uses headers stored before each returned memory
	   block and which are easily corrupted by out of bounds writes).

	Here is a simple example: */
#if 0 /* SAMPLE */
void *foo(...)
{
  void *ptr1, *ptr2, *ptr3;
  char *depth= gpl_ws_cleanup(NULL); /* Get current depth and perform
				      pre-cleanup. */

  /* Allocate memory for workspaces. */
  ptr1= gpl_ws_create(depth, SIZE_1);
  ptr2= gpl_ws_create(depth, SIZE_2);
  ptr3= gpl_ws_create(depth, SIZE_3);
  ...;
  gpl_ws_delete(ptr2);                /* Intermediate free: PTR2 no more
				       needed. */
  ...;
  gpl_ws_detach(ptr3);                /* Detach PTR3 before returning it to the
				       caller. */

  gpl_ws_cleanup(depth);              /* Final cleanup (actually delete PTR1
				       but not PTR2 nor PTR3). */
  return ptr3;
}
#endif /* SAMPLE */

#if GPL_WS_NO_ELIB
/* Standard codes returned by those routines: */
#define GPL_OK     0 /* Routine completes successfully. */
#define GPL_ERROR -1 /* Routine fails, generally, an error message is left that
                      can be retrieved with gpl_error_info(CLEAR) -- the CLEAR
                      argument indicates whether (CLEAR==non-zero) or not
                      (CLEAR==0) the error message should be discarded. */

extern GPL_WS_CONST char *gpl_error_info GPL_WS_ARGS((int clear));
/*	Return last workspace routines error message.  Clear the error message
	if CLEAR is non-zero.  Always return a valid static string (i.e. even
	when no error occurred). */
#endif /* GPL_WS_NO_ELIB */

extern char *gpl_ws_cleanup GPL_WS_ARGS((char *depth));
/*	Perform garbage collection among managed workspaces.  This routine
	should be called at least twice by the user routine:
	 - on entry with a NULL value for DEPTH to get the current stack
	   depth and delete workspaces deeper or at the returned depth;
         - before exiting with the current depth as argument in order to
	   delete all workspaces at or deeper than DEPTH. */

typedef void gpl_ws_delete_t GPL_WS_ARGS((void *data));
/*	Prototype of routine called to delete (free) attached workspace. */

extern int gpl_ws_attach GPL_WS_ARGS((char *depth, void *data,
				  gpl_ws_delete_t *delete));
/*	Attach workspace DATA at given DEPTH.  The gpl_ws_cleanup/gpl_ws_delete
	routines will use the provided DELETE routine to eventually release
	DATA.  A standard result is returned: GPL_OK or GPL_ERROR (error
	message can be retrieved by calling gpl_error_info).  Note: use
	gpl_ws_detach, if you want to get back the control of the attached
	workspace.  Do not attach a workspace with NULL address and free(3)
	routine.  Do not attach the same workspace twice (for reasons of
	efficiency this is not checked) even if your code performs as many
	detach as attach. */

extern int gpl_ws_detach GPL_WS_ARGS((void *data));
/*	Detach workspace DATA (does the opposite of gpl_ws_attach).  The
	detached workspace is no longer managed by the workspace routines.
	A standard result is returned: GPL_OK or GPL_ERROR (error message can
	be retrieved by calling gpl_error_info). */

extern int gpl_ws_delete GPL_WS_ARGS((void *data));
/*	Delete workspace at address DATA.  Return standard result: GPL_OK or
	GPL_ERROR (use gpl_error_info() to retrieve the error message).  Note:
	this routine is intended to allow the deallocation of any attached
	workspace before cleanup, it may be rather expensive in terms CPU
	time (i.e. unless you really need to free some momory, preferably
	just call gpl_ws_cleanup at the end of your routine). */

extern GPL_WS_PTR gpl_ws_create GPL_WS_ARGS((char *depth, size_t size));
/*	Allocate a SIZE byte workspace at given DEPTH.  Return workspace
	address or NULL in case of error (use gpl_error_info() to retrieve
	the error message). */

extern void gpl_ws_info GPL_WS_ARGS((void));
/* Print out workspace information. */



/*---------------------------------------------------------------------------*/
#ifdef __cplusplus
} /* C++ */
#endif
#endif /* _GPL_WORKSPACE_H ----------------------------------------------------*/
