/*
 * yeti_channel.c --
 *
 *	Implement regular/compressed file input/output functions in Yeti.
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 2001-2002 Eric THIEBAUT.
 *
 *	This file is part of Yeti.
 *
 *	Yeti is  free software;  you can redistribute  it and/or  modify it
 *	under the terms  of the GNU General Public  License as published by
 *	the Free Software  Foundation; either version 2 of  the License, or
 *	(at your option) any later version.
 *
 *	Yeti is distributed in the hope that it will be useful, but WITHOUT
 *	ANY WARRANTY; without even  the implied warranty of MERCHANTABILITY
 *	or FITNESS  FOR A PARTICULAR  PURPOSE.  See the GNU  General Public
 *	License for more details.
 *
 *	You should have  received a copy of the  GNU General Public License
 *	along with  Yeti (file "COPYING"  in the top source  directory); if
 *	not, write to the Free  Software Foundation, Inc., 59 Temple Place,
 *	Suite 330,  Boston, MA  02111-1307 USA.
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id: yeti_channel.c,v 1.1 2005/05/24 13:12:07 eric Exp $
 *	$Log: yeti_channel.c,v $
 *	Revision 1.1  2005/05/24 13:12:07  eric
 *	Initial revision
 *
 */

#include "yeti.h"
#if (YORICK_VERSION_MAJOR == 1) && (YORICK_VERSION_MINOR <= 4)
# include <stdio.h>
#else
# include "pstdio.h"
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#ifndef HAVE_ZLIB
# define HAVE_ZLIB 0
#endif
#if HAVE_ZLIB
# include "zlib.h"
#else
/* Even if ZLib support is unwanted, we need the value of some constants. */
# define Z_NO_COMPRESSION         0
# define Z_BEST_SPEED             1
# define Z_BEST_COMPRESSION       9
# define Z_DEFAULT_COMPRESSION  (-1)
# define Z_FILTERED               1
# define Z_HUFFMAN_ONLY           2
# define Z_DEFAULT_STRATEGY       0
#endif

#include "yio.h"

#include "prmtyp.h"
#undef YETI_INT32
#undef YETI_UINT32
#undef YETI_BYTE_ORDER
#if ! defined(YETI_INT32) && (P_INT_SIZE == 4)
# define YETI_INT32       int
# define YETI_BYTE_ORDER  P_INT_ORDER
#endif
#if ! defined(YETI_INT32) && (P_LONG_SIZE == 4)
# define YETI_INT32       long
# define YETI_BYTE_ORDER  P_LONG_ORDER
#endif
#if ! defined(YETI_INT32) && (P_SHORT_SIZE == 4)
# define YETI_INT32       short
# define YETI_BYTE_ORDER  P_SHORT_ORDER
#endif
#if ! defined(YETI_INT32) && (P_INT_SIZE >= 4)
# define YETI_INT32       int
# define YETI_BYTE_ORDER  P_INT_ORDER
#endif
#if ! defined(YETI_INT32) && (P_LONG_SIZE >= 4)
# define YETI_INT32       long
# define YETI_BYTE_ORDER  P_LONG_ORDER
#endif
#if ! defined(YETI_INT32) && (P_SHORT_SIZE >= 4)
# define YETI_INT32       short
# define YETI_BYTE_ORDER  P_SHORT_ORDER
#endif
#if ! defined(YETI_INT32)
# error Cannot find suitable integer type.
#endif
#define YETI_UINT32 unsigned YETI_INT32

/*---------------------------------------------------------------------------*/
/* Built-in functions defined in this file: */
extern BuiltIn Y_chn_open, Y_chn_close;
extern BuiltIn Y_chn_read, Y_chn_write;
extern BuiltIn Y_chn_tell, Y_chn_seek, Y_chn_rewind;
extern BuiltIn Y_filepath, Y_get_path;

/* Yorick functions defined elsewhere: */
extern void BuildDimList(Symbol *stack, int argc);  /* ops3.c */
extern DataBlock *ForceToDB(Symbol *s);             /* ops3.c */

/*---------------------------------------------------------------------------*/
typedef struct _chn_t chn_t;
typedef struct _chn_ops_t chn_ops_t;

static chn_t *get_channel(Symbol *stack);
/*----- Returns i/o channel table stored by symbol STACK.  FIXME: STACK get
	replaced by the referenced object if it is a reference symbol. */

struct _chn_t {
  /* Common part of all Yorick's DataBlocks: */
  int references;     /* reference counter */
  Operations *ops;    /* virtual function table */

  /* Specific to this class of objects: */
  char *path;
  char mode[8];
  union {
    int    fd; /* normal file descriptor */
#if HAVE_ZLIB
    gzFile gz; /* file managed by zlib library */
#endif /* HAVE_ZLIB */
  } chn;
  chn_ops_t *chn_ops;
};

typedef void close_op_t(chn_t *chn);
typedef int    eof_op_t(chn_t *chn);
typedef long  tell_op_t(chn_t *chn);
typedef long  seek_op_t(chn_t *chn, long offset, int whence);
typedef long  read_op_t(chn_t *chn, void *ptr, long nbytes);
typedef void  gets_op_t(chn_t *chn, char *str, long nbytes);
typedef void write_op_t(chn_t *chn, void *ptr, long nbytes);
struct _chn_ops_t {
  close_op_t *close;
  tell_op_t  *tell;
  seek_op_t  *seek;
  read_op_t  *read;
  write_op_t *write;
};

#define CHN_CLOSE(this)                ((this)->chn_ops->close(this))
#define CHN_EOF(this)                  ((this)->chn_ops->eof(this))
#define CHN_TELL(this)                 ((this)->chn_ops->tell(this))
#define CHN_SEEK(this, offset, whence) ((this)->chn_ops->seek(this, offset, \
                                                              whence))
#define CHN_READ(this, ptr, nbytes)    ((this)->chn_ops->read(this, ptr, \
                                                              nbytes))
#define CHN_GETS(this, s, nbytes)      ((this)->chn_ops->gets(this, str, \
                                                              nbytes))
#define CHN_WRITE(this, ptr, nbytes)   ((this)->chn_ops->write(this, ptr, \
                                                               nbytes))

/* Channel operation tables. */

static  read_op_t bogus_read;
static write_op_t bogus_write;
static  tell_op_t bogus_tell;
static  seek_op_t bogus_seek;
static close_op_t bogus_close;
static chn_ops_t bogus_chn_ops = {
  bogus_close, bogus_tell, bogus_seek, bogus_read, bogus_write
};

#if HAVE_ZLIB
static  read_op_t gz_read;
static write_op_t gz_write;
static  tell_op_t gz_tell;
static  seek_op_t gz_seek;
static close_op_t gz_close;
static chn_ops_t gz_chn_ops = {
  gz_close, gz_tell, gz_seek, gz_read, gz_write
};
#endif /* HAVE_ZLIB */

static  read_op_t fd_read;
static write_op_t fd_write;
static  tell_op_t fd_tell;
static  seek_op_t fd_seek;
static close_op_t fd_close;
static chn_ops_t fd_chn_ops = {
  fd_close, fd_tell, fd_seek, fd_read, fd_write
};

/*--------------------------------------------------------------------------*/
/* IMPLEMENTATION OF I/O CHANNELS AS OPAQUE YORICK OBJECTS */

extern PromoteOp PromXX;
extern UnaryOp ToAnyX, NegateX, ComplementX, NotX, TrueX;
extern BinaryOp AddX, SubtractX, MultiplyX, DivideX, ModuloX, PowerX;
extern BinaryOp EqualX, NotEqualX, GreaterX, GreaterEQX;
extern BinaryOp ShiftLX, ShiftRX, OrX, AndX, XorX;
extern BinaryOp AssignX, MatMultX;
extern UnaryOp EvalX, SetupX, PrintX;
extern MemberOp GetMemberX;
static UnaryOp PrintChannel;
static void FreeChannel(void *addr);  /* ******* Use Unref(obj) ******* */

Operations channelOps = {
  &FreeChannel, T_OPAQUE, 0, T_STRING, "channel",
  {&PromXX, &PromXX, &PromXX, &PromXX, &PromXX, &PromXX, &PromXX, &PromXX},
  &ToAnyX, &ToAnyX, &ToAnyX, &ToAnyX, &ToAnyX, &ToAnyX, &ToAnyX,
  &NegateX, &ComplementX, &NotX, &TrueX,
  &AddX, &SubtractX, &MultiplyX, &DivideX, &ModuloX, &PowerX,
  &EqualX, &NotEqualX, &GreaterX, &GreaterEQX,
  &ShiftLX, &ShiftRX, &OrX, &AndX, &XorX,
  &AssignX, &EvalX, &SetupX, &GetMemberX, &MatMultX, &PrintChannel
};


/* FreeChannel is automatically called by Yorick to delete an object instance
   that is no longer referenced. */
static void FreeChannel(void *addr)
{
  chn_t *this = (chn_t *)addr;
  CHN_CLOSE(this);
  if (this->path) p_free(this->path);
  p_free(addr);
}

/* PrintChannel is used by Yorick's info command. */
static void PrintChannel(Operand *op)
{
  chn_t *this = (chn_t *)op->value;
  char line[80];
  ForceNewline();
  PrintFunc("Object of type: ");
  PrintFunc(this->ops->typeName);
  sprintf(line, " (references=%d, path=\"", this->references);
  PrintFunc(line);
  PrintFunc(this->path);
  PrintFunc("\", mode=\"");
  PrintFunc(this->mode);
  PrintFunc("\")");
  ForceNewline();
}

/* DOCUMENT chn_open(path, mode)
     Open a channel.

   SEE ALSO chn_seek, chn_read, chn_write. */

/*---------------------------------------------------------------------------*/
/* Operations for regular files. */

static void posix_error(void);

static void posix_error(void)
{
  YError(strerror(errno));
}

static void fd_close(chn_t *this)
{
  this->chn_ops = &bogus_chn_ops;
  if (close(this->chn.fd) == -1) posix_error();
}

static long fd_tell(chn_t *this)
{
  long offset = lseek(this->chn.fd, 0, SEEK_CUR);
  if (offset == -1) posix_error();
  return offset;
}

static long fd_seek(chn_t *this, long offset, int whence)
{
  offset = lseek(this->chn.fd, offset, whence);
  if (offset == -1) posix_error();
  return offset;
}

static long fd_read(chn_t *this, void *ptr, long nbytes)
{
  long count = read(this->chn.fd, ptr, nbytes);
  if (count == -1) posix_error();
  return count;
}

static void fd_write(chn_t *this, void *ptr, long nbytes)
{
  long count = write(this->chn.fd, ptr, nbytes);
  if (count != nbytes) {
    if (count == -1) posix_error();
    YError("cannot write all data");
  }
}

/*---------------------------------------------------------------------------*/
/* Operations for temporarily disabled files. */

static void bogus_close(chn_t *this)
{
  /* no-operation */
}

static long bogus_tell(chn_t *this)
{
  YError("cannot tell: corrupted channel");
  return -1;
}

static long bogus_seek(chn_t *this, long offset, int whence)
{
  YError("cannot tell: corrupted channel");
  return -1;
}

static long bogus_read(chn_t *this, void *ptr, long nbytes)
{
  YError("cannot read: corrupted channel");
  return 0;
}

static void bogus_write(chn_t *this, void *ptr, long nbytes)
{
  YError("cannot write: corrupted channel");
}

/*---------------------------------------------------------------------------*/
/* Operations for compressed files. */

#if HAVE_ZLIB
static void gz_error(chn_t *this);
static void gz_error(chn_t *this)
{
  int errnum;
  const char *msg = gzerror(this->chn.gz, &errnum);
  if (errnum == Z_ERRNO) posix_error();
  YError(msg);
}

static void gz_close(chn_t *this)
{
  this->chn_ops = &bogus_chn_ops;
  if (gzclose(this->chn.gz) != Z_OK) gz_error(this);
}

static long gz_tell(chn_t *this)
{
  long offset = gztell(this->chn.gz);
  if (offset == -1) gz_error(this);
  return offset;
}

static long gz_seek(chn_t *this, long offset, int whence)
{
  offset = gzseek(this->chn.gz, offset, whence);
  if (offset == -1) gz_error(this);
  return offset;
}

static long gz_read(chn_t *this, void *ptr, long nbytes)
{
  /* return value is -1 in case of error, 0 for end-of-file */
  long count = gzread(this->chn.gz, ptr, nbytes);
  if (count == -1) gz_error(this);
  return count;
}

static void gz_write(chn_t *this, void *ptr, long nbytes)
{
  long count = gzwrite(this->chn.gz, ptr, nbytes);
  if (count != nbytes) gz_error(this);
}
#endif /* HAVE_ZLIB */

/*---------------------------------------------------------------------------*/
/* BUILTIN FUNCTIONS */

void Y_chn_open(int argc)
{
  int compression = Z_NO_COMPRESSION;
  int strategy = Z_DEFAULT_STRATEGY;
  int i, j, c;
  int access = 0;     /* 'r', 'w', or 'a' */
  int read_write = 0; /* true if '+' character present in mode */
  int binary = 0;     /* true if 'b' character present in mode */
  chn_t *this;
  char *path, *mode, zmode[8];

  if (argc != 2) YError("chn_open takes exactly 2 arguments");
  path = yeti_get_string(sp-1);
  mode = yeti_get_string(sp);

  /* Parse mode string. */
  i = 0;
  c = mode[i++];
  if (c == 'r' || c == 'w' || c == 'a') access = c;
  else YError("first character of MODE must be 'r', 'w', or 'a'");
  c = mode[i++];
  if (c == 'b') {
    binary = 1;
    c = mode[i++];
    if (c == '+') {
      read_write = 1;
      c = mode[i++];
    }
  } else if (c == '+') {
    read_write = 1;
    c = mode[i++];
    if (c == 'b') {
      binary = 1;
      c = mode[i++];
    }
  }
  if (c == 'z') {
    c = mode[i++];
    if (c >= '1' && c <= '9') {
      compression = c - '0';
      c = mode[i++];
    } else {
      compression = Z_DEFAULT_COMPRESSION;
    }
    if (c == 'f') {
      strategy = Z_FILTERED;
      c = mode[i++];
    } else if (c == 'h') {
      strategy = Z_HUFFMAN_ONLY;
      c = mode[i++];
    }
  }
  if (c != 0) {
    YError("bad MODE string");
  }

  /* Create instance. */
  this = p_malloc(sizeof(chn_t));
  memset(this, 0, sizeof(chn_t));
  this->ops = &channelOps;
  this->chn_ops = &bogus_chn_ops;
  PushDataBlock(this);
  this->path = YExpandName(path);

  /* Set instance mode. */
  i = j = 0;
  mode = this->mode;
  mode[i++] = zmode[j++] = access;
  if (binary) mode[i++] = zmode[j++] = 'b';
  if (read_write) mode[i++] = zmode[j++] = '+';
  if (compression != Z_NO_COMPRESSION) {
    mode[i++] = 'z';
    if (compression != Z_DEFAULT_COMPRESSION)
      mode[i++] = zmode[j++] = '0' + compression;
    if (strategy == Z_FILTERED) mode[i++] = zmode[j++] = 'f';
    else if (strategy == Z_HUFFMAN_ONLY) mode[i++] = zmode[j++] = 'h';
  }
  mode[i] = zmode[j] = 0;

  /* Open file. */
  if (compression == Z_NO_COMPRESSION) {
    int flags;
    if (access == 'r') {
      flags = (read_write ? O_RDWR : O_RDONLY);
    } else if (access == 'w') {
      flags = (read_write ? (O_RDWR  |O_TRUNC|O_CREAT)
                          : (O_WRONLY|O_TRUNC|O_CREAT));
    } else /* access == 'a' */ {
      flags = (read_write ? (O_RDWR  |O_CREAT)
                          : (O_WRONLY|O_CREAT));
    }
    if ((this->chn.fd = open(this->path, flags)) == -1 ||
	(access == 'a' && lseek(this->chn.fd, 0, SEEK_END) == -1)) {
      posix_error();
    }
    this->chn_ops = &fd_chn_ops;
  } else {
#if HAVE_ZLIB
    if (access == 'a')
      YError("compressed file cannot be open in append mode");
    if (read_write)
      YError("compressed file cannot be open for reading _and_ writing");
    if (! (this->chn.gz = gzopen(this->path, zmode))) {
      if (errno != 0) posix_error();
      else YError("insufficient memory");
    }
    this->chn_ops = &gz_chn_ops;
#else /* !HAVE_ZLIB */
    YError("no support for reading/writing of compressed files");
#endif /* HAVE_ZLIB */
  }
}

#if 0
void Y_chn_close(int argc)
{
  chn_t *this;
  if (argc != 1) YError("chn_close takes exactly 1 argument");
  this = get_channel(sp);
  ...;
  PushDataBlock(RefNC(&nilDB));
}
#endif

void Y_chn_tell(int argc)
{
  chn_t *this;
  long offset;
  if (argc != 1) YError("chn_tell takes exactly 1 argument");
  this = get_channel(sp);
  offset = CHN_TELL(this);
  PushLongValue(offset);
}

void Y_chn_seek(int argc)
{
  chn_t *this;
  int whence;
  long offset;
  if (argc != 3) YError("chn_seek takes exactly 3 arguments");
  this = get_channel(sp-2);
  offset = yeti_get_integer(sp-1);
  whence = yeti_get_integer(sp);
  if      (whence == 0) whence = SEEK_SET;
  else if (whence == 1) whence = SEEK_CUR;
  else if (whence == 2) whence = SEEK_END;
  else YError("WHENCE must be 0, 1 or 2");
  offset = CHN_SEEK(this, offset, whence);
  PushLongValue(offset);
}

void Y_chn_write(int argc)
{
  int i;
  long nbytes, nbytes_written=0;
  chn_t *this;
  Symbol *stack = sp-argc+1;
  if (argc < 2) YError("chn_write takes at least 2 arguments");
  this = get_channel(stack);
  for (i=1 ; i<argc ; ++i) {
    Symbol *s = stack+i;
    if (s->ops) {
      Operand op;
      s->ops->FormOperand(s, &op);
      if (op.ops->typeID >= T_CHAR && op.ops->typeID <= T_COMPLEX) {
	nbytes = op.type.number*op.type.base->size;
	CHN_WRITE(this, op.value, nbytes);
	nbytes_written += nbytes;
	continue;
      } else if (op.ops->typeID == T_STRING) {
	char **list = op.value;
	long j, number=op.type.number;
	for (j=0 ; j<number ; ++j) {
	  char *str = list[j];
	  if (str && (nbytes = strlen(str)) > 0) {
	    CHN_WRITE(this, str, nbytes);
	    nbytes_written += nbytes;
	  }
	}
	continue;
      }
    }
    YError("only basic array types are supported");
  }
  PushLongValue(nbytes_written);
}

void Y_chn_read(int argc)
{
  Array *array;
  long count, nbytes;
  chn_t *this;
  Symbol *s, *stack = sp-argc+1;

  if (argc < 2) YError("chn_read takes at least 2 arguments");
  this = get_channel(stack);
  s = stack+1;
  if (s->ops == &referenceSym) s = &globTab[s->index];
  if (s->ops == &dataBlockSym && s->value.db->ops == &structDefOps) {
    StructDef *base = (StructDef *)s->value.db;
    if (base->dataOps->typeID < T_CHAR || base->dataOps->typeID > T_COMPLEX)
      YError("only basic data types are supported");
    BuildDimList(stack+2, argc-2);
    array = yeti_push_new_array(base, tmpDims);
    nbytes = array->type.number*array->type.base->size;
    count = CHN_READ(this, array->value.c, nbytes);
    if (count != nbytes) YError(count == 0 ? "end of file" : "short file");
  } else if (argc == 2) {
    nbytes = yeti_get_integer(s);
    if (nbytes <= 0) YError("bad number of bytes");
    Drop(1); /* we may need one more scratch item onto the stack */
    array = yeti_push_new_char_array(yeti_first_dimension(nbytes, 1));
    count = CHN_READ(this, array->value.c, nbytes);
    if (count != nbytes) {
      if (count > 0) {
	char *old = array->value.c;
	array = yeti_push_new_char_array(yeti_first_dimension(count, 1));
	memcpy(array->value.c, old, count);
      } else {
	PushDataBlock(RefNC(&nilDB));
      }
    }
  } else {
    YError("bad arguments");
  }
}

/*---------------------------------------------------------------------------*/
/* FILE ROUTINES */

extern BuiltIn Y_filepath;

/* The following definitions _must_ match those in "ascio.c" */
extern Operations textOps;

struct TextStream {
  int references;      /* reference counter */
  Operations *ops;     /* virtual function table */
#if (YORICK_VERSION_MAJOR == 1) && (YORICK_VERSION_MINOR <= 4)
  FILE *stream;        /* 0 indicates file has been closed */
#else
  p_file *stream;      /* 0 indicates file has been closed */
#endif
  char *fullname;      /* filename after YExpandName */
  int permissions;     /* +1 read permission, +2 write permission
			  +4 append mode, +8 binary mode, +16 pipe */
  /* ------ begin specific text stream part ------- */
  long lastLineRead;   /* 1-origin line number of last line read */
  long readPosition;   /* file position (ftell) after lastLineRead */
  long lastPosition;   /* file position (ftell) before lastLineRead --
			  after backup, lastPosition==readPosition,
			  and lastPosition is not valid */
  int readWrite;       /* 0 initially, 1 after read, 2 after write */
  long fileID;         /* unique number used to recognize this file */
};

void Y_filepath(int argc)
{
  Dimension *dims;
  Operand op;
  char **input, **output;
  long i, n;

  if (argc != 1) YError("filepath function takes exactly one argument");
  op.ops= 0;
  if (sp->ops) sp->ops->FormOperand(sp, &op);
  if (op.ops == &stringOps) {
    input = YGet_Q(sp, 0, &dims);
    n = TotalNumber(dims);
    output = yeti_push_new_string_array(dims)->value.q;
    for (i=0 ; i<n ; ++i) output[i] = input[i] ? YExpandName(input[i]) : 0;
  } else if (op.ops == &streamOps) {
    output = yeti_push_new_string_array(NULL)->value.q;
    output[0] = p_strcpy(((IOStream *)op.value)->fullname);
  } else if (op.ops == &textOps) {
    output = yeti_push_new_string_array(NULL)->value.q;
    output[0] = p_strcpy(((TextStream *)op.value)->fullname);
  } else if (op.ops == &voidOps) {
    PushDataBlock(RefNC(&nilDB));
  } else {
    YError("bad argument: expecting text/binary file or file name(s)");
  }
}

/*---------------------------------------------------------------------------*/
/* PRIVATE (STATIC) FUNCTIONS */

static chn_t *get_channel(Symbol *stack)
{
  DataBlock *db;
  Symbol *sym = (stack->ops == &referenceSym) ? &globTab[stack->index] : stack;
  if (sym->ops != &dataBlockSym || sym->value.db->ops != &channelOps)
    YError("expected i/o channel object");
  db = sym->value.db;
#if 1
  if (sym != stack) {
    /* Replace reference onto the stack (equivalent to the statement
       ReplaceRef(s); see ydata.c for actual code of this routine). */
    stack->value.db = Ref(db);
    stack->ops = &dataBlockSym;     /* change ops only AFTER value updated */
  }
#endif
  return (chn_t *)db;
}

/*---------------------------------------------------------------------------*/
