#include <math.h>
#include "yeti_nnls.h"
#define diff nnls_diff

static void h12(integer_t mode, integer_t lpivot, integer_t l1, integer_t m,
		real_t u[], integer_t iue, real_t *up, real_t c[],
		integer_t ice, integer_t icv, integer_t ncv);
static void g1(real_t a, real_t b, real_t *cterm, real_t *sterm, real_t *sig);

/*C     Last change:  DE   14 Apr 101    9:39 am*/

/*C     SUBROUTINE NNLS  (A,MDA,M,N,B,X,RNORM,W,ZZ,INDEX,MODE)
C
C  Algorithm NNLS: NONNEGATIVE LEAST SQUARES
C
c  The original version of this code was developed by
c  Charles L. Lawson and Richard J. Hanson at Jet Propulsion Laboratory
c  1973 JUN 15, and published in the book
c  "SOLVING LEAST SQUARES PROBLEMS", Prentice-HalL, 1974.
c  Revised FEB 1995 to accompany reprinting of the book by SIAM.
c
C     GIVEN AN M BY N MATRIX, A, AND AN M-VECTOR, B,  COMPUTE AN
C     N-VECTOR, X, THAT SOLVES THE LEAST SQUARES PROBLEM
C
C                      A * X = B  SUBJECT TO X .GE. 0
C     ------------------------------------------------------------------
c                     Subroutine Arguments
c
C     A(),MDA,M,N     MDA IS THE FIRST DIMENSIONING PARAMETER FOR THE
C                     ARRAY, A().   ON ENTRY A() CONTAINS THE M BY N
C                     MATRIX, A.           ON EXIT A() CONTAINS
C                     THE PRODUCT MATRIX, Q*A , WHERE Q IS AN
C                     M BY M ORTHOGONAL MATRIX GENERATED IMPLICITLY BY
C                     THIS SUBROUTINE.
C     B()     ON ENTRY B() CONTAINS THE M-VECTOR, B.   ON EXIT B() CON-
C             TAINS Q*B.
C     X()     ON ENTRY X() NEED NOT BE INITIALIZED.  ON EXIT X() WILL
C             CONTAIN THE SOLUTION VECTOR.
C     RNORM   ON EXIT RNORM CONTAINS THE EUCLIDEAN NORM OF THE
C             RESIDUAL VECTOR.
C     W()     AN N-ARRAY OF WORKING SPACE.  ON EXIT W() WILL CONTAIN
C             THE DUAL SOLUTION VECTOR.   W WILL SATISFY W(I) = 0.
C             FOR ALL I IN SET P  AND W(I) .LE. 0. FOR ALL I IN SET Z
C     ZZ()     AN M-ARRAY OF WORKING SPACE.
C     INDEX()     AN INTEGER WORKING ARRAY OF LENGTH AT LEAST N.
C                 ON EXIT THE CONTENTS OF THIS ARRAY DEFINE THE SETS
C                 P AND Z AS FOLLOWS..
C
C                 INDEX(1)   THRU INDEX(NSETP) = SET P.
C                 INDEX(IZ1) THRU INDEX(IZ2)   = SET Z.
C                 IZ1 = NSETP + 1 = NPP1
C                 IZ2 = N
C     MODE    THIS IS A SUCCESS-FAILURE FLAG WITH THE FOLLOWING
C             MEANINGS.
C             1     THE SOLUTION HAS BEEN COMPUTED SUCCESSFULLY.
C             2     THE DIMENSIONS OF THE PROBLEM ARE BAD.
C                   EITHER M .LE. 0 OR N .LE. 0.
C             3    ITERATION COUNT EXCEEDED.  MORE THAN 3*N ITERATIONS.
C
C     ------------------------------------------------------------------*/

void nnls(real_t a[], integer_t mda, integer_t m, integer_t n, real_t b[],
	  real_t x[], real_t *rnorm, real_t w[], real_t zz[],
	  integer_t index[], integer_t *mode)
     /* integer INDEX(N)
	real A(MDA,N), B(M), W(N), X(N), ZZ(M)
     */
{
  const real_t factor = 0.01;
  const real_t two = 2.0;
  const real_t zero = 0.0;
  integer_t i, ii, ip, iter, itmax, iz, iz1, iz2, izmax, j, jj, jz, l;
  integer_t npp1, nsetp, rtnkey;
  real_t alpha, asave, cc;
  real_t sm, ss, t, temp, unorm, up, wmax, ztest;

  /* Adjustements and definitions to make C arrays look like FORTRAN arrays. */
#undef index_1
#define index_1(a1) index[a1-1]
#undef zz_1
#define zz_1(a1) zz[a1-1]
#undef x_1
#define x_1(a1) x[a1-1]
#undef w_1
#define w_1(a1) w[a1-1]
#undef b_1
#define b_1(a1) b[a1-1]
#undef a_2
#define a_2(a1,a2) a[a1-1+mda*(a2-1)]

  /*------------------------------------------------------------------*/
  if (m <= 0 || n <= 0) {
    *mode = 2;
    return;
  }
  *mode = 1;
  iter = 0;
  itmax = 3*n;

  /* Initialize the arrays INDEX() and X(). */
  for (i=1 ; i<=n ; ++i) {
    x_1(i) = zero;
    index_1(i) = i;
  }
  iz2 = n;
  iz1 = 1;
  nsetp = 0;
  npp1 = 1;

  /*******  MAIN LOOP BEGINS HERE  *******/
 L_30:
  /* Quit if all coefficients are already in the solution or if M columns
     of A have been triangularized. */
  if (iz1 > iz2 || nsetp >= m) goto L_350;
  /* Compute components of the dual (negative gradient) vector W(). */
  for (iz=iz1 ; iz<=iz2 ; ++iz) {
    j = index_1(iz);
    sm = zero;
    for (l=npp1 ; l<=m ; ++l) {
      sm += a_2(l,j)*b_1(l);
    }
    w_1(j) = sm;
  }

  /* Find largest positive W(J). */
 L_60:
  wmax = zero;
#warning FIXME
  izmax = 0; /* avoid compiler warning about unitialized variable */
  for (iz=iz1 ; iz<=iz2 ; ++iz) {
    j = index_1(iz);
    if (w_1(j) > wmax) {
      wmax = w_1(j);
      izmax = iz;
    }
  }

  /* If WMAX <= 0, goto termination.  This indicates satisfaction of the
     Kuhn-Tucker conditions. */
  if (wmax <= zero) goto L_350;
  iz = izmax;
  j = index_1(iz);

  /* The sign of W(J) is ok for J to be moved to set P.  Begin the
     transformation and check new diagonal element to avoid near linear
     dependence. */
  asave = a_2(npp1,j);
  h12(1, npp1, npp1+1, m, &a_2(1,j), 1, &up, (real_t *)0, 1, 1, 0);
  unorm = zero;
  if (nsetp != 0) {
    for (l=1 ; l<=nsetp ; ++l) {
#warning FIXME
      real_t tmp = a_2(l,j);
      unorm += tmp*tmp;
    }
    unorm = sqrt(unorm);
  }
  if (diff(unorm + fabs(a_2(npp1,j))*factor, unorm) > zero) {
    /* Column J is sufficiently independent.  Copy B into ZZ, update ZZ
       and solve for ZTEST = proposed new value for X(J). */
#warning "use memcpy"
    for (l=1 ; l<=m ; ++l) zz_1(l) = b_1(l);
    h12(2, npp1, npp1+1, m, &a_2(1,j), 1, &up, zz, 1, 1, 1);
    ztest = zz_1(npp1)/a_2(npp1,j);
    /* See if ZTEST is positive. */
    if (ztest > zero) goto L_140;
  }
  /* Reject J as a candidate to be moved from set Z to set P.  Restore
     A(NPP1,J), set W(J)=ZERO, and loop back to test dual coefficients
     again. */
  a_2(npp1,j) = asave;
  w_1(j) = zero;
  goto L_60;
  /* The index J=INDEX(IZ) has been selected to be moved from set Z to set
     P.  Update B, update indices, apply Householder transformations to
     columns in new set Z, zero subdiagonal elements in column J, set
     W(J)=ZERO.  */
 L_140:
#warning "use memcpy"
  for (l=1 ; l<=m ; ++l) b_1(l) = zz_1(l);
  index_1(iz) = index_1(iz1);
  index_1(iz1) = j;
  ++iz1;
  nsetp = npp1;
  ++npp1;
  if (iz1 <= iz2) {
    for (jz=iz1 ; jz<=iz2 ; ++jz) {
      jj = index_1(jz);
      h12(2, nsetp, npp1, m, &a_2(1,j), 1, &up, &a_2(1,jj), 1, mda, 1);
    }
  }
  if (nsetp != m) {
#warning "FIXME: NPP1 = NSETP+1 always?"
    for (l=npp1 ; l<=m ; ++l) a_2(l,j) = zero;
  }
  w_1(j)=zero;

  /* Solve the triangular system.
     Store the solution temporarily in ZZ(). */
  rtnkey = 1;
  goto L_400;
 L_200:

  /*******  SECONDARY LOOP BEGINS HERE *******/

  /* Iteration counter. */
 L_210:
  if (++iter > itmax) {
    /* Too many iterations. */
    *mode = 3;
    goto L_350;
  }

  /* See if all new constrained coefficients are feasible.
     If not, compute ALPHA.  */
  alpha = two;
  for (ip=1 ; ip<=nsetp ; ++ip) {
    l = index_1(ip);
    if (zz_1(ip) <= zero) {
#warning "local tmp var T"
      t = x_1(l)/(x_1(l) - zz_1(ip));
      if (alpha > t) {
	alpha = t;
	jj = ip;
      }
    }
  }

  /* If all new constrained coefficients are feasible then ALPHA will
     still be equal to 2.  If so, exit from secondary loop to main loop. */
  if (alpha == two) goto L_330;

  /* Otherwise use ALPHA which will be between 0 and 1 to interpolate
     between the old X and the new ZZ. */
  for (ip=1 ; ip<=nsetp ; ++ip) {
    l = index_1(ip);
    x_1(l) += alpha*(zz_1(ip) - x_1(l));
  }

  /* Modify A and B and the index arrays to move I-th coefficient
     from set P to set Z. */
  i = index_1(jj);
 L_260:
  x_1(i) = zero;
  if (jj != nsetp) {
    ++jj;
    for (j=jj ; j<=nsetp ; ++j) {
      ii = index_1(j);
      index_1(j-1) = ii;
#warning "FIXME local var CC SS"
      g1(a_2(j-1,ii), a_2(j,ii), &cc, &ss, &a_2(j-1,ii));
      a_2(j,ii) = zero;
      for (l=1 ; l<=n ; ++l) {
	if (l != ii) {
	  /* Apply procedure G2(CC,SS,A(J-1,L),A(J,L)) */
#warning "use 2 temp. vars"
	  temp = a_2(j-1,l);
	  a_2(j-1,l) = ss*a_2(j,l) + cc*temp;
	  a_2(j,l)   = cc*a_2(j,l) - ss*temp;
	}
      }
      /* Apply procedure G2(CC,SS,B(J-1),B(J)) */
#warning "use 2 temp. vars"
      temp = b_1(j-1);
      b_1(j-1) = ss*b_1(j) + cc*temp;
      b_1(j)   = cc*b_1(j) - ss*temp;
    }
  }

#define G2(type_t, cs, sn, var1, var2) {	\
    type_t t1 = var1, t2 = var2;		\
    var1 = cs*t1 + sn*t2;			\
    var2 = cs*t2 - sn*t1;			\
  }

  npp1 = nsetp;
  --nsetp;
  --iz1;
  index_1(iz1) = i;
  /* See if the remaining coefficient in set P are feasible.  They should
     be because of the way ALPHA was determined.  If any are infeasible it
     is due to round-off error.  Any that are nonpositive will be set to
     zero and moved from set P to set Z. */
  for (jj=1 ; jj<=nsetp ; ++jj) {
    i = index_1(jj);
    if (x_1(i) <= zero) goto L_260;
  }

  /* Copy B( ) into ZZ( ).  Then solve again and loop back. */
#warning "use memcpy"
  for (i=1 ; i<=m ; ++i) zz_1(i) = b_1(i);
  rtnkey = 2;
  goto L_400;
 L_320:
  goto L_210;
  /*******  END OF SECONDARY LOOP  *******/
 L_330:
  for (ip=1 ; ip<=nsetp ; ++ip) {
    i = index_1(ip);
    x_1(i) = zz_1(ip);
  }
  /* All new coefficient are positive.  Loop back to beginning. */
  goto L_30;
  /*******  END OF MAIN LOOP  *******/
  /* Come to here for termination.  Compute the norm of the final residual
     vector. */
 L_350:
  sm = zero;
  if (npp1 <= m) {
    for (i=npp1 ; i<=m ; ++i) {
      real_t b_i = b_1(i);
      sm += b_i*b_i;
    }
  } else {
#warning "use memset"
    for (j=1 ; j<=n ; ++j) w_1(j) = zero;
  }
  *rnorm = sqrt(sm);
  return;

  /* The following block of code is used as an internal subroutine to
     solve the triangular system, putting the solution in ZZ(). */
 L_400:
  for (l=1 ; l<=nsetp ; ++l) {
    ip = nsetp + 1 - l;
    if (l != 1) {
      real_t zz_ipp1 = zz_1(ip+1);
      for (ii=1 ; ii<=ip ; ++ii) zz_1(ii) -= a_2(ii,jj)*zz_ipp1;
    }
    jj = index_1(ip);
    zz_1(ip) /= a_2(ip,jj);
  }
  if (rtnkey == 1) goto L_200;
  goto L_320;
}

/*     SUBROUTINE H12(MODE,LPIVOT,L1,M,U,IUE,UP,C,ICE,ICV,NCV)
 *
 * Construction and/or application of a single
 * Householder transformation:     Q = I + U*(U**T)/B
 *
 * The original version of this code was developed by Charles L. Lawson
 * and Richard J. Hanson at Jet Propulsion Laboratory 1973 JUN 12, and
 * published in the book "SOLVING LEAST SQUARES PROBLEMS", Prentice-HalL,
 * 1974.  Revised FEB 1995 to accompany reprinting of the book by SIAM.
 *
 *    ------------------------------------------------------------------
 *		      Subroutine Arguments
 *
 *    MODE   = 1 OR 2   Selects Algorithm H1 to construct and apply a
 *	     Householder transformation, or Algorithm H2 to apply a
 *	     previously constructed transformation.
 *    LPIVOT is the index of the pivot element.
 *    L1,M   If L1 .LE. M,  the transformation will be constructed to
 *	     zero elements indexed from L1 through M.   If L1 GT. M
 *	     the subroutine does an identity transformation.
 *    U(),IUE,UP    On entry with MODE = 1, U() contains the pivot
 *	     vector.  IUE is the storage increment between elements.
 *	     On exit when MODE = 1, U() and UP contain quantities
 *	     defining the vector U of the Householder transformation.
 *	     on entry with MODE = 2, U() and UP should contain
 *	     quantities previously computed with MODE = 1.  These will
 *	     not be modified during the entry with MODE = 2.
 *    C()    On entry with MODE = 1 or 2, C() contains a matrix which
 *	     will be regarded as a set of vectors to which the
 *	     Householder transformation is to be applied.  Nothing is
 *           done with C() if NCV = 0. 
 *	     On exit C() contains the set of transformed vectors.
 *    ICE    Storage increment between elements of vectors in C().
 *    ICV    Storage increment between vectors in C().
 *    NCV    Number of vectors in C() to be transformed. If NCV <= 0,
 *	     no operations will be done on C().
 *    ------------------------------------------------------------------*/

static void h12(integer_t mode, integer_t lpivot, integer_t l1, integer_t m,
		real_t u[], integer_t iue, real_t *up, real_t c[],
		integer_t ice, integer_t icv, integer_t ncv)
{
  const real_t one = 1.0e0;
  const real_t zero = 0.0e0;
  integer_t i, i2, i3, i4, incr, j;
  real_t tmp, b, cl, clinv, sm;
#warning "optimize indices (always U(1,J))"
#warning "use double for sums"
#undef c_1
#define c_1(a1) c[a1-1]
  /* REAL U(IUE,M) */
#undef u_2
#define u_2(a1,a2) u[a1-1+iue*(a2-1)]

  if (0 >= lpivot || lpivot >= l1 || l1 > m) return;
  cl = fabs(u_2(1,lpivot));
  if (mode == 2) {
    if (cl <= zero) return;
  } else {
    /***
     *** Construct the transformation.
     ***/
    for (j=l1 ; j<=m ; ++j) {
      if ((tmp = fabs(u_2(1,j))) > cl) cl = tmp;
    }
    if (cl <= zero) return;
    clinv = one/cl;
    tmp = u_2(1,lpivot)*clinv;
    sm = tmp*tmp;
    for (j=l1 ; j<=m ; ++j) {
      tmp = u_2(1,j)*clinv;
      sm += tmp*tmp;
    }
    cl *= sqrt(sm);
    if (u_2(1,lpivot) > zero) cl = -cl;
    *up = u_2(1,lpivot) - cl;
    u_2(1,lpivot) = cl;
  }

  /***
   *** Apply the transformation  I+U*(U**T)/B  to  C.
   ***/
  if (ncv <= 0) return;
  /* B must be non-positive here, otherwise, return. */
  if ((b = (*up)*u_2(1,lpivot)) >= zero) return;
  b = one/b;
  i2 = 1 - icv + ice*(lpivot - 1);
  incr = ice*(l1 - lpivot);
  for (j=1 ; j<=ncv ; ++j) {
    i2 += icv;
    i4 = i3 = i2 + incr;
    sm = c_1(i2)*(*up);
    for (i=l1 ; i<=m ; ++i) {
      sm += c_1(i3)*u_2(1,i);
      i3 += ice;
    }
    if (sm != zero) {
      sm *= b;
      c_1(i2) += sm*(*up);
      for (i=l1 ; i<=m ; ++i) {
	c_1(i4) += sm*u_2(1,i);
	i4 += ice;
      }
    }
  }
}

/*
 *    COMPUTE ORTHOGONAL ROTATION MATRIX..
 *
 * The original version of this code was developed by Charles L. Lawson and
 * Richard J. Hanson at Jet Propulsion Laboratory 1973 JUN 12, and published
 * in the book "SOLVING LEAST SQUARES PROBLEMS", Prentice-Hall, 1974.
 * Revised FEB 1995 to accompany reprinting of the book by SIAM.
 *
 *    COMPUTE.. MATRIX   (C, S) SO THAT (C, S)(A) = (SQRT(A**2+B**2))
 *			 (-S,C)         (-S,C)(B)   (   0          )
 *    COMPUTE SIG = SQRT(A**2+B**2)
 *	 SIG IS COMPUTED LAST TO ALLOW FOR THE POSSIBILITY THAT
 *	 SIG MAY BE IN THE SAME LOCATION AS A OR B .
 *------------------------------------------------------------------ */
static void g1(real_t a, real_t b, real_t *cterm, real_t *sterm, real_t *sig)
{
  const real_t one = 1.0e0;
  const real_t zero = 0.0e0;
  real_t  xr, yr, abs_a, abs_b;

  if ((abs_a = fabs(a)) > (abs_b = fabs(b))) {
    xr = b/a;
    yr = sqrt(one + xr*xr);
    *cterm = (a >= zero ? one/yr : -one/yr);
    *sterm = (*cterm)*xr;
    *sig = abs_a*yr;
  } else if (b != zero) {
    xr = a/b;
    yr = sqrt(one + xr*xr);
    *sterm = (b >= zero ? one/yr : -one/yr);
    *cterm = (*sterm)*xr;
    *sig = abs_b*yr;
  } else {
    *sig = zero;
    *cterm = zero;
    *sterm = one;
  }
}

