
/* load dynamic code with builtin functions (before any reshape operations) */
if (is_func(load_plugin)) load_plugin;

extern __nnls;
/* PROTOTYPE
     void nnls(double array a, long mda, long m, long n, double array b,
               double array x, double array rnorm, double array w,
	       double array zz, long array index, long array mode);
*/
func nnls(a, b, &rnorm)
/* DOCUMENT nnls(a, b)
       -or- nnls(a, b, rnorm)
     Compute the non-negative least squares solution X to the linear
     problem:
       A(,+)*X(+) = B
     the result is:
       arg min sum((A(,+)*X(+) - B)^2)   subject to X(i) >= 0 for all i
*/
{
  if (! is_array(a) || structof(a) == complex || (dims = dimsof(a))(1) != 2)
    error, "A must be a non-complex matrix";
  lda = m = dims(2);
  n = dims(3);
  if (! is_array(b) || structof(b) == complex || (dims = dimsof(b))(1) != 1)
    error, "B must be a non-complex vector";
  if (dims(2) != m) error, "non conformable dimensions";
  a = double(a); /* convert to double and make private copy */
  b = double(b); /* convert to double and make private copy */
  rnorm = 0.0;
  x = w = array(double, n);
  z = array(double, m);
  index = array(long, n);
  mode = 0;
  __nnls, a, m, m, n, b, x, rnorm, w, z, index, mode;
  if (mode != 1) {
    if (mode == 3) error, "iteration count exceeded";
    error, "unexpected result (BUG)";
  }
  return x;
}

func psf_as_matrix(psf)
{
  /* build the matrix A */
  if (! is_array(psf) || dimsof(psf)(1) != 1) error, "expecting vector PSF";
  n = numberof(psf);
  a = array(double, n, n);
  a(,1) = psf;
  for (i=1 ; i<n ; ++i) {
    a(i+1:n,i+1) = psf(1:n-i);
    a(1:i,i+1) = psf(n-i+1:n);
  }
  return a;
}

func nnls_deconv(data, psf, &rnorm)
{
  a = psf_as_matrix(psf);
  return nnls(a, data, rnorm);
}

func genpsf(x, y, wx, wy, rhox, rhoy, p)
{
  if (is_void(y)) {
    psf = 1.0/(1.0 + abs(x/wx)^(p*(1.0 + abs(x/rhox))));
  } else {
    psf = 1.0/(1.0 + abs(x/wx, y/wy)^(p*(1.0 + abs(x/rhox, y/rhoy))));
    
  }
  return psf/sum(psf);
}

func nnls_vmlmb(a, b, x0, frtol=, fatol=, verb=, maxiter=, maxeval=)
{
  require, "Eric/optim_pack.i";
  if (! is_array(a) || structof(a) == complex || (dims = dimsof(a))(1) != 2)
    error, "A must be a non-complex matrix";
  lda = m = dims(2);
  n = dims(3);
  if (! is_array(b) || structof(b) == complex || (dims = dimsof(b))(1) != 1)
    error, "B must be a non-complex vector";
  if (dims(2) != m) error, "non conformable dimensions";
  a = double(a); /* convert to double and make private copy */
  b = double(b); /* convert to double and make private copy */
  x = (is_void(x0) ? array(double, n) : double(x0)); 
  
  task = "start";
  if (is_void(frtol)) frtol = 1e-8;
  if (is_void(fatol)) fatol = 0.0;
  if (is_void(ndirs)) ndirs = 5;
  ws = optim_vmlm_setup(ndirs, fmin=0.0, fatol=fatol, frtol=frtol, prj=1);
  msg = string(0);
  if (verb) {
    elapsed = array(double, 3);
    timer, elapsed;
    cpu_start = elapsed(1);
  }
  eval = iter = 0;
  stp = 0.0;
  for (;;) {
    /* Setup STAGE according to TASK. */
    if (task == "fg") {
      stage = 1;
    } else if (task == "start") {
      stage = 0;
    } else if ((ident = strpart(task, 1:4)) == "newx") {
      stage = 2;
    } else if (ident == "conv") {
      stage = 3;
    } else {
      stage = 4;
    }

    if (stage <= 1) {
      /* Compute function value and gradient at X
         (TASK is "fg" or "start"). */
      x = max(x, 0.0);
      ax = a(,+)*x(+);
      r = ax - b;
      gx = a(+,)*r(+);
      fx = 0.5*sum(r*r);
      bounded = (x <= 0.0)&(gx >= 0.0);
      ++eval;
    }

    /* Unless TASK is "fg", decide to continue/stop and/or
       display information. */
    if (stage != 1) {
      if (stage == 2 || stage == 3) ++iter;
      if (! (stop = (stage >= 3))) {
        if ((stop = (! is_void(maxiter) && iter > maxiter))) {
          msg = swrite(format="warning: too many iterations (%d)", iter);
        } else if ((stop = (! is_void(maxeval) && eval > maxeval))) {
          msg = swrite(format="warning: too many function evaluations (%d)",
                       eval);
        }
      }
      if (verb) {
        if (eval == 1) {
          write, format="%s  %s\n%s  %s\n",
            " ITER    EVAL     CPU [s]            FUNC             GNORM   ",
            " STEPLEN",
            "------  ------  ----------  -----------------------  ---------",
            "---------";
        }
        if (stop || iter%verb == 0) {
          timer, elapsed;
          cpu = elapsed(1) - cpu_start;
          write, format="%6d  %6d %11.2e  %23.16e  %9.1e  %9.1e\n",
            iter, eval, cpu, fx, optim_nrm2(gx), stp;
        }
      }
      if (stop) {
        if (msg) write, format="%s\n", msg;
        return x;
      }
    }
    
    /* Call optimizer. */
    optim_vmlm, x, fx, gx, task, ws, bounded;
    stp = optim_vmlm_step(ws);
  }
}
