/*                                                                      -*-C-*-
 * yeti_plugin.c --
 *
 *	Dynamic loading of code for Yeti.
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 2004, Eric Thibaut (Observatoire de Lyon).
 *
 *	This file is part of Yeti.
 *
 *	Yeti is  free software;  you can redistribute  it and/or  modify it
 *	under the terms  of the GNU General Public  License as published by
 *	the Free Software  Foundation; either version 2 of  the License, or
 *	(at your option) any later version.
 *
 *	Yeti is distributed in the hope that it will be useful, but WITHOUT
 *	ANY WARRANTY; without even  the implied warranty of MERCHANTABILITY
 *	or FITNESS  FOR A PARTICULAR  PURPOSE.  See the GNU  General Public
 *	License for more details.
 *
 *	You should have  received a copy of the  GNU General Public License
 *	along with  Yeti (file "COPYING"  in the top source  directory); if
 *	not, write to the Free  Software Foundation, Inc., 59 Temple Place,
 *	Suite 330,  Boston, MA  02111-1307 USA.
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id: yeti_plugin.c,v 1.1 2005/05/24 13:25:42 eric Exp $
 *	$Log: yeti_plugin.c,v $
 *	Revision 1.1  2005/05/24 13:25:42  eric
 *	Initial revision
 *
 */

#include "ydata.h"

/* Public symbols exported by this file. */
extern BuiltIn Y_load_plugin, Y_debug_plugin;

#ifndef HAVE_DL
# define HAVE_DL 0
#endif

#if !HAVE_DL
# error "no support for dynamic loading in this version of Yeti"
#endif /* not HAVE_DL */

#ifndef __WINDOWS__
#  ifdef __WIN32__
#    define __WINDOWS__
#  endif
#endif

#include <sys/types.h>
#include <pwd.h>
#include <errno.h>
#include <unistd.h>
#include <dlfcn.h>
#include <string.h>
#include <stdio.h>
#include "pstdlib.h"
#include "play.h"
#include "yio.h"

#ifndef DEBUG
# define DEBUG 0
#endif
#define DIR_SEP '/'
#ifndef DL_SUFFIX
# define DL_SUFFIX ".so"
#endif

#define DLOPEN_FLAGS (RTLD_NOW /*RTLD_LAZY*/ | RTLD_GLOBAL)

#define offset_of(class, member) ((char*)&(((class*)0)->member) - (char*)0)

/* Private stuff. */
typedef struct plugin plugin_t;
struct plugin {
  void *handle;        /* opaque handle to dynamic library */
  plugin_t *next;      /* next loaded library */
  char *path;          /* filename of the dynamic code */
  const char **name;   /* array of symbol names */
  unsigned int nfuncs; /* number of builtin functions */
  unsigned int nvars;  /* number of global variables */
  DataBlock *db[1];    /* array of data blocks for builtin functions and
			  global variables referenced by this plugin;
			  actual size is adjuted when allocated; must be
			  _last_ member */
};

static plugin_t *first_plugin = (plugin_t *)0;
static plugin_t *old_plugin = (plugin_t *)0;
static BuiltIn safe_func;
static int     safe_var = -1;
static void cleanup(void);
static void report_error(const char *text, const char *dlfunc,
			 const char *dlerr);

void Y_debug_plugin(int argc)
{
  plugin_t *plugin;
  unsigned int i, n;

  for (plugin=first_plugin ; plugin ; plugin=plugin->next) {
    if (plugin != first_plugin) {
      fprintf(stderr, "\n");
    }
    fprintf(stderr, "plugin path: %s\n",
	    (plugin->path ? plugin->path : "<NULL>"));
    fprintf(stderr, "plugin handle: 0x%lx\n", (unsigned long)plugin->handle);
    n = plugin->nfuncs + plugin->nvars;
    for (i=0 ; i<n ; ++i) {
      if (plugin->db[i]) {
	fprintf(stderr, "plugin %s %s has %d reference%s\n",
		(i<plugin->nfuncs ? "function" : "variable"),
		(plugin->name[i] ? plugin->name[i] : "<NULL>"),
		plugin->db[i]->references,
		(plugin->db[i]->references > 1 ? "s" : ""));
      }
    }
  }
  PushDataBlock(RefNC(&nilDB));
}

void Y_load_plugin(int argc)
{
  plugin_t *plugin, *previous;
  void *handle;
  const char *errmsg, *home;
  char *path, *buf;
  unsigned int i, n, nfuncs, nvars, nbytes, len, len1, len2, len3, offset;
  int replace_extension;
  BuiltIn **builtin = (BuiltIn **)0;
  void **global = (void **)0;
  const char **name = (const char **)0;
  const char *suffix = DL_SUFFIX;

  /* Release ressources provided by an old plugin (i.e. because of an
     interrupt), if any. */
  if (old_plugin) cleanup();

  /* Get name of dynamic library file to load (we want an absolute path to
     avoid confusion about which shared object is loaded). */
  if (argc > 1) YError("load_plugin takes zero or one argument");
  if (argc == 0) {
    if (nYpIncludes <= 0 || ! ypIncludes[nYpIncludes-1].filename) {
      YError("not parsing a source file");
    }
    path = ypIncludes[nYpIncludes-1].filename;
    replace_extension = 1; /* will replace the '.i' extension by SUFFIX */
    len3 = strlen(suffix);
  } else {
    path = YGetString(sp);
    replace_extension = 0;
    len3 = 0;
  }
  len = strlen(path);
  home = (const char *)0;
  offset = 0;
  if (path[0] == '~') {
    /* Deal with '~' and '~USER' notation. */
    struct passwd *pw;
    while (++offset < len) {
      if (path[offset] == DIR_SEP) break;
    }
    if (offset == 1) {
      if (! (home = getenv("HOME"))) {
        pw = getpwuid(getuid());
        if (pw) home = pw->pw_dir;
      }
    } else {
      if (path[offset]) {
	buf = p_malloc(offset);
	memcpy(buf, path + 1, offset - 1);
	buf[offset - 1] = '\0';
      } else {
	buf = &path[1];
      }
      pw = getpwnam(buf);
      if (pw) home = pw->pw_dir;
      if (buf != &path[1]) p_free(buf);
    }
  }
  if (home) {
    len1 = strlen(home);
    len2 = len - offset;
    path += offset;
  } else {
    len1 = 0;
    len2 = len;
  }
  /* At this stage:
   *   LEN  is the length of the original name
   *   LEN1 is the length of the head part if any (in HOME),
   *   LEN2 is the length of the tail part (in PATH) including a leading '/'
   *   LEN3 is the length of SUFFIX, if extension must be replaced;
   *        otherwise LEN3=0.
   */
  if (replace_extension) {
    if (len2 < 2 || path[len2 - 1] != 'i' || path[len2 - 2] != '.') {
      YError("expecting \".i\" extension");
    }
    len2 -= 2; /* now LEN2 is the length of the tail part without the
		  extension -- 2 = strlen(".i") */
  }
  len = len1 + len2 + len3; /* now LEN is the length of the path to load */
  buf = p_malloc(len + 1);
  if (len1) memcpy(buf, home, len1);
  if (len2) memcpy(buf + len1, path, len2);
  if (len3) memcpy(buf + len1 + len2, suffix, len3);
  buf[len] = '\0';
  path = buf; /* now PATH is a dynamic string */
  Drop(argc);
  if (! CalledAsSubroutine()) PushDataBlock(RefNC(&nilDB));

  /* The function dlopen() returns the _same_ handle for a dynamic library
     already loaded even if the file has changed (at least in Linux
     implementation).  This is used to detect that we are attempting to
     reload a plugin.  If same plugin is already loaded, we must replace
     the compiled function pointer of every builtin function provided by
     the previously loaded plugin to that of "safe_func" (to avoid further
     reference to a non-existing compiled function). Note that dlclose must
     be called onto the handle as many times as dlopen was called for the
     same plugin (twice in this case).  Cleanup of previous plugin must be
     done in several pass to be more robust with respect to interrupts, to
     completely avoid further reference to compiled symbols that no longer
     exist and to limit warnings about freeing builtin functions. */

  /* Open dynamic library. */
#if DEBUG
  fprintf(stderr, "dynamic library path: \"%s\"\n", path);
#endif
  handle = dlopen(path, DLOPEN_FLAGS);
#if DEBUG
  fprintf(stderr, "plugin: 0x%x\n", (unsigned)handle);
#endif
  if (! handle) {
    p_free(path);
    report_error("cannot load plugin", "dlopen", dlerror());
  }

  /* Search for a plugin with the same handle. */
  previous = (plugin_t *)0;
  for (plugin=first_plugin ; plugin ; plugin=plugin->next) {
    if (plugin->handle == handle) break;
    previous = plugin;
  }
  if (plugin) {    
    /* Substitute address of a permanent function for every
       built-in functions provided by the existing plugin. */
    n = plugin->nfuncs;
    for (i=0 ; i<n ; ++i) {
      DataBlock *db = plugin->db[i];
      if (db) {
	((BIFunction *)db)->function = safe_func;
      }
    }

    /* Substitute address of a permanent L-value for every
       global variables provided by the existing plugin. */
    n = plugin->nfuncs + plugin->nvars;
    for (i=plugin->nfuncs ; i<n ; ++i) {
      DataBlock *db = plugin->db[i];
      if (db) {
#warning "lv->owner and lv->strider"
	LValue *lv = (LValue *)db;
	StructDef *base = lv->type.base;
	Dimension *dims = lv->type.dims;
	Strider *strider = lv->strider;
	lv->type.base = Ref(&intStruct); /* BEFORE Unref */
	lv->type.dims = (Dimension *)0; /* BEFORE FreeDimension */
	lv->type.number = 1;
	lv->strider = (Strider *)0;
	lv->address.m = (char *)&safe_var;
	if (dims) FreeDimension(dims);
	if (strider) FreeStrider(strider);
	Unref(base);
      }
    }

    /* Remove the plugin from the chain of loaded plugins and save address
       of old plugin for further cleanup (in case of interrupts). */
    if (previous) {
      previous->next = plugin->next;
    } else {
      first_plugin = plugin->next;
    }
    old_plugin = plugin;
    
    /* Remove two references of the shared object handle, then
       reload the DLL. */
    if (dlclose(handle) || dlclose(handle)) {
      p_free(path);
      report_error("cannot unload plugin", "dlclose", dlerror());
    }
    handle = dlopen(path, DLOPEN_FLAGS);
#if DEBUG
    fprintf(stderr, "plugin: 0x%x\n", (unsigned)handle);
#endif
    if (! handle) {
      p_free(path);
      report_error("cannot reload plugin", "dlopen", dlerror());
    }
  }

  /* Get tables of builtin functions and global variables provided by the
     plugin and count them. */
  (void)dlerror(); /* clear error state */
  if (! (builtin = dlsym(handle, "yplgBuiltIns")) ||
      ! (global  = dlsym(handle, "yplgGlobals")) ||
      ! (name    = dlsym(handle, "yplgLinkNames"))) {
    errmsg = dlerror();
    dlclose(handle);
    p_free(path);
    report_error("invalid plugin", (errmsg ? "dlsym" : 0), errmsg);
  }
  nfuncs = -1;
  while (builtin[++nfuncs])
    ;
  nvars = -1;
  while (global[++nvars])
    ;
  n = nfuncs + nvars;

  /* Allocate clean plugin descriptor. */
  nbytes = offset_of(plugin_t, db) + n*sizeof(DataBlock*);
  plugin = p_malloc(nbytes);
  memset(plugin, 0, nbytes);
  plugin->handle = handle;
  plugin->path = path;
  plugin->name = name;
  plugin->nfuncs = nfuncs;
  plugin->nvars = nvars;
  plugin->next = first_plugin;
  first_plugin = plugin;

  /* Link Yorick symbols in the global table to the dynamic code. */ 
  for (i=0 ; i<n ; ++i) {
    long index = Globalize(name[i], 0L);
    volatile Symbol *sym = &globTab[index];
    DataBlock *db = (sym->ops == &dataBlockSym ? sym->value.db : 0);
    if (i < nfuncs) {
      /* Install built-in function. */
      if (db && db->ops == &builtinOps) {
	/* Re-use existing builtin-function object.  Any references to the
	   previous function will use the new one.  This is an advantage if
	   the old function no longer exists because it was defined into
	   the unloaded plugin.  This also avoids spurious warning messages
	   about freeing builtin function. */
	((BIFunction *)db)->function = builtin[i];
      } else {
	/* Allocate new builtin-function object and unreference existing
	   data block. */
	sym->value.db = (DataBlock *)NewBIFunction(builtin[i], index);
	sym->ops = &dataBlockSym;
	Unref(db);
      }
    } else {
      /* Install global variable. Note (from std0.c): everything starts out
	 as a scalar char at the correct address -- the reshape operation
	 when the associated package include file is read (below) will set
	 the correct data type and dimension list */
      sym->value.db = (DataBlock *)NewLValueM((Array *)0, global[i - nfuncs],
					      &charStruct, (Dimension *)0);
      sym->ops = &dataBlockSym;
      Unref(db);
    }

    /* The plugin object itself sets a reference to the DataBlock.  This is
       required to make sure that the DataBlock still exists in any case
       until the plugin is eventually unloaded.  Because when the plugin
       get unloaded we replace the linked functions or variables by a fake
       (but safe) address so that further usage of these symbols will not
       break the interpreter. */
    db = sym->value.db;
    plugin->db[i] = RefNC(db);
  }

  /* Decrement reference count of every Yorick symbols provided by the old
     plugin, if any, and release memory. */
  if (old_plugin) cleanup();
}


/* Remove the references set by the plugin and release memory used by the
   plugin. */
static void cleanup(void)
{
  plugin_t *plugin = old_plugin;
  if (plugin) {
    unsigned int i, n = plugin->nfuncs + plugin->nvars;
    for (i=0 ; i<n ; ++i) {
      DataBlock *db = plugin->db[i];
      plugin->db[i] = (DataBlock *)0; /* BEFORE Unref() */
      Unref(db); /* remove reference set by the plugin */
    }
    old_plugin = (plugin_t *)0; /* better never free than free twice */
    if (plugin->path) p_free(plugin->path);
    p_free(plugin);
  }
}

static void report_error(const char *text, const char *dlfunc,
			 const char *dlerr)
{
#define MSGSIZ 256
  char msg[MSGSIZ], *ptr = msg;
  size_t maxlen = MSGSIZ - 1, len;
#define APPEND(STR)					\
  if ((len = strlen(STR)) > maxlen) len = maxlen;	\
  if (len) memcpy(ptr, STR, len);			\
  ptr += len;						\
  maxlen -= len

  if (text && text[0]) {
    APPEND(text);
  }
  if (dlfunc && dlfunc[0]) {
    text = " -- error in ";
    APPEND(text);
    APPEND(dlfunc);
    if (dlerr && dlerr[0]) {
      text = "(): ";
      APPEND(text);
      APPEND(dlerr);
    } else {
      text = "()";
      APPEND(text);
    }
  } else {
    if (dlerr && dlerr[0]) {
      text = " -- ";
      APPEND(text);
      APPEND(dlerr);
    }
  }
  ptr[0] = 0;
#undef APPEND
#undef MSGSIZ
  YError(msg);
}

static void safe_func(int argc)
{
  YError("attempting to execute function from un-loaded plugin");
}

/*---------------------------------------------------------------------------*/
