/*
 * yeti_tiff.c --
 *
 *	Implement support for TIFF images in Yorick.
 *
 *-----------------------------------------------------------------------------
 *
 *	Copyright (C) 2003 Eric THIEBAUT.
 *
 *	This file is part of Yeti.
 *
 *	Yeti is  free software;  you can redistribute  it and/or  modify it
 *	under the terms  of the GNU General Public  License as published by
 *	the Free Software  Foundation; either version 2 of  the License, or
 *	(at your option) any later version.
 *
 *	Yeti is distributed in the hope that it will be useful, but WITHOUT
 *	ANY WARRANTY; without even  the implied warranty of MERCHANTABILITY
 *	or FITNESS  FOR A PARTICULAR  PURPOSE.  See the GNU  General Public
 *	License for more details.
 *
 *	You should have  received a copy of the  GNU General Public License
 *	along with  Yeti (file "COPYING"  in the top source  directory); if
 *	not, write to the Free  Software Foundation, Inc., 59 Temple Place,
 *	Suite 330,  Boston, MA  02111-1307 USA.
 *
 *-----------------------------------------------------------------------------
 *
 * History:
 *	$Id: yeti_tiff.c,v 1.3 2005/05/24 13:22:40 eric Exp $
 *	$Log: yeti_tiff.c,v $
 *	Revision 1.3  2005/05/24 13:22:40  eric
 *	 - macro USE_TIFF renamed as HAVE_TIFF
 *	 - change in error messages
 *
 *	Revision 1.2  2004/05/06 14:59:00  eric
 *	 - Check macros SAMPLEFORMAT_COMPLEXINT and SAMPLEFORMAT_COMPLEXIEEEFP
 *	   which may not be defined for old versions of TIFF library.
 *
 *	Revision 1.1  2003/08/26 14:39:47  eric
 *	Initial revision
 *
 */

#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include "play.h"
#include "pstdlib.h"
#include "ydata.h"
#include "yio.h"

/*---------------------------------------------------------------------------*/
/* PUBLIC ROUTINES */
extern BuiltIn Y_tiff_open;
extern BuiltIn Y_tiff_read_directory;
extern BuiltIn Y_tiff_read_image;
extern BuiltIn Y_tiff_read_pixels;
extern BuiltIn Y_tiff_debug;

/*---------------------------------------------------------------------------*/
#ifndef HAVE_TIFF
# define HAVE_TIFF 0
#endif
#if HAVE_TIFF
#include <tiff.h>
#include <tiffio.h>

/*---------------------------------------------------------------------------*/
/* DATA TYPES */
typedef struct _member member_t;

/* PRIVATE DATA */
static char message[2048];
static int debug = 0;

/* PRIVATE ROUTINES */
static void  push_string_value(const char *value);
static void *push_new_array(StructDef *base, long dim1, long dim2, long dim3);
static void *push_workspace(long nbytes);
static void  load_pixels(TIFF *tiff);
static int   get_boolean(Symbol *s);
static void  error_handler(const char* module, const char* fmt, va_list ap);
static void  warning_handler(const char* module, const char* fmt, va_list ap);
static member_t *find_member(char *name, long index, int stop_on_error);
static int cmapbits(unsigned int n,
		    const uint16 r[],
		    const uint16 g[],
		    const uint16 b[]);
static void missing_required_tag(const char *tagname);

/*---------------------------------------------------------------------------*/
/* OPAQUE OBJECTS */

extern PromoteOp PromXX;
extern UnaryOp ToAnyX, NegateX, ComplementX, NotX, TrueX;
extern BinaryOp AddX, SubtractX, MultiplyX, DivideX, ModuloX, PowerX;
extern BinaryOp EqualX, NotEqualX, GreaterX, GreaterEQX;
extern BinaryOp ShiftLX, ShiftRX, OrX, AndX, XorX;
extern BinaryOp AssignX, MatMultX;
extern UnaryOp EvalX, SetupX, PrintX;
extern MemberOp GetMemberX;

typedef struct _object object_t;
static object_t *get_object(Symbol *stack);
static void free_tiff(void *addr);
static UnaryOp print_tiff;
static void get_member(Operand *op, char *name);

static Operations tiff_ops= {
  &free_tiff, T_OPAQUE, 0, T_STRING, "TIFF file handle",
  {&PromXX, &PromXX, &PromXX, &PromXX, &PromXX, &PromXX, &PromXX, &PromXX},
  &ToAnyX, &ToAnyX, &ToAnyX, &ToAnyX, &ToAnyX, &ToAnyX, &ToAnyX,
  &NegateX, &ComplementX, &NotX, &TrueX,
  &AddX, &SubtractX, &MultiplyX, &DivideX, &ModuloX, &PowerX,
  &EqualX, &NotEqualX, &GreaterX, &GreaterEQX,
  &ShiftLX, &ShiftRX, &OrX, &AndX, &XorX,
  &AssignX, &EvalX, &SetupX, &get_member, &MatMultX, &print_tiff
};

struct _object {
  int references;                   /* reference counter */
  Operations *ops;                  /* virtual function table */
  TIFF *handle;                     /* TIFF file handle */
  char *path;                       /* full path */
};

/*
 * The tags understood by libtiff,  the number of parameter values, and the
 * expected types for the parameter values are shown below.  The data types
 * are: char* is  null-terminated string and corresponds to  the ASCII data
 * type; uint16 is  an unsigned 16-bit value; uint32  is an unsigned 32-bit
 * value; uint16* is an array of unsigned 16-bit values.  void* is an array
 * of data values of unspecified type.
 *
 * Consult the  TIFF specification for  information on the meaning  of each
 * tag.
 *
 * Tag Name                        Count  Types              Notes
 * TIFFTAG_ARTIST                  1      char*
 * TIFFTAG_BADFAXLINES             1      uint32
 * TIFFTAG_BITSPERSAMPLE           1      uint16             -
 * TIFFTAG_CLEANFAXDATA            1      uint16
 * TIFFTAG_COLORMAP                3      uint16*            1<<BitsPerSample arrays
 * TIFFTAG_COMPRESSION             1      uint16             -
 * TIFFTAG_CONSECUTIVEBADFAXLINES  1      uint32
 * TIFFTAG_COPYRIGHT               1      char*
 * TIFFTAG_DATETIME                1      char*
 * TIFFTAG_DOCUMENTNAME            1      char*
 * TIFFTAG_DOTRANGE                2      uint16
 * TIFFTAG_EXTRASAMPLES            2      uint16,uint16*     - count & types array
 * TIFFTAG_FAXMODE                 1      int                - G3/G4 compression pseudo-tag
 * TIFFTAG_FAXFILLFUNC             1      TIFFFaxFillFunc    G3/G4 compression pseudo-tag
 * TIFFTAG_FILLORDER               1      uint16             -
 * TIFFTAG_GROUP3OPTIONS           1      uint32             -
 * TIFFTAG_GROUP4OPTIONS           1      uint32             -
 * TIFFTAG_HALFTONEHINTS           2      uint16
 * TIFFTAG_HOSTCOMPUTER            1      char*
 * TIFFTAG_IMAGEDESCRIPTION        1      char*
 * TIFFTAG_IMAGEDEPTH              1      uint32             -
 * TIFFTAG_IMAGELENGTH             1      uint32
 * TIFFTAG_IMAGEWIDTH              1      uint32             -
 * TIFFTAG_INKNAMES                1      char*
 * TIFFTAG_INKSET                  1      uint16             -
 * TIFFTAG_JPEGTABLES              2      uint32*,void*      - count & tables
 * TIFFTAG_JPEGQUALITY             1      int                JPEG pseudo-tag
 * TIFFTAG_JPEGCOLORMODE           1      int                - JPEG pseudo-tag
 * TIFFTAG_JPEGTABLESMODE          1      int                - JPEG pseudo-tag
 * TIFFTAG_MAKE                    1      char*
 * TIFFTAG_MATTEING                1      uint16             -
 * TIFFTAG_MAXSAMPLEVALUE          1      uint16
 * TIFFTAG_MINSAMPLEVALUE          1      uint16
 * TIFFTAG_MODEL                   1      char*
 * TIFFTAG_ORIENTATION             1      uint16
 * TIFFTAG_PAGENAME                1      char*
 * TIFFTAG_PAGENUMBER              2      uint16
 * TIFFTAG_PHOTOMETRIC             1      uint16
 * TIFFTAG_PLANARCONFIG            1      uint16             -
 * TIFFTAG_PREDICTOR               1      uint16             -
 * TIFFTAG_PRIMARYCHROMATICITIES   1      float*             6-entry array
 * TIFFTAG_REFERENCEBLACKWHITE     1      float*             - 2*SamplesPerPixel array
 * TIFFTAG_RESOLUTIONUNIT          1      uint16
 * TIFFTAG_ROWSPERSTRIP            1      uint32             - must be > 0
 * TIFFTAG_SAMPLEFORMAT            1      uint16             -
 * TIFFTAG_SAMPLESPERPIXEL         1      uint16             - value must be <= 4
 * TIFFTAG_SMAXSAMPLEVALUE         1      double
 * TIFFTAG_SMINSAMPLEVALUE         1      double
 * TIFFTAG_SOFTWARE                1      char*
 * TIFFTAG_STONITS                 1      double             -
 * TIFFTAG_SUBFILETYPE             1      uint32
 * TIFFTAG_SUBIFD                  2      uint16,uint32*     count & offsets array
 * TIFFTAG_TARGETPRINTER           1      char*
 * TIFFTAG_THRESHHOLDING           1      uint16
 * TIFFTAG_TILEDEPTH               1      uint32             -
 * TIFFTAG_TILELENGTH              1      uint32             - must be a multiple of 8
 * TIFFTAG_TILEWIDTH               1      uint32             - must be a multiple of 8
 * TIFFTAG_TRANSFERFUNCTION        1 or 3 =                  uint16*1<<BitsPerSample entry arrays
 * TIFFTAG_XPOSITION               1      float
 * TIFFTAG_XRESOLUTION             1      float
 * TIFFTAG_WHITEPOINT              1      float*             2-entry array
 * TIFFTAG_YCBCRCOEFFICIENTS       1      float*             - 3-entry array
 * TIFFTAG_YCBCRPOSITIONING        1      uint16             -
 * TIFFTAG_YCBCRSAMPLING           2      uint16             -
 * TIFFTAG_YPOSITION               1      float
 * TIFFTAG_YRESOLUTION             1      float
 * TIFFTAG_ICCPROFILE              2      uint32,void*       count, profile data*
 * - Tag may not have its values changed once data is written.
 * =  If  SamplesPerPixel is one, then a single array is passed; otherwise
 * three arrays should be passed.
 * * The contents of this field are quite complex.  See  The  ICC  Profile
 * Format  Specification, Annex B.3 "Embedding ICC Profiles in TIFF Files"
 * (available at http://www.color.org) for an explanation.
 */


#if 0

static void get_tag_uint16(object_t *this, member_t *m);
static void get_tag_uint32(object_t *this, member_t *m);
static void get_tag_int(object_t *this, member_t *m);
static void get_tag_float(object_t *this, member_t *m);
static void get_tag_double(object_t *this, member_t *m);
static void get_tag_string(object_t *this, member_t *m);
static void get_tag_colormap(object_t *this, member_t *m);

static void set_tag_read_only(object_t *this, member_t *m, Symbol *s);
#if 0
static void set_tag_uint16(object_t *this, member_t *m, Symbol *s);
static void set_tag_uint32(object_t *this, member_t *m, Symbol *s);
static void set_tag_int(object_t *this, member_t *m, Symbol *s);
static void set_tag_float(object_t *this, member_t *m, Symbol *s);
static void set_tag_double(object_t *this, member_t *m, Symbol *s);
static void set_tag_string(object_t *this, member_t *m, Symbol *s);
static void set_tag_colormap(object_t *this, member_t *m, Symbol *s);
#else
#define set_tag_uint16   set_tag_read_only
#define set_tag_uint32   set_tag_read_only
#define set_tag_int      set_tag_read_only
#define set_tag_float    set_tag_read_only
#define set_tag_double   set_tag_read_only
#define set_tag_string   set_tag_read_only
#define set_tag_colormap set_tag_read_only
#endif

struct _member {
  int tag;
  char *name;
  void (*get)(object_t *this, member_t *m);
  void (*set)(object_t *this, member_t *m, Symbol *s);
  long index; /* in globTab */
};

#define TAG_UINT16(tag, name) {tag, name, get_tag_uint16, set_tag_uint16, -1}
#define TAG_UINT32(tag, name) {tag, name, get_tag_uint32, set_tag_uint32, -1}
#define TAG_INT(tag, name)    {tag, name, get_tag_int,    set_tag_int,    -1}
#define TAG_FLOAT(tag, name)  {tag, name, get_tag_float,  set_tag_float,  -1}
#define TAG_DOUBLE(tag, name) {tag, name, get_tag_double, set_tag_double, -1}
#define TAG_STRING(tag, name) {tag, name, get_tag_string, set_tag_string, -1}

static member_t member_table[] = {
  /* Due to the "slow" way I search this table, order is important (more
     frequently used tags should appears first). */
  TAG_STRING(TIFFTAG_ARTIST, "artist"),
  TAG_UINT16(TIFFTAG_BITSPERSAMPLE, "bitspersample"),
  TAG_UINT16(TIFFTAG_CLEANFAXDATA, "cleanfaxdata"),
  {TIFFTAG_COLORMAP, "colormap", get_tag_colormap, set_tag_colormap, -1},
  TAG_UINT16(TIFFTAG_COMPRESSION, "compression"),
  TAG_UINT32(TIFFTAG_CONSECUTIVEBADFAXLINES, "consecutivebadfaxlines"),
  TAG_STRING(TIFFTAG_COPYRIGHT, "copyright"),
  TAG_UINT16(TIFFTAG_DATATYPE, "datatype"),
  TAG_STRING(TIFFTAG_DATETIME, "datetime"),
  TAG_STRING(TIFFTAG_DOCUMENTNAME, "documentname"),
  TAG_INT(   TIFFTAG_FAXMODE, "faxmode"),
  TAG_UINT16(TIFFTAG_FILLORDER, "fillorder"),
  TAG_UINT32(TIFFTAG_GROUP3OPTIONS, "group3options"),
  TAG_UINT32(TIFFTAG_GROUP4OPTIONS, "group4options"),
  TAG_STRING(TIFFTAG_HOSTCOMPUTER, "hostcomputer"),
  TAG_UINT32(TIFFTAG_IMAGEDEPTH, "imagedepth"),
  TAG_STRING(TIFFTAG_IMAGEDESCRIPTION, "imagedescription"),
  TAG_UINT32(TIFFTAG_IMAGELENGTH, "imagelength"),
  TAG_UINT32(TIFFTAG_IMAGEWIDTH, "imagewidth"),
  TAG_STRING(TIFFTAG_INKNAMES, "inknames"),
  TAG_UINT16(TIFFTAG_INKSET, "inkset"),
  TAG_INT(   TIFFTAG_JPEGQUALITY, "jpegquality"),
  TAG_INT(   TIFFTAG_JPEGCOLORMODE, "jpegcolormode"),
  TAG_INT(   TIFFTAG_JPEGTABLESMODE, "jpegtablesmode"),
  TAG_STRING(TIFFTAG_MAKE, "make"),
  TAG_UINT16(TIFFTAG_MATTEING, "matteing"),
  TAG_UINT16(TIFFTAG_MAXSAMPLEVALUE, "maxsamplevalue"),
  TAG_UINT16(TIFFTAG_MINSAMPLEVALUE, "minsamplevalue"),
  TAG_STRING(TIFFTAG_MODEL, "model"),
  TAG_UINT16(TIFFTAG_ORIENTATION, "orientation"),
  TAG_STRING(TIFFTAG_PAGENAME, "pagename"),
/*TAG_UINT16(TIFFTAG_PAGENUMBER, "pagenumber"),*/
  TAG_UINT16(TIFFTAG_PHOTOMETRIC, "photometric"),
  TAG_UINT16(TIFFTAG_PLANARCONFIG, "planarconfig"),
  TAG_UINT16(TIFFTAG_PREDICTOR, "predictor"),
  TAG_UINT16(TIFFTAG_RESOLUTIONUNIT, "resolutionunit"),
  TAG_UINT32(TIFFTAG_ROWSPERSTRIP, "rowsperstrip"),
  TAG_UINT16(TIFFTAG_SAMPLEFORMAT, "sampleformat"),
  TAG_UINT16(TIFFTAG_SAMPLESPERPIXEL, "samplesperpixel"),
  TAG_DOUBLE(TIFFTAG_SMAXSAMPLEVALUE, "smaxsamplevalue"),
  TAG_DOUBLE(TIFFTAG_SMINSAMPLEVALUE, "sminsamplevalue"),
  TAG_STRING(TIFFTAG_SOFTWARE, "software"),
  TAG_FLOAT( TIFFTAG_XPOSITION, "xposition"),
  TAG_FLOAT( TIFFTAG_XRESOLUTION, "xresolution"),
  TAG_FLOAT( TIFFTAG_YPOSITION, "yposition"),
  TAG_FLOAT( TIFFTAG_YRESOLUTION, "yresolution"),
  TAG_DOUBLE(TIFFTAG_STONITS, "stonits"),
  TAG_UINT32(TIFFTAG_SUBFILETYPE, "subfiletype"),
  TAG_STRING(TIFFTAG_TARGETPRINTER, "targetprinter"),
  TAG_UINT16(TIFFTAG_THRESHHOLDING, "threshholding"),
  TAG_UINT32(TIFFTAG_TILEDEPTH, "tiledepth"),
  TAG_UINT32(TIFFTAG_TILELENGTH, "tilelength"),
  TAG_UINT32(TIFFTAG_TILEWIDTH, "tilewidth"),
  TAG_UINT16(TIFFTAG_YCBCRPOSITIONING, "ycbcrpositioning"),
/*TAG_UINT16(TIFFTAG_YCBCRSAMPLING, "ycbcrsampling"),*/
  {0, 0, 0, 0},
};
#undef TAG_UINT16
#undef TAG_UINT32
#undef TAG_INT
#undef TAG_FLOAT
#undef TAG_DOUBLE
#undef TAG_STRING

/* attempt to allow for '.' operator */
static void get_member(Operand *op, char *name)
{
  /* We can globalize name since it is always a valid symbol name. */
  member_t *m = find_member(name, Globalize(name, strlen(name)), 1);
  object_t *this = (object_t *)op->value;
  m->get(this, m);
  PopTo(op->owner);
}

#define GET_MEMBER(name, type_t, pusher)				  \
static void name(object_t *this, member_t *m)				  \
{									  \
  type_t value;								  \
  message[0] = 0;							  \
  if (TIFFGetFieldDefaulted(this->handle, m->tag, &value)) pusher(value); \
  else PushDataBlock(RefNC(&nilDB));					  \
}
GET_MEMBER(get_tag_int,    int,    PushLongValue)
GET_MEMBER(get_tag_uint16, uint16, PushLongValue)
GET_MEMBER(get_tag_uint32, uint32, PushLongValue)
GET_MEMBER(get_tag_float,  float,  PushDoubleValue)
GET_MEMBER(get_tag_double, double, PushDoubleValue)
GET_MEMBER(get_tag_string, char *, push_string_value)
#undef GET_MEMBER
static void get_tag_colormap(object_t *this, member_t *m)
{
  TIFF *tiff = this->handle;
  uint16 *red, *green, *blue, bitsPerSample;

#if 0
  uint16 photometric;
  if (! TIFFGetFieldDefaulted(tiff, TIFFTAG_PHOTOMETRIC, &photometric))
    YError("cannot get TIFF photometric");
  if (photometric == PHOTOMETRIC_PALETTE) get_palette;
  else no_palette;
#endif
  message[0] = 0;
  if (! TIFFGetFieldDefaulted(tiff, TIFFTAG_BITSPERSAMPLE, &bitsPerSample))
    YError("cannot get TIFF bits per sample");
  if (TIFFGetFieldDefaulted(tiff, TIFFTAG_COLORMAP, &red, &green, &blue)) {
    unsigned int i, number = 1<<(unsigned int)bitsPerSample;
    int colormapBits = cmapbits(number, red, green, blue);
    unsigned char *cmap = push_new_array(&charStruct, number, 3, 0);
    if (colormapBits == 8) {
      for (i=0 ; i<number ; ++i) cmap[i] = red[i];
      cmap += number;
      for (i=0 ; i<number ; ++i) cmap[i] = green[i];
      cmap += number;
      for (i=0 ; i<number ; ++i) cmap[i] = blue[i];
    } else {
      for (i=0 ; i<number ; ++i) cmap[i] = red[i]>>8;
      cmap += number;
      for (i=0 ; i<number ; ++i) cmap[i] = green[i]>>8;
      cmap += number;
      for (i=0 ; i<number ; ++i) cmap[i] = blue[i]>>8;
    }
  } else {
    /* Unable to get colormap. */
    if (message[0]) YError(message);
    PushDataBlock(RefNC(&nilDB));
  }
}

#else

struct _member {
  int   type; /* data type, one of the following: */
#define TAG_UINT16   0
#define TAG_UINT32   1
#define TAG_INT      2
#define TAG_FLOAT    3
#define TAG_DOUBLE   4
#define TAG_STRING   5
#define TAG_COLORMAP 6
  char *name;
  int   tag;
  long  index; /* in globTab */
};

static member_t member_table[] = {
  {TAG_STRING,                 "artist", TIFFTAG_ARTIST,                 -1},
  {TAG_UINT16,          "bitspersample", TIFFTAG_BITSPERSAMPLE,          -1},
  {TAG_UINT16,           "cleanfaxdata", TIFFTAG_CLEANFAXDATA,           -1},
  {TAG_COLORMAP,             "colormap", TIFFTAG_COLORMAP,               -1},
  {TAG_UINT16,            "compression", TIFFTAG_COMPRESSION,            -1},
  {TAG_UINT32, "consecutivebadfaxlines", TIFFTAG_CONSECUTIVEBADFAXLINES, -1},
  {TAG_STRING,              "copyright", TIFFTAG_COPYRIGHT,              -1},
  {TAG_UINT16,               "datatype", TIFFTAG_DATATYPE,               -1},
  {TAG_STRING,               "datetime", TIFFTAG_DATETIME,               -1},
  {TAG_STRING,           "documentname", TIFFTAG_DOCUMENTNAME,           -1},
  {TAG_INT,                   "faxmode", TIFFTAG_FAXMODE,                -1},
  {TAG_UINT16,              "fillorder", TIFFTAG_FILLORDER,              -1},
  {TAG_UINT32,          "group3options", TIFFTAG_GROUP3OPTIONS,          -1},
  {TAG_UINT32,          "group4options", TIFFTAG_GROUP4OPTIONS,          -1},
  {TAG_STRING,           "hostcomputer", TIFFTAG_HOSTCOMPUTER,           -1},
  {TAG_UINT32,             "imagedepth", TIFFTAG_IMAGEDEPTH,             -1},
  {TAG_STRING,       "imagedescription", TIFFTAG_IMAGEDESCRIPTION,       -1},
  {TAG_UINT32,            "imagelength", TIFFTAG_IMAGELENGTH,            -1},
  {TAG_UINT32,             "imagewidth", TIFFTAG_IMAGEWIDTH,             -1},
  {TAG_STRING,               "inknames", TIFFTAG_INKNAMES,               -1},
  {TAG_UINT16,                 "inkset", TIFFTAG_INKSET,                 -1},
  {TAG_INT,               "jpegquality", TIFFTAG_JPEGQUALITY,            -1},
  {TAG_INT,             "jpegcolormode", TIFFTAG_JPEGCOLORMODE,          -1},
  {TAG_INT,            "jpegtablesmode", TIFFTAG_JPEGTABLESMODE,         -1},
  {TAG_STRING,                   "make", TIFFTAG_MAKE,                   -1},
  {TAG_UINT16,               "matteing", TIFFTAG_MATTEING,               -1},
  {TAG_UINT16,         "maxsamplevalue", TIFFTAG_MAXSAMPLEVALUE,         -1},
  {TAG_UINT16,         "minsamplevalue", TIFFTAG_MINSAMPLEVALUE,         -1},
  {TAG_STRING,                  "model", TIFFTAG_MODEL,                  -1},
  {TAG_UINT16,            "orientation", TIFFTAG_ORIENTATION,            -1},
  {TAG_STRING,               "pagename", TIFFTAG_PAGENAME,               -1},
/*{TAG_UINT16,             "pagenumber", TIFFTAG_PAGENUMBER,             -1},*/
  {TAG_UINT16,            "photometric", TIFFTAG_PHOTOMETRIC,            -1},
  {TAG_UINT16,           "planarconfig", TIFFTAG_PLANARCONFIG,           -1},
  {TAG_UINT16,              "predictor", TIFFTAG_PREDICTOR,              -1},
  {TAG_UINT16,         "resolutionunit", TIFFTAG_RESOLUTIONUNIT,         -1},
  {TAG_UINT32,           "rowsperstrip", TIFFTAG_ROWSPERSTRIP,           -1},
  {TAG_UINT16,           "sampleformat", TIFFTAG_SAMPLEFORMAT,           -1},
  {TAG_UINT16,        "samplesperpixel", TIFFTAG_SAMPLESPERPIXEL,        -1},
  {TAG_DOUBLE,        "smaxsamplevalue", TIFFTAG_SMAXSAMPLEVALUE,        -1},
  {TAG_DOUBLE,        "sminsamplevalue", TIFFTAG_SMINSAMPLEVALUE,        -1},
  {TAG_STRING,               "software", TIFFTAG_SOFTWARE,               -1},
  {TAG_FLOAT,               "xposition", TIFFTAG_XPOSITION,              -1},
  {TAG_FLOAT,             "xresolution", TIFFTAG_XRESOLUTION,            -1},
  {TAG_FLOAT,               "yposition", TIFFTAG_YPOSITION,              -1},
  {TAG_FLOAT,             "yresolution", TIFFTAG_YRESOLUTION,            -1},
  {TAG_DOUBLE,                "stonits", TIFFTAG_STONITS,                -1},
  {TAG_UINT32,            "subfiletype", TIFFTAG_SUBFILETYPE,            -1},
  {TAG_STRING,          "targetprinter", TIFFTAG_TARGETPRINTER,          -1},
  {TAG_UINT16,          "threshholding", TIFFTAG_THRESHHOLDING,          -1},
  {TAG_UINT32,              "tiledepth", TIFFTAG_TILEDEPTH,              -1},
  {TAG_UINT32,             "tilelength", TIFFTAG_TILELENGTH,             -1},
  {TAG_UINT32,              "tilewidth", TIFFTAG_TILEWIDTH,              -1},
  {TAG_UINT16,       "ycbcrpositioning", TIFFTAG_YCBCRPOSITIONING,       -1},
/*{TAG_UINT16,          "ycbcrsampling", TIFFTAG_YCBCRSAMPLING,          -1},*/
  {0, 0, 0, 0},
};

static void get_member(Operand *op, char *name)
{
  /* We can globalize name since it is always a valid symbol name. */
  long len = strlen(name);
  member_t *m = find_member(name, Globalize(name, len), 1);
  object_t *this = (object_t *)op->value;
  TIFF *tiff = this->handle;
  int status;

  message[0] = 0;
  switch (m->type) {
#define GET_FIELD(type_t, pusher) {				\
    type_t value;						\
    status = TIFFGetFieldDefaulted(tiff, m->tag, &value);	\
    if (status) pusher(value); }
  case TAG_UINT16: GET_FIELD(uint16, PushLongValue) break;
  case TAG_UINT32: GET_FIELD(uint32, PushLongValue) break;
  case TAG_INT:    GET_FIELD(int,    PushLongValue) break;
  case TAG_FLOAT:  GET_FIELD(float,  PushDoubleValue) break;
  case TAG_DOUBLE: GET_FIELD(double, PushDoubleValue) break;
  case TAG_STRING: GET_FIELD(char *, push_string_value) break;
#undef GET_FIELD
  case TAG_COLORMAP:
    {
      uint16 *r, *g, *b, bitsPerSample;
      status = TIFFGetFieldDefaulted(tiff, TIFFTAG_COLORMAP, &r, &g, &b);    
      if (status) {
	if (TIFFGetFieldDefaulted(tiff, TIFFTAG_BITSPERSAMPLE,
				  &bitsPerSample)) {
	  unsigned int i, number = 1<<(unsigned int)bitsPerSample;
	  unsigned char *cmap = push_new_array(&charStruct, number, 3, 0);
	  if (cmapbits(number, r, g, b) == 8) {
	    for (i=0 ; i<number ; ++i) cmap[i] = r[i];
	    cmap += number;
	    for (i=0 ; i<number ; ++i) cmap[i] = g[i];
	    cmap += number;
	    for (i=0 ; i<number ; ++i) cmap[i] = b[i];
	  } else {
	    for (i=0 ; i<number ; ++i) cmap[i] = r[i]>>8;
	    cmap += number;
	    for (i=0 ; i<number ; ++i) cmap[i] = g[i]>>8;
	    cmap += number;
	    for (i=0 ; i<number ; ++i) cmap[i] = b[i]>>8;
	  }
	} else {
	  YError("cannot get TIFF bits per sample");
	}
      }
    }
    break;
  default:
    status = 0;
    sprintf(message, "unkonwn data type for TIFF field \"%.40s%s\"",
	    name, (len>40 ? "..." : ""));
  }
  if (! status) {
    if (message[0]) YError(message);
    PushDataBlock(RefNC(&nilDB));
  }
  PopTo(op->owner);
}
#endif

static member_t *find_member(char *name, long index, int stop_on_error)
{
  member_t *m;
  if (index >= 0) {
    /* Valid index in globTab. */
    for (m=member_table ; m->index ; ++m) {
      if (m->index == index) return m;
    }
  } else {
    /* We don't want to globalize name in case name is
       not a valid symbol name. */
    int c = name[0];
    for (m=member_table ; m->name ; ++m) {
      if (m->name[0] == c && ! strcmp(m->name, name)) return m;
    }
  }
  if (stop_on_error) {
    sprintf(message, "non-existing TIFF field \"%.40s%s\"",
	    name, (strlen(name)>40 ? "..." : ""));
    YError(message);
  }
  return 0;
}

static void error_handler(const char* module, const char* fmt, va_list ap)
{
  char *ptr = message;
  strcpy(ptr, "TIFF");
  if (module) {
    strcat(ptr, " [");
    strcat(ptr, module);
    strcat(ptr, "]: ");
  } else {
    strcat(ptr, ": ");
  }
  vsprintf(ptr + strlen(ptr), fmt, ap);
}

static void warning_handler(const char* module, const char* fmt, va_list ap)
{
  if (debug) {
    fputs("TIFF WARNING", stderr);
    if (module) {
      fputs(" [", stderr);
      fputs(module, stderr);
      fputs("]: ", stderr);
    } else {
      fputs(": ", stderr);
    }
    vfprintf(stderr, fmt, ap);
    fputs("\n", stderr);
    fflush(stderr);
  }
}

/* free_tiff is automatically called by Yorick to delete an object instance
   that is no longer referenced */
static void free_tiff(void *addr)
{
  if (addr) {
    object_t *this = (object_t *)addr;
    if (this->handle) TIFFClose(this->handle);
    if (this->path) p_free(this->path);
    p_free(this);
  }
}

/* print_tiff is used by Yorick's info command */
static void print_tiff(Operand *op)
{
  object_t *this = (object_t *)op->value;
  ForceNewline();
  PrintFunc(this->ops->typeName);
  PrintFunc(": path=\"");
  PrintFunc(this->path);
  PrintFunc("\"");
  ForceNewline();
}

static void bad_arg_list(const char *function)
{
  sprintf(message, "bad argument list to %s function", function);
  YError(message);
}

void Y_tiff_debug(int nargs)
{
  int prev = debug;
  if (nargs!=1) bad_arg_list("tiff_debug");
  debug = get_boolean(sp);
  PushIntValue(prev);
}

void Y_tiff_open(int nargs)
{
  static int first_time = 1;
  Symbol *stack= sp-nargs+1;
  object_t *this;
  char *filename, *filemode, *fullname;

  /* Initialization. */
  if (first_time) {
    member_t *m;
    TIFFSetErrorHandler(error_handler);
    TIFFSetWarningHandler(warning_handler);
    for (m=member_table ; m->name ; ++m) m->index = Globalize(m->name, 0);
    first_time = 0;
  }
  message[0] = 0;

  if (nargs<1 || nargs>2) bad_arg_list("tiff_open");
  filename = YGetString(stack);
  filemode = (nargs >= 2 ? YGetString(stack+1) : "r");
  fullname = YExpandName(filename);

  /* Push new opaque object on the stack (which will be automatically
     destroyed in case of error). */
  this = p_malloc(sizeof(*this));
  this->references = 0;
  this->ops = &tiff_ops;
  this->handle = NULL;
  this->path = fullname;
  PushDataBlock(this);

  /* Open file. */
  this->handle = TIFFOpen(fullname, filemode);
  if (! this->handle) YError(message);
}

void Y_tiff_read_directory(int nargs)
{
  int result;
  if (nargs != 1) bad_arg_list("tiff_read_directory");
  message[0] = 0;
  result = TIFFReadDirectory(get_object(sp)->handle);
  if (! result && message[0]) YError(message);
  PushIntValue(result);
}


void Y_tiff_read_image(int nargs)
{
  Symbol *stack= sp-nargs+1;
  uint16 photometric, bitsPerSample;
  uint32 width, height, depth;
  void *raster;
  object_t *this;
  TIFF *tiff;
  int stopOnError;

  if (nargs < 1 || nargs > 2) bad_arg_list("tiff_read_image");
  this = get_object(stack);
  tiff = this->handle;
  stopOnError = (nargs >= 2 ? YGetInteger(stack + 1) : 0);

  /* Get TIFF information to figure out the image type. */
  message[0] = 0; /* reset error message buffer */
  if (! TIFFGetFieldDefaulted(tiff, TIFFTAG_PHOTOMETRIC, &photometric))
    missing_required_tag("photometric");
  if (! TIFFGetFieldDefaulted(tiff, TIFFTAG_IMAGEDEPTH, &depth))
    missing_required_tag("depth");
  if (depth != 1) YError("TIFF depth != 1 not yet supported");

  /* Push image on top of the stack. */
  switch (photometric) {
  case PHOTOMETRIC_MINISWHITE: /* Grey and binary images. */
  case PHOTOMETRIC_MINISBLACK: /* Grey and binary images. */
    load_pixels(tiff);
    break;

  case PHOTOMETRIC_RGB: /* RGB Full color images */
  case PHOTOMETRIC_PALETTE: /* palette color images */
    /* Read RGBA image (do not stop on error). */
    if (! TIFFGetFieldDefaulted(tiff, TIFFTAG_BITSPERSAMPLE, &bitsPerSample))
      missing_required_tag("bitsPerSample");
    if (! TIFFGetFieldDefaulted(tiff, TIFFTAG_IMAGEWIDTH, &width))
      missing_required_tag("imageWidth");
    if (! TIFFGetFieldDefaulted(tiff, TIFFTAG_IMAGELENGTH, &height))
      missing_required_tag("imageLength");
    raster = push_new_array(&charStruct, 4, width, height);
    if (! TIFFReadRGBAImage(tiff, width, height, raster, stopOnError)) {
      if (! message[0])
	strcpy(message, "TIFFReadRGBAImage failed to read complete image");
      if (stopOnError) YError(message);
      fprintf(stderr, "TIFF WARNING: %s\n", message);
    }
    break;

  default:
    YError("unknown photometric in TIFF file");
  }
}

void Y_tiff_read_pixels(int nargs)
{
  Symbol *stack= sp-nargs+1;
  if (nargs != 1) bad_arg_list("tiff_read_pixels");
  load_pixels(get_object(stack)->handle);
}

#if 0
void Y_tiff_SetField(int nargs)
{
  Symbol *stack= sp-nargs+1;
  object_t *this;
  TIFF *tiff;

  if (nargs < 1 && nargs % 2 != 1) bad_arg_list("TIFFSetField");
  this = get_object(stack);
  tiff = this->handle;
}
#endif

static void missing_required_tag(const char *tagname)
{
  if (! message[0]) sprintf(message, "missing required TIFF tag \"%s\"",
			    tagname);
  YError(message);
}

#if 0 /* unused */
static void set_tag_read_only(object_t *this, member_t *m, Symbol *s)
{
  sprintf(message, "TIFF field \"%s\" is readonly", m->name);
  YError(message);
}
#endif /* unused */

#if 0 /* unused */
static void cannot_set_member(member_t *m);
static void cannot_set_member(member_t *m)
{
  if (! message[0])
    sprintf(message, "cannot set value of TIFF field \"%s\"", m->name);
  YError(message);
}
#endif /* unused */

static int cmapbits(unsigned int n,
		    const uint16 r[],
		    const uint16 g[],
		    const uint16 b[])
{
  unsigned int i;
  for (i=0 ; i<n ; ++i) {
    if (r[i] >= 256 || g[i] >= 256 || b[i] >= 256) return 16;
  }
  return 8; /* assume 8-bit colormap */
}

static object_t *get_object(Symbol *stack)
{
  object_t *obj;
  Symbol *s = (stack->ops == &referenceSym) ? &globTab[stack->index] : stack;
  if (s->ops != &dataBlockSym || s->value.db->ops != &tiff_ops) {
    YError("expecting a TIFF handle");
  }
  obj = (object_t *)s->value.db;
#if 0
  if (s != stack) {
    /* Replace reference onto the stack (equivalent to the statement
       ReplaceRef(stack); see ydata.c for actual code of this routine). */
    stack->value.db = Ref(s->value.db);
    stack->ops = &dataBlockSym;     /* change ops only AFTER value updated */
  }
#endif
  return obj;
}

/*---------------------------------------------------------------------------*/

static void *push_new_array(StructDef *base, long dim1, long dim2, long dim3)
{
  extern Dimension *tmpDims;
  Array *array;
  if (tmpDims) {
    Dimension *_dims = tmpDims;
    tmpDims = NULL;
    FreeDimension(_dims);
  }
  if (dim1 >= 1) {
    tmpDims = NewDimension(dim1, 1, tmpDims);
    if (dim2 >= 1) {
      tmpDims = NewDimension(dim2, 1, tmpDims);
      if (dim3 >= 1) tmpDims = NewDimension(dim3, 1, tmpDims);
    }
  }
  array = (Array *)PushDataBlock(NewArray(base, tmpDims));
  switch (base->dataOps->typeID) {
  case T_CHAR:    return array->value.c;
  case T_SHORT:   return array->value.s;
  case T_INT:     return array->value.i;
  case T_LONG:    return array->value.l;
  case T_FLOAT:   return array->value.f;
  case T_DOUBLE: /* same as complex */
  case T_COMPLEX: return array->value.d;
  case T_STRING:  return array->value.q;
  case T_POINTER: return array->value.p;
  }
  YError("unexpected array data type");
  return NULL;
}

static void *push_workspace(long nbytes)
{
  extern Dimension *tmpDims;
  if (tmpDims) {
    Dimension *_dims = tmpDims;
    tmpDims = NULL;
    FreeDimension(_dims);
  }
  tmpDims = NewDimension(nbytes, 1, tmpDims);
  return ((Array *)PushDataBlock(NewArray(&charStruct, tmpDims)))->value.c;
}

/* 
 * TIFFTAG_PHOTOMETRIC             262     // photometric interpretation
 *     PHOTOMETRIC_MINISWHITE      0       // min value is white
 *     PHOTOMETRIC_MINISBLACK      1       // min value is black
 *     PHOTOMETRIC_RGB             2       // RGB color model
 *     PHOTOMETRIC_PALETTE         3       // color map indexed
 *     PHOTOMETRIC_MASK            4       // $holdout mask
 *     PHOTOMETRIC_SEPARATED       5       // !color separations
 *     PHOTOMETRIC_YCBCR           6       // !CCIR 601
 *     PHOTOMETRIC_CIELAB          8       // !1976 CIE L*a*b*
 *     PHOTOMETRIC_ITULAB          10      // ITU L*a*b*
 *     PHOTOMETRIC_LOGL            32844   // CIE Log2(L)
 *     PHOTOMETRIC_LOGLUV          32845   // CIE Log2(L) (u',v')
 */
static void load_pixels(TIFF *tiff)
{
  void *buf, *raster;
  tstrip_t strip, numberOfStrips;
  uint16 orientation;
  uint16 sampleFormat;
  uint16 samplesPerPixel;
  uint16 bitsPerSample;
  uint16 planarConfig;
  uint16 photometric;
  uint32 width, height, depth;
  uint32 rowsPerStrip;
  uint32 stripSize;
  uint32 x, y, y1, y0, rowLength, rowSize;
  int single, complexData = 0;
  StructDef *sdef;
  
  /* Get information about pixels organization. */
  message[0] = 0;
  if (TIFFIsTiled(tiff))
    YError("reading of tiled TIFF images not yet implemented");
  if (! TIFFGetFieldDefaulted(tiff, TIFFTAG_PHOTOMETRIC, &photometric))
    missing_required_tag("photometric");
  if (! TIFFGetFieldDefaulted(tiff, TIFFTAG_SAMPLESPERPIXEL, &samplesPerPixel))
    missing_required_tag("samplesPerPixel");
  if (! TIFFGetFieldDefaulted(tiff, TIFFTAG_BITSPERSAMPLE, &bitsPerSample))
    missing_required_tag("bitsPerSample");
  if (! TIFFGetFieldDefaulted(tiff, TIFFTAG_IMAGEWIDTH,  &width))
    missing_required_tag("imageWidth");
  if (! TIFFGetFieldDefaulted(tiff, TIFFTAG_IMAGELENGTH, &height))
    missing_required_tag("imageLength");
  if (! TIFFGetFieldDefaulted(tiff, TIFFTAG_PLANARCONFIG, &planarConfig))
    missing_required_tag("planarConfig");
  if (! TIFFGetFieldDefaulted(tiff, TIFFTAG_IMAGEDEPTH,  &depth))
    missing_required_tag("imageDepth");
  if (! TIFFGetFieldDefaulted(tiff, TIFFTAG_ROWSPERSTRIP, &rowsPerStrip))
    missing_required_tag("rowsPerStrip");
  if (! TIFFGetFieldDefaulted(tiff, TIFFTAG_SAMPLEFORMAT, &sampleFormat))
    missing_required_tag("sampleFormat");
  if (! TIFFGetFieldDefaulted(tiff, TIFFTAG_ORIENTATION, &orientation))
    missing_required_tag("orientation");
    
  /* Figure out which data type to use for the result. */
  sdef = NULL;
  if (sampleFormat == SAMPLEFORMAT_UINT ||
      sampleFormat == SAMPLEFORMAT_INT) {
    if (bitsPerSample == 1 ||
	bitsPerSample == 2 ||
	bitsPerSample == 4 ||
	bitsPerSample == 8) {
      sdef = &charStruct;
    } else if (bitsPerSample == 8*sizeof(long)) {
      sdef = &longStruct;
    } else if (bitsPerSample == 8*sizeof(short)) {
      sdef = &shortStruct;
    } else if (bitsPerSample == 8*sizeof(int)) {
      sdef = &intStruct;
    }
  } else if (sampleFormat == SAMPLEFORMAT_IEEEFP) {
    if (bitsPerSample == 8*sizeof(float)) {
      sdef = &floatStruct;
    } else if (bitsPerSample == 8*sizeof(double)) {
      sdef = &doubleStruct;
    }
#ifdef SAMPLEFORMAT_COMPLEXINT
  } else if (sampleFormat == SAMPLEFORMAT_COMPLEXINT) {
    complexData = 1;
    sdef = &complexStruct;
#endif
#ifdef SAMPLEFORMAT_COMPLEXIEEEFP
  } else if (sampleFormat == SAMPLEFORMAT_COMPLEXIEEEFP) {
    complexData = 1;
    sdef = &complexStruct;
#endif
  }
  if (complexData
#ifdef SAMPLEFORMAT_COMPLEXIEEEFP
      && (sampleFormat != SAMPLEFORMAT_COMPLEXIEEEFP ||
	  bitsPerSample == 8*sizeof(double))
#endif
      ) {
    YError("unsupported TIFF complex sample");
  }
  if (! sdef) {
    sprintf(message,
	    "unsupported TIFF image data/format (BitsPerSample=%d, SampleFormat=%d)",
	    (int)bitsPerSample, (int)sampleFormat);
    YError(message);
  }
  
  /* We need to push 2 buffers onto the stack: a scanline/tile buffer and
     an image buffer. */
  CheckStack(2);

  /* Allocate a strip buffer onto the stack. */
  numberOfStrips = TIFFNumberOfStrips(tiff);
  stripSize = TIFFStripSize(tiff);
  buf = push_workspace(stripSize);

  /* Allocate raster image. */
  single = (samplesPerPixel == 1);
  if (single) raster = push_new_array(sdef, width, height, 0);
  else        raster = push_new_array(sdef, samplesPerPixel, width, height);
  
  rowLength = samplesPerPixel*width;
  rowSize = ((bitsPerSample + 7)/8)*rowLength;
  if (planarConfig == PLANARCONFIG_CONTIG) {
    for (strip=0, y0=0 ; y0<height ; y0=y1, ++strip) {
      /* Localization of next strip. */
      unsigned char *dst = ((unsigned char *)raster) + rowSize*y0;
      unsigned char *src = buf;

      /* Read next strip. */
      if (strip >= numberOfStrips) YError("bad number of strips");
      TIFFReadEncodedStrip(tiff, strip, buf, stripSize);
      
      /* Convert every rows in the strip. */
      if ((y1 = y0 + rowsPerStrip) > height) y1 = height;
      if (bitsPerSample%8 == 0) {
	/* Just copy values. */
	memcpy(dst, src, rowSize*(y1 - y0));
      } else if (bitsPerSample == 1) {
	unsigned int mask=1, value;
	uint32 n = (rowLength/8)*8;
	for (y=y0 ; y<y1 ; ++y, dst+=rowLength) {
	  for (x=0 ; x < n ; x+=8) {
	    value = *src++;
	    dst[x]   =  value     & mask;
	    dst[x+1] = (value>>1) & mask;
	    dst[x+2] = (value>>2) & mask;
	    dst[x+3] = (value>>3) & mask;
	    dst[x+4] = (value>>4) & mask;
	    dst[x+5] = (value>>5) & mask;
	    dst[x+6] = (value>>6) & mask;
	    dst[x+7] = (value>>6) & mask;
	  }
	  if (x < rowLength) {
	    for (value = *src++ ; x < rowLength ; ++x, value >>= 1) {
	      dst[x] = value & mask;
	    }
	  }
	}
      } else if (bitsPerSample == 2) {
	unsigned int mask=3, value;
	uint32 n = (rowLength/4)*4;
	for (y=y0 ; y<y1 ; ++y, dst+=rowLength) {
	  for (x=0 ; x < n ; x+=4) {
	    value = *src++;
	    dst[x]   =  value     & mask;
	    dst[x+1] = (value>>2) & mask;
	    dst[x+2] = (value>>4) & mask;
	    dst[x+3] = (value>>6) & mask;
	  }
	  if (x < rowLength) {
	    for (value = *src++ ; x < rowLength ; ++x, value >>= 2) {
	      dst[x] = value & mask;
	    }
	  }
	}
      } else if (bitsPerSample == 4) {
	unsigned int mask=15, value;
	uint32 n = (rowLength/2)*2;
	for (y=y0 ; y<y1 ; ++y, dst+=rowLength) {
	  for (x=0 ; x < n ; x+=2) {
	    value = *src++;
	    dst[x]   =  value     & mask;
	    dst[x+2] = (value>>4) & mask;
	  }
	  if (x < rowLength) {
	    for (value = *src++ ; x < rowLength ; ++x, value >>= 4) {
	      dst[x] = value & mask;
	    }
	  }
	}
      }
    }
  } else {
    /* Must be PLANARCONFIG_SEPARATE. */
    YError("unsupported TIFF planar configuration");
  }

#warning "take orientation into account..."
#if 0
  int transpose=0, revx=0, revy=0;
  switch (orientation) {
  case ORIENTATION_TOPLEFT:  /* row 0 top, col 0 lhs */
    revx = 0; revy = 1; transpose = 0; break;
  case ORIENTATION_TOPRIGHT: /* row 0 top, col 0 rhs */
    revx = 1; revy = 1; transpose = 0; break;
  case ORIENTATION_BOTRIGHT: /* row 0 bottom, col 0 rhs */
    revx = 1; revy = 0; transpose = 0; break;
  case ORIENTATION_BOTLEFT:  /* row 0 bottom, col 0 lhs */
    revx = 0; revy = 0; transpose = 0; break;
  case ORIENTATION_LEFTTOP:  /* row 0 lhs, col 0 top */
    revx = 0; revy = 0; transpose = 1; break;
  case ORIENTATION_RIGHTTOP: /* row 0 rhs, col 0 top */
    revx = 0; revy = 0; transpose = 1; break;
  case ORIENTATION_RIGHTBOT: /* row 0 rhs, col 0 bottom */
    revx = 0; revy = 0; transpose = 1; break;
  case ORIENTATION_LEFTBOT:  /* row 0 lhs, col 0 bottom */
    revx = 0; revy = 0; transpose = 1; break;
  }
  if (revx) {
    if (samplesPerPixel == 1) {
      for (y=0 ; y<height ; ++y) {
	type_t *p0 = ((type_t *)raster) + y*rowLength;
	type_t *p1 = p0 + rowLength - 1;
	while (p0 < p1) {
	  tmp = *p0;
	  *p0++ = *p1;
	  *p1-- = tmp;
	}
      }
    } else {
      unsigned int s, n=samplesPerPixel;
      for (y=0 ; y<height ; ++y) {
	type_t *p0 = ((type_t *)raster) + y*rowLength;
	type_t *p1 = p0 + rowLength - samplesPerPixel;
	while (p0 < p1) {
	  for (s=0 ; s<n ; ++s) {
	    type_t tmp = p0[s]; p0[s] = p1[s]; p1[s] = tmp;
	  }
	  p0 += n;
	  p1 -= n;
	}
      }
      }
    }
  }
#endif

  /* Revert to MinIsBlack photometric (if possible). */
  if (photometric == PHOTOMETRIC_MINISWHITE) {
    int warn=0;
    uint32 i, number = width*height*samplesPerPixel;
    if (sampleFormat == SAMPLEFORMAT_UINT ||
	sampleFormat == SAMPLEFORMAT_INT) {
      switch (bitsPerSample) {
#define REVERSE(type_t, ones) {					\
        type_t complement = ones, *ptr = (type_t *)raster;	\
	for (i=0 ; i<number ; ++i) ptr[i] ^= complement; }
      case   1: REVERSE(unsigned char, 0x00000001) break;
      case   2: REVERSE(unsigned char, 0x00000003) break;
      case   4: REVERSE(unsigned char, 0x0000000f) break;
      case   8: REVERSE(unsigned char, 0x000000ff) break;
      case  16: REVERSE(uint16,        0x0000ffff) break;
      case  32: REVERSE(uint32,        0xffffffff) break;
#undef REVERSE
      default:
	warn=1;
      }
    } else {
      warn=1;
    }
    if (warn) {
      fprintf(stderr, "warning: TIFF photometric MinIsWhite left unchanged\n");
    }
  }
}

static void push_string_value(const char *value)
{
  ((Array *)PushDataBlock(NewArray(&stringStruct, NULL)))->value.q[0] = 
    (value ? p_strcpy((char *)value) : NULL);
}

static int get_boolean(Symbol *s)
{
  if (s->ops == &referenceSym) s = &globTab[s->index];
  if (s->ops == &intScalar)    return (s->value.i != 0);
  if (s->ops == &longScalar)   return (s->value.l != 0L);
  if (s->ops == &doubleScalar) return (s->value.d != 0.0);
  if (s->ops == &dataBlockSym) {
    Operand op;
    s->ops->FormOperand(s, &op);
    if (! op.type.dims) {
      switch (op.ops->typeID) {
      case T_CHAR:   return (*(char   *)op.value != 0);
      case T_SHORT:  return (*(short  *)op.value != 0);
      case T_INT:    return (*(int    *)op.value != 0);
      case T_LONG:   return (*(long   *)op.value != 0L);
      case T_FLOAT:  return (*(float  *)op.value != 0.0F);
      case T_DOUBLE: return (*(double *)op.value != 0.0);
      case T_COMPLEX:return (((double *)op.value)[0] != 0.0 ||
			     ((double *)op.value)[1] != 0.0);
      case T_STRING: return (op.value != NULL);
      case T_VOID:   return 0;
      }
    }
  }
  YError("bad non-boolean argument");
  return 0; /* avoid compiler warning */
}

/*---------------------------------------------------------------------------*/
#else /* not HAVE_TIFF */

static void no_tiff_support();

void Y_tiff_debug(int nargs) { no_tiff_support(); }
void Y_tiff_open(int nargs) { no_tiff_support(); }
void Y_tiff_read_directory(int nargs) { no_tiff_support(); }
void Y_tiff_read_image(int nargs) { no_tiff_support(); }
void Y_tiff_read_pixels(int nargs) { no_tiff_support(); }

static void no_tiff_support()
{
  YError("your version of Yorick was compiled without support for TIFF files");
}
#endif /* not HAVE_TIFF */
/*---------------------------------------------------------------------------*/
